// 
// blasvis - graphics routines for BLASIUS / 3dvom
//
// Blasvis : Main class to start the program and
// provide the UI.
//
// (C) 2001-2004 Andrew Ross <aross@env.leeds.ac.uk>


#include "dataplot.h"
#include "blasfile.h"
#include "vomfile.h"
#include "blasvis.h"
#include "escstring.h"
#include <iostream>
#include <cstdlib>
#include <cctype>
#include <cstring>
#include <cmath>
#include <limits.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

using namespace std;

//
// Constructor - does nothing at present
//
Blasvis::Blasvis() 
{
  data1 = NULL;
  data2 = NULL;
  data3 = NULL;
  vect = NULL;
  tdata1 = NULL;
  tdata2 = NULL;
  temperature = NULL;
  runno = 0;
  time = 0;
  verbose = false;
  model = none;

  for (int i=0;i<10;i++) {
    varvals[i] = 0;  
  }
}

Blasvis::~Blasvis() 
{
}

//
// Run an instance of the program using this class.
//
int Blasvis::Run(int argc, char** argv) 
{
  char optchar;
  char cmdfile[PATH_MAX];
  char datfile[PATH_MAX];
  bool gotfile = false;
  bool cmd = false;
  bool dat = false;
  bool endian = false;
  bool norec = true;
  bool quiet = false;
  useLES = false;
  progname = argv[0] + strlen(argv[0]);
  int ver;
  int maj, min, rev;
  
#ifdef HAVE_GETOPT_H  
  int opt_index = 0;
  static struct option long_options[] = {
	  {"version",0,NULL,'V'},
	  {"help",0,NULL,'h'}
  };
#endif

  // This will only work on unix with a "/" as a directory separator
  while ((progname > argv[0]) && (*(progname-1) != '/')) {
    progname--;
  }
  

  if (strncmp(progname,"blasvis",7) == 0) {
    model = blasius;
    ver = 0;
  }
  else if (strncmp(progname,"lemvis",6) == 0) {
    model = lem;
    ver = 0;
  }
  else if (strncmp(progname,"vomvis",6) == 0) {
    model = vom;
    ver = 265;
  }
  else if (strncmp(progname,"nvomvis",7) == 0) {
    model = nvom;
    ver = 160;
  }
  else {
    model = none;
    ver = 0;
  }

  // Set up dev null as an output if running command files
  foutput.open("/dev/null");

#ifdef HAVE_GETOPT_H
  while ( (optchar = getopt_long(argc,argv, "bhf:c:elLm:n:qrRvV",
				  long_options, &opt_index)) != EOF) {
#else
  while ( (optchar = getopt(argc,argv, "bhf:c:elLm:n:qrRvV")) != EOF) {
#endif

    switch (optchar) {
    case 'h':
    case '?':
      // Print help and exit   
      cout << "Usage: " << progname << " [OPTIONS]\n";
      cout << "Extract data and plot graphs from BLASIUS / 3D(n)vom data files.\n\n";
      cout << "  -b             Read BLASIUS files\n";
      cout << "  -c FILE        Read in commands from FILE\n";
      cout << "  -e             Convert between big and little endian\n";
      cout << "  -f FILE        Select the data file FILE\n";
      cout << "  -h, --help     Print this help and exit\n";
      cout << "  -l             Use LES average not instantaneous fields in BLASIUS\n";
      cout << "  -L             Read LEM files\n";
      cout << "  -m ver         Read 3dvom files\n";
      cout << "  -n ver         Read 3dnvom files\n";
      cout << "  -q             Supress screen output\n";
      cout << "  -r             No record length in files (BLASIUS 4, default)\n";
      cout << "  -R             Record length in files (BLASIUS 3)\n";
      cout << "  -v             Verbose command mode\n";
      cout << "  -V, --version  Print version and exit\n\n";
      cout << "Please report bugs to Andrew Ross <aross@env.leeds.ac.uk>\n" << endl;
      return 0;
    case 'b':
      model = blasius;
      break;
    case 'f':
      dat = true;
      strncpy(datfile,optarg,PATH_MAX);
      datfile[PATH_MAX-1]='\0';
      break;
    case 'c':
      cmd = true;
      strncpy(cmdfile,optarg,PATH_MAX);
      cmdfile[PATH_MAX-1]='\0';
      break;
    case 'e':
      endian = true;
      break;
    case 'l':
      useLES = true;
      break;
    case 'L':
      model = lem;
      break;
    case 'm':
      if (sscanf(optarg,"%i.%i.%i",&maj,&min,&rev) != 3) {
	cerr << "Invalid version number " << optarg << endl;
      }
      else {
	model = vom;
	ver = maj*100+min*10+rev;
      }
      break;
    case 'n':
      if (sscanf(optarg,"%i.%i.%i",&maj,&min,&rev) != 3) {
	cerr << "Invalid version number " << optarg << endl;
      }
      else {
	model = nvom;
	ver = maj*100+min*10+rev;
      }
      break;
    case 'q':
      quiet = true;
      break;
    case 'r':
      norec = true;
      break;
    case 'R':
      norec = false;
      break;
    case 'v':
      verbose = true;
      break;
    case 'V':
      // Print out version and copyright and exit.
      cout << progname << " " << version << "\n\n";
      cout << "Copyright (C) 2001-2004 Andrew Ross\n";
      cout << "This is free software; see the source for copying conditions.  There is NO\n";
      cout << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << endl;
      return 0;
    }
  }

  if ((model == blasius)||(model == lem)) {
    datafile = new BlasFile();
  }
  else if ((model == vom)||(model == nvom)) {
    datafile = new VomFile();
  }
  else {
    cerr << "Unknown model type" << endl;
    return -1;
  }

  datafile->SetVersion(ver,model);
  datafile->SwapEndian(endian);
  datafile->NoFortranRecords(norec);
  datafile->UseLESAv(useLES);

  if (dat) {
    datafile->OpenFile(datfile);
    runno = GetRunNumber(datfile);
    //time = GetTime(datfile);
    time = (int)floor(datafile->GetParm("TIME"));
    plot.SetRunParms(runno,time);
    temperature = datafile->GetData("T");
    gotfile = true;
  }

  if (cmd) {
      finput.open(cmdfile);
      input = &finput;
      if (verbose) {
	output = &cout;
	*output << "Command file " << cmdfile << endl;
	if (dat) {
	  *output << "Data file " << datfile << endl;
	}
      }
      else {
	output = &foutput;
      }
      isCmdFile = true;
  }
  else {
    input = &cin;
    if (quiet) {
      output = &foutput;
    }
    else {
      output = &cout;
    }
    isCmdFile = false;
  }

//   if (!gotfile) {
//     GetInputFile();
//   }

  MainMenu();
  
  return 0;

}

void Blasvis::MainMenu() 
{
  char line[10];
  char menuopt;

  while (true) {

    *output << "\n " << progname << " - graphics for orographic models\n\n";
    *output << " 1. Input data file options    J. Set variable\n";
    *output << " 2. Command file               K. Calculate drag\n";
    *output << " 3. Set interpolation grid     L. Get value at specific point\n";
    *output << " 4. Select data set            N. Plot misc\n";
    *output << " 5. Select output device       O. Show headers\n";
    *output << " 6. Set range of axes          P. Output time series to file\n";
    *output << " 7. Set contour spacing        R. Plot time series\n";
    *output << " 8. Set plot titles\n";
    *output << " 9. Set plot options\n";
    *output << " A. Start plot\n";
    *output << " B. Plot data set\n";
    *output << " C. Finish plot\n";
    *output << " D. Output data to file\n";
    *output << " E. Output data statistics\n";
    *output << " F. Start new page\n";
    *output << " G. Check for separation\n";
    *output << " H. Data derivatives\n";
    *output << " I. Normalise data\n";
    *output << " Q. Quit\n" << endl;

    input->getline(line,10);
  
    menuopt = tolower(line[0]);

    switch (menuopt) {
    case '1':
      InputFileOpts();
      break;
    case '2':
      GetCommandFile();
      break;
    case '3':
      SetInterpolationGrid();
      break;
    case '4':
      SelectData();
      break;
    case '5':
      SelectOutputDev();
      break;
    case '6':
      SetRange();
      break;
    case '7':
      SetContours();
      break;
    case '8':
      SetTitles();
      break;
    case '9':
      SetPlotOpts();
      break;
    case 'a':
      StartPlot();
      break;
    case 'b':
      if (data1 == NULL) {
	cerr << "Must select data set first" << endl;
      }
      else {
	PlotData();
      }
      break;
    case 'c':
      FinishPlot();
      break;
    case 'd':
      OutputData();
      break;
    case 'e':
      OutputDataStats();
      break;
    case 'f':
      NextPage();
      break;
    case 'g':
      CheckSeparation();
      break;
    case 'h':
      DerivData();
      break;
    case 'i':
      NormData();
      break;
    case 'j':
      SetVariable();
      break;
    case 'k':
      GetDrag();
      break;
    case 'l':
      GetValues();
      break;
    case 'n':
      PlotMisc();
      break;
    case 'o':
      if (datafile) {
	datafile->DisplayHeader();
      }
      break;
    case 'p':
      OutputTimeSeries();
      break;
    case 'r':
      PlotTimeSeries();
      break;
    case 'q':
      return;
    default:
      cerr << "Unknown option " << menuopt << endl;
      if (isCmdFile) return;
    }
  }
}

//
// Get name of Blasius data file
//
void Blasvis::InputFileOpts() {
  char fname[PATH_MAX];
  char menuopt, line[10];

  *output << "    Data file options\n";
  *output << " 1. Open data file\n";
  *output << " 2. Close data file\n";
  *output << " 3. Swap byte order\n";
  *output << " M. Return to main menu\n" << endl;

  do {
    input->getline(line,10);
    menuopt = tolower(line[0]);
    
    switch (menuopt) {
    case '1':
      *output << "Enter file name " << endl;
      input->getline(fname,PATH_MAX);
      
      datafile->OpenFile(fname);
      runno = GetRunNumber(fname);
      time = GetTime(fname);
      plot.SetRunParms(runno,time);
      temperature = datafile->GetData("T");
      break;
    case '2':
      datafile->CloseFile();
      break;
    case '3':
      *output << "Swap byte order of data (y/n)?" << endl;
      datafile->SwapEndian(GetYesNo());
      break;
    case 'm':
      break;
    default:
      menuopt = '\0';
    }
  } while (menuopt == '\0');
}

//
// Control whether input is from file or from keyboard
//
void Blasvis::GetCommandFile() {
  char fname[PATH_MAX];
  char line[10],menuopt;
  
  while (true) {
    *output << "    Command input\n" ;
    *output << " 1. Keyboard\n";
    *output << " 2. File\n";
    *output << " Q. Quit\n" << endl;
    
    input->getline(line,10);

    menuopt = tolower(line[0]);

    if (menuopt == '1') {
      input = &cin;
      output = &cerr;
      isCmdFile = false;
      return;
    }
    else if (menuopt == '2') {
      *output << "Enter file name " << endl;
      input->getline(fname,PATH_MAX);
      finput.open(fname);
      input = &finput;
      if (verbose) {
	output = &cout;
      }
      else {
	output = &foutput;
      }
      isCmdFile = true;
      return;
    }
    else if (menuopt == 'q') {
      return;
    }
  }      
}

void Blasvis::SetInterpolationGrid() {

  bool transform;
  int ix, iy;
  char line[80];

  *output << "Transform vertical coordinate? (y/n)" << endl;
  transform = GetYesNo();
  *output << "Set horizontal interpolation grid size" << endl;
  input->getline(line,80);
  ix = atoi(line);
  *output << "Set vertical interpolation grid size" << endl;
  input->getline(line,80);
  iy = atoi(line);

  plot.SetTransform(transform);
  plot.SetInterpolation(ix,iy);
  
}

 
void Blasvis::SelectData() {
  char dataopt, subopt;
  char label[8],line[10];
  DataVect *oldvect;
  char fname[PATH_MAX];

  *output << "    Select data \n\n";
  *output << " 0. 0-D\n";
  *output << " 1. 1-D\n";
  *output << " 2. 2-D\n";
  *output << " 3. 3-D\n";
  *output << " 4. Vector field\n";
  *output << " 5. LES average fields\n";
  *output << " 6. Time series\n";
  *output << " M. Return to main menu\n"<< endl;

  do {
    input->getline(line,10);
    dataopt = tolower(line[0]);

    switch (dataopt) {
    case '0':
      datafile->DisplayParms();
      break;
    case '1':    
      *output << "\n    1-D field to plot\n\n";
      *output << " 1. Background temperature (TI)\n";
      *output << " 2. Background energy (EI)\n";
      *output << " 3. Background water vapour (QI)\n";
      *output << " 4. Background wind (UI)\n";
      *output << " 5. Background wind (VI)\n";
      *output << " 6. User data set\n";
      *output << " 7. Background liquid water (QLI)\n";
      *output << " M. Return to main menu\n"<< endl;
      do {

       input->getline(line,10);
       subopt = tolower(line[0]);

       switch (subopt) {
       case '1':
         data1 = datafile->GetData("TI");
         break;
       case '2':
         data1 = datafile->GetData("EI");
         break;
       case '3':
         data1 = datafile->GetData("QI");
         break;
       case '4':
         data1 = datafile->GetData("UI");
         break;
       case '5':
         data1 = datafile->GetData("VI");
         break;
       case '6':
         *output << "Enter data label" << endl;
         input->getline(label,80);
         data1 = datafile->GetData(label);
         break;
       case '7':
         data1 = datafile->GetData("QLI");
         break;
       case 'm':
         break;
       default:
         cerr << "Invalid option - please try again" << endl;
         if (isCmdFile) return;
         subopt = '\0';
       }
      } while (subopt == '\0');
      if (subopt != 'm') {
       data2 = NULL;
       data3 = NULL;
       if ((vect != NULL) && (!vect->IsCached() )) {
         delete vect;
       }
       vect = NULL;
      }
      break;;
    case '2':
      *output << "\n    2-D field to plot\n\n";
      *output << " 1. Surface height (ZSN)\n";
      *output << " 2. Surface buoyancy flux (SFBFLX)\n";
      *output << " 3. Upstream slope surface stress (TAUX)\n";
      *output << " 4. Across slope surface stress (TAUY)\n";
      //*output << " 5. Streamfunction\n";
      *output << " 6. Boundary layer depth (BLDEPTH)\n";
      *output << " 7. User data set\n";
      *output << " M. Return to main menu\n"<< endl;

      do {
    
	input->getline(line,10);
      	subopt = tolower(line[0]);
      
	switch (subopt) {
	case '1':
	  data1 = datafile->GetData("ZSN");
	  break;
	case '2':
	  data1 = datafile->GetData("SFBFLX");
	  break;
	case '3':
	  data1 = datafile->GetData("TAUX");
	  break;
	case '4':
	  data1 = datafile->GetData("TAUY");
	  break;
	case '5':
	  cerr << "Not yet implemented" << endl;
	  if (isCmdFile) return;
	  //	data1 = datafile->GetData("ZSN");
	  break;
	case '6':
	  data1 = datafile->GetBLDepth();
	  break;
	case '7':
	  *output << "Enter data label" << endl;
	  input->getline(label,80);
	  data1 = datafile->GetData(label);
	  break;
	case 'm':
	  break;
	default:
	  cerr << "Invalid option - please try again" << endl;
	  if (isCmdFile) return;
	  subopt = '\0';
	}
      } while (subopt == '\0');
      if (subopt != 'm') {
	data2 = NULL;
	data3 = NULL;
	if ((vect != NULL) && (!vect->IsCached() )) {
	  delete vect;
	}
	vect = NULL;
      }
      break;
    case '3':
      *output << "\n    3-D field to plot\n\n";
      *output << " 1. Temperature (T)                  J. Tau_xx (TAU11)\n";
      *output << " 2. Energy (E)                       K. Tau_xy (TAU12)\n";
      *output << " 3. Water vapour (Q)                 L. Tau_xz (TAU13)\n";
      *output << " 4. Density (RHO)                    N. Tau_yy (TAU22)\n";
      *output << " 5. Pressure (P)                     O. Tau_yz (TAU23)\n";
      *output << " 6. Viscosity (VIS)                  P. Tau_zz (TAU33)\n";
      *output << " 7. Horizontal viscosity (VISH)      Q. <U^2> (<U^2>)\n";
      *output << " 8. Reference temp (THETAR)          R. <W^2> (<W^2>)\n";
      *output << " 9. Richardson number (RIGRAD)       S. <UW> (B13)\n";
      *output << " A. Windspeed                        T. Dissipation (EPSLN)\n";
      *output << " B. Relative speedup                 U. Streamfunction\n";
      *output << " C. Corrected pressure               V. B11\n";
      *output << " D. Temp perturbation (T')           W. B33\n";
      *output << " E. Potential vorticity (PV)         X. <UV> (B12)\n";
      *output << " F. Bernoulli potential              Y. <VW> (B23)\n";
      *output << " G. CFL timstep                      Z. PV source\n";
      *output << " H. User data set                    -. Height\n";
      *output << " I. PV (alt scheme)                  =. <V^2> (V^2>)\n";
      *output << " +. ASCII data set                   !. Liquid water \n";
      *output << "\n M. Return to main menu\n" << endl;

      do {
    
	input->getline(line,10);
      	subopt = tolower(line[0]);
      
	switch (subopt) {
	case '1':
	  data1 = temperature;
	  break;
	case '2':
	  data1 = datafile->GetData("E");
	  if (data1 == NULL) {
	    data1 = datafile->GetData("TKE");
	  }
	  break;
	case '3':
	  data1 = datafile->GetData("Q");
	  break;
	case '4':
	  data1 = datafile->GetData("RHO");
	  break;
	case '5':
	  if (useLES) 
	    data1 = datafile->GetData("PRAV");
	  else
	    data1 = datafile->GetData("P");
	  break;
	case '6':
	  data1 = datafile->GetData("VIS");
	  break;
	case '7':
	  data1 = datafile->GetData("VISH");
	  break;
	case '8':
	  data1 = datafile->GetData("THETAR");
	  break;
	case '9':
	  data1 = datafile->GetData("RIGRAD");
	  break;
	case 'a':
	  data1 = datafile->GetWindspeed();
	  break;
	case 'b':
	  data1 = datafile->GetSpeedup();
	  break;
	case 'c':
	  data1 = datafile->GetPressureCorr();
	  break;
	case 'd':
	  data1 = datafile->GetDeltaT();
	  break;
	case 'e':
	  data1 = datafile->GetPV();
	  break;
	case 'f':
	  data1 = datafile->GetBernoulli();
	  break;
	case 'g':
	  data1 = datafile->GetTimestep();
	  break;
	case 'h':
	  *output << "Enter data label" << endl;
	  input->getline(label,80);
	  data1 = datafile->GetData(label);
	  break;
	case 'i':
	  data1 = datafile->GetPVAlt();
	  break;
	case 'j':
	  if (useLES) 
	    data1 = datafile->GetData("T11AV");
	  else 
	    data1 = datafile->GetData("TAU11");
	  break;
	case 'k':
	  if (useLES) 
	    data1 = datafile->GetData("T12AV");
	  else 
	    data1 = datafile->GetData("TAU12");
	  break;
	case 'l':
	  if (useLES) 
	    data1 = datafile->GetData("T13AV");
	  else 	    
	    data1 = datafile->GetData("TAU13");
	  break;
	case 'n':
	  if (useLES) 
	    data1 = datafile->GetData("T22AV");
	  else 
	    data1 = datafile->GetData("TAU22");
	  break;
	case 'o':
	  if (useLES) 
	    data1 = datafile->GetData("T23AV");
	  else 
	    data1 = datafile->GetData("TAU23");
	  break;
	case 'p':
	  if (useLES) 
	    data1 = datafile->GetData("T33AV");
	  else 
	    data1 = datafile->GetData("TAU33");
	  break;
	case 'q':
	  if (useLES) {
	    data1 = datafile->GetData("U2BAR");
	  }
	  else {
	    data1 = datafile->GetData("E");
	    if (data1 != NULL) {
	      data2 = datafile->GetData("TAU11");
	      if (data2 != NULL) {
		// NB This syntax will make a copy and keep the original data unchanged
		data1 = data1->copy();
		*data1 *= 2.0/3.0;
		*data1 -= *data2;
	      }
	      else {
		data2 = datafile->GetData("B11");
		if (data2 != NULL) {
		  // NB This syntax will make a copy and keep the original data unchanged
		  data1 = data1->copy();
		  *data1 *= 2.0/3.0;
		  *data1 += *data2;
		}
	      }
	      if (data2 == NULL) {
		data1 = NULL;
		data2 = NULL;
	      }
	    }
	  }
	  break;
	case 'r':
	  if (useLES) {
	    data1 = datafile->GetData("W2BAR");
	  }
	  else {
	    data1 = datafile->GetData("E");
	    if (data1 != NULL) {
	      data2 = datafile->GetData("TAU33");
	      if (data2 != NULL) {
		// NB This syntax will make a copy and keep the original data unchanged
		data1 = data1->copy();
		*data1 *= 2.0/3.0;
		*data1 -= *data2;
	      }
	      else {
		data2 = datafile->GetData("B33");
		if (data2 != NULL) {
		  // NB This syntax will make a copy and keep the original data unchanged
		  data1 = data1->copy();
		  *data1 *= 2.0/3.0;
		  *data1 += *data2;
		}
	      }
	      if (data2 == NULL) {
		data1 = NULL;
		data2 = NULL;
	      }
	    }
	  }
	  break;
	case 's':
	  data1 = datafile->GetData("B13");
	  break;
	case 't':
	  data1 = datafile->GetData("EPSLN");
	  break;
	case 'u':
	  data1 = datafile->GetStreamFn();
	  break;
	case 'v':
	  data1 = datafile->GetData("B11");
	  break;
	case 'w':
	  data1 = datafile->GetData("B33");
	  break;
	case 'x':
	  data1 = datafile->GetData("B12");
	  break;
	case 'y':
	  data1 = datafile->GetData("B23");
	  break;
	case 'z':
	  data1 = datafile->GetPVSource();
	  break;
	case '-':
	  data1 = datafile->GetHeight();
	  break;
	case '=':
	  if (useLES) {
	    data1 = datafile->GetData("V2BAR");
	  }
	  else {
	    data1 = datafile->GetData("E");
	    if (data1 != NULL) {
	      data2 = datafile->GetData("TAU22");
	      if (data2 != NULL) {
		// NB This syntax will make a copy and keep the original data unchanged
		data1 = data1->copy();
		*data1 *= 2.0/3.0;
		*data1 -= *data2;
	      }
	      else {
		data2 = datafile->GetData("B11");
		data3 = datafile->GetData("B33");
		if ((data2 != NULL) && (data3 != NULL)) {
		  // NB This syntax will make a copy and keep the original data unchanged
		  data1 = data1->copy();
		  *data1 *= 2.0/3.0;
		  *data1 -= *data2;
		  *data1 -= *data3;
		}
	      }
	      if (data2 == NULL) {
		data1 = NULL;
		data2 = NULL;
		data3 = NULL;
	      }	    
	    }
	  }
	  break;
	case '+':
	  *output << "Enter file name " << endl;
	  input->getline(fname,PATH_MAX);
	  data1 = datafile->GetASCIIData(fname);
	  break;
	case '!':
	  data1 = datafile->GetData("QL");
	  break;
	case 'm':
	  break;
	default:
	  cerr << "Invalid option - please try again" << endl;
	  if (isCmdFile) return;
	  subopt = '\0';
	}
      } while (subopt == '\0');
      if (subopt != 'm') {
	data2 = NULL;
	data3 = NULL;
	if ((vect != NULL) && (!vect->IsCached() )) {
	  delete vect;
	}
	vect = NULL;
      }
      break;
    case '4':
      *output << "\n    3-D field to plot\n\n";
      *output << " 1. Velocity (U,V,W)\n";
      *output << " 2. Vorticity (VORTX,VORTY,VORTZ)\n";
      *output << " 3. Heat flux (HTFLXX,HTFLXY,HTFLXZ)\n";
      *output << " 4. PV flux\n";
      *output << " 5. Grad temp\n";
      *output << " 6. PV flux (alternate)\n";
      *output << " 7. Surface stress\n";      
      *output << " 8. User data set\n";
      *output << " M. Return to main menu\n" << endl;

      oldvect = vect;

      vect = NULL;

      do {
    
	input->getline(line,10);
      	subopt = tolower(line[0]);
      
	switch (subopt) {
	case '1':
	  if (useLES) {
	    data1 = datafile->GetData("UAV");
	    data2 = datafile->GetData("VAV");
	    data3 = datafile->GetData("WAV");
	  }
	  else {
	    data1 = datafile->GetData("U");
	    data2 = datafile->GetData("V");
	    data3 = datafile->GetData("W");
	  }
	  break;
	case '2':
	  data1 = datafile->GetData("VORTX");
	  data2 = datafile->GetData("VORTY");
	  data3 = datafile->GetData("VORTZ");
	  break;
	case '3':
	  data1 = datafile->GetData("HTFLXX");
	  data2 = datafile->GetData("HTFLXY");
	  data3 = datafile->GetData("HTFLXZ");
	  break;
	case '4':
	  vect = datafile->GetPVFlux();
	  data1 = vect->GetX();
	  data2 = vect->GetY();
	  data3 = vect->GetZ();
	  break;
	case '5':
	  vect = datafile->GetGradT();
	  data1 = vect->GetX();
	  data2 = vect->GetY();
	  data3 = vect->GetZ();
	  break;
	case '6':
	  vect = datafile->GetPVFluxAlt();
	  data1 = vect->GetX();
	  data2 = vect->GetY();
	  data3 = vect->GetZ();
	  break;
	case '7':
	  vect = datafile->GetTau();
	  data1 = vect->GetX();
	  data2 = vect->GetY();
	  data3 = vect->GetZ();
	  break;
	case '8':
	  *output << "Enter first data label" << endl;
	  input->getline(label,80);
	  data1 = datafile->GetData(label);
	  *output << "Enter second data label" << endl;
	  input->getline(label,80);
	  data2 = datafile->GetData(label);
	  *output << "Enter third data label" << endl;
	  input->getline(label,80);
	  data3 = datafile->GetData(label);
	case 'm':
	  break;
	default:
	  cerr << "Invalid option - please try again" << endl;
	  if (isCmdFile) return;
	  subopt = '\0';
	}
      } while (subopt == '\0');
      if (subopt != 'm') {
	if (data1 != NULL) {
	  data1->SetIsVector(1);
	  data2->SetIsVector(2);
	  if (data3)
	    data3->SetIsVector(3);
	}
	if ((oldvect != NULL) && (!oldvect->IsCached()) ) {
	  delete oldvect;
	}
	if (vect == NULL) {
	  vect = new DataVect(data1, data2, data3);
	}
      }
      else {
	vect = oldvect;
      }
      break;
    case '5':
      // Do LES average fields
      *output << "\n    3-D field to plot\n\n";
      *output << " 1. Velocity average (UAV,VAV,WAV)\n";
      *output << " 2. Velocity^2 average (U2AV,V2AV,W2AV)\n";
      *output << " 3. W^3 average (W3AV)\n";
      *output << " 4. UV average (UVAV)\n";
      *output << " 5. UW average (UWAV)\n";
      *output << " 6. VW average (VWAV)\n";
      *output << " 7. TAU11 average (T11AV)\n";
      *output << " 8. TAU12 average (T12AV)\n";
      *output << " 9. TAU13 average (T13AV)\n";
      *output << " A. TAU22 average (T22AV)\n";
      *output << " B. TAU23 average (T23AV)\n";
      *output << " C. TAU33 average (T33AV)\n";
      *output << " D. Pressure average (PRAV)\n";
      *output << " E. Pressure^2 average (PR2AV)\n";
      *output << " F. LES resolved u'w'\n";
      *output << " G. LES resolved u'u'\n";
      *output << " H. LES resolved v'v'\n";
      *output << " I. LES resolved w'w'\n";
      *output << " J. LES resolved TKE\n";
      *output << " K. Subgrid energy (SGEAV)\n";
      *output << " L. Subgrid dissipation (DISAV)\n";
      *output << "\n M. Return to main menu\n" << endl;

      do {
    
	input->getline(line,10);
      	subopt = tolower(line[0]);
      
	switch (subopt) {
	case '1':
	  data1 = datafile->GetData("UAV");
	  data2 = datafile->GetData("VAV");
	  data3 = datafile->GetData("WAV");
	  break;
	case '2':
	  data1 = datafile->GetData("U2AV");
	  data2 = datafile->GetData("V2AV");
	  data3 = datafile->GetData("W2AV");
	  break;
	case '3':
	  data1 = datafile->GetData("W3AV");
	  break;
	case '4':
	  data1 = datafile->GetData("UVAV");
	  break;
	case '5':
	  data1 = datafile->GetData("UWAV");
	  break;
	case '6':
	  data1 = datafile->GetData("VWAV");
	  break;
	case '7':
	  data1 = datafile->GetData("T11AV");
	  break;
	case '8':
	  data1 = datafile->GetData("T12AV");
	  break;
	case '9':
	  data1 = datafile->GetData("T13AV");
	  break;
	case 'a':
	  data1 = datafile->GetData("T22AV");
	  break;
	case 'b':
	  data1 = datafile->GetData("T23AV");
	  break;
	case 'c':
	  data1 = datafile->GetData("T33AV");
	  break;
	case 'd':
	  data1 = datafile->GetData("PRAV");
	  break;
	case 'e':
	  data1 = datafile->GetData("PR2AV");
	  break;
	case 'f':
	  data1 = datafile->GetLESuw();
	  break;
	case 'g':
	  data1 = datafile->GetLESuu();
	  break;
	case 'h':
	  data1 = datafile->GetLESvv();
	  break;
	case 'i':
	  data1 = datafile->GetLESww();
	  break;
	case 'j':
	  data1 = datafile->GetLESenergy();
	  break;
	case 'k':
	  data1 = datafile->GetData("SGEAV");
	  break;
	case 'l':
	  data1 = datafile->GetData("DISAV");
	  break;
	case 'm':
	  break;
	default:
	  cerr << "Invalid option - please try again" << endl;
	  if (isCmdFile) return;
	  subopt = '\0';
	}
	if (subopt != '0') {
	  if ((subopt == '1') || (subopt == '2')) {
	    oldvect = vect;
	    vect = NULL;
	    if (data1 != NULL) {
	      data1->SetIsVector(1);
	      data2->SetIsVector(2);
	      if (data3)
		data3->SetIsVector(3);
	    }
	    if ((oldvect != NULL) && (!oldvect->IsCached()) ) {
	      delete oldvect;
	    }
	    if (vect == NULL) {
	      vect = new DataVect(data1, data2, data3);
	    }
	  }
	  else if (subopt != 'm') {
	    data2 = NULL;
	    data3 = NULL;
	    if ((vect != NULL) && (!vect->IsCached() )) {
	      delete vect;
	    }
	    vect = NULL;
	  }
	}
      } while (subopt == '0');
      break;         
    case '6':
      if (model != blasius) {
	cerr << "Time series only available with BLASIUS" << endl;
	break;
      }
      *output << "\n    Time series to plot\n\n";
      *output << " 1. Single series\n";
      *output << " 2. Pair of series\n";
      *output << " M. Return to main menu\n" << endl;

      do {
    
	input->getline(line,10);
      	subopt = tolower(line[0]);
      
	switch (subopt) {
	case '1':
	  *output << "Enter data label" << endl;
	  input->getline(label,80);
	  tdata1 = ((BlasFile *) datafile)->GetTimeSeries(label);
	  tdata2 = NULL;
	  break;
	case '2':
	  *output << "Enter first data label" << endl;
	  input->getline(label,80);
	  tdata1 = ((BlasFile *) datafile)->GetTimeSeries(label);
	  *output << "Enter second data label" << endl;
	  input->getline(label,80);
	  tdata2 = ((BlasFile *) datafile)->GetTimeSeries(label);
	  break;
	default:
	  cerr << "Invalid option - please try again" << endl;
	  if (isCmdFile) return;
	  subopt = '\0';
	}
      } while (subopt == '\0');
    case 'm':
      break;
    default:
      cerr << "Invalid option - please try again" << endl;
      if (isCmdFile) return;
      dataopt = '\0';
    } 
  } while (dataopt == '\0');
}

void Blasvis::SelectOutputDev() {
  char devopt,line[10];
  DataPlot::devtype dev = DataPlot::none;

  *output << "    Select output device\n" << endl;
  *output << " Current device " << endl;
  *output << " 1. X Window (xwin)\n";
  *output << " 2. Mono postscript (ps)\n";
  *output << " 3. Colour postscript (psc)\n";
  *output << " 4. XFig (xfig)\n";
  *output << " 5. PBM (pbm)\n";
  *output << " 6. PNG (png)\n";
  *output << " 7. Save data to file\n" ;
  *output << " 8. JPEG (jpeg)\n";
  *output << " M. Return to main menu\n" << endl;

  while (dev == DataPlot::none) {
    input->getline(line,10);

    devopt = tolower(line[0]);

    switch (devopt) {
    case '1':
      dev = DataPlot::xwin;
      break;
    case '2':
      dev = DataPlot::ps;
      break;
    case '3':
      dev = DataPlot::psc;
      break;
    case '4':
      dev = DataPlot::xfig;
      break;
    case '5':
      dev = DataPlot::pbm;
      break;
    case '6':
      dev = DataPlot::png;
      break;
    case '7':
      dev = DataPlot::file;
      break;
    case '8':
      dev = DataPlot::jpeg;
      break;
    case 'm':
      return;
    default:
      cerr << "Unknown option - please try again" << endl;
      if (isCmdFile) return;
    }
  }

  plot.SetDevice(dev);
}

void Blasvis::SetRange() {
  bldata xmin, xmax, ymin, ymax, zmin, zmax;

  *output << "Set X axis automatically? (y/n)" << endl;
  if (GetYesNo()) {
    xmin = 0.0;
    xmax = 0.0;
  }
  else {
    *output << "Enter minimum and maximum values" << endl;
    xmin = GetReal();
    xmax = GetReal();
  }

  *output << "Set Y axis automatically? (y/n)" << endl;
  if (GetYesNo()) {
    ymin = 0.0;
    ymax = 0.0;
  }
  else {
    *output << "Enter minimum and maximum values" << endl;
    ymin = GetReal();
    ymax = GetReal();
  }

  *output << "Set Z axis automatically? (y/n)" << endl;
  if (GetYesNo()) {
    zmin = 0.0;
    zmax = 0.0;
  }
  else {
    *output << "Enter minimum and maximum values" << endl;
    zmin = GetReal();
    zmax = GetReal();
  }

  plot.SetRange(xmin,xmax,ymin,ymax,zmin,zmax);
}

void Blasvis::SetContours() {
  bldata zmin, zmax;
  int nlev;
  char line[80];

  *output << "Set contour range automatically? (y/n)" << endl;
  if (GetYesNo()) {
    zmin = 0.0;
    zmax = 0.0;
  }
  else {
    *output << "Enter minimum and maximum values" << endl;
    zmin = GetReal();
    zmax = GetReal();
  }

  do {
    *output << "Enter number of contour levels" << endl;
    input->getline(line,80);
    nlev = atoi(line);
  } while (nlev <= 0);

  plot.SetContours(zmin,zmax,nlev);
}

void Blasvis::SetTitles() {
  char title[80];
  char xtitle[40];
  char ytitle[40];
  char ztitle[40];

  *output << "Enter plot title" << endl;
  input->getline(title,80);

  *output << "Enter x axis title" << endl;
  input->getline(xtitle,40);

  *output << "Enter y axis title" << endl;
  input->getline(ytitle,40);

  *output << "Enter z axis title (surface plot)" << endl;
  input->getline(ztitle,40);

  plot.SetTitles(title,xtitle,ytitle,ztitle);
}

void Blasvis::SetPlotOpts() {
  
  char plotopt, line[80];
  int axis,ph;
  bldata alt, az;
  bool err, b1, b2;

  plotopt = '\0';
  err = false;

  while (plotopt != 'm') {
    if (err == false) {
      *output << "    Set plotting options\n\n";
      *output << " 1. Plot legend                     J. Set text scaling\n";
      *output << " 2. Plot hill                       K. Set plot aspect ratio\n";
      *output << " 3. Line plot offset\n";
      *output << " 4. Line plot scale\n";
      *output << " 5. Vector plot scale\n";
      *output << " 6. Invert horizontal axes\n";
      *output << " 7. Set colour scheme\n";
      *output << " 8. Set axis scaling factor\n";
      *output << " 9. Set data scaling factor\n";
      *output << " A. Plot mesh?\n";
      *output << " B. Set view angles\n";
      *output << " C. Set angle to rotate horizontal axes\n";
      *output << " D. Label line contours\n";
      *output << " E. Set line width\n";
      *output << " F. Set log contour spacing\n";
      *output << " G. Set line style\n";
      *output << " H. Set colour\n";
      *output << " I. Set vector style\n";
      *output << " M. Return to main menu\n" << endl;
    }
    else {
      err = false;
    }

    input->getline(line,10);

    plotopt = tolower(line[0]);

    switch (plotopt) {
    case '1':
      *output << "\nPlot legend? (y/n)" << endl;
      plot.SetPlotLegend(GetYesNo());
      break;
    case '2':
      ph = 0;
      *output << "\nPlot hill base? (y/n)" << endl;
      if (GetYesNo()) {
	ph |= DataPlot::PLOTBASE;
      }
      *output << "\nPlot hill intersection? (y/n)" << endl;
      if (GetYesNo()) {
	ph |= DataPlot::PLOTINTERSECT;
      }
      plot.SetPlotHill(ph);
      break;
    case '3':
      *output << "\nEnter offset" << endl;
      plot.SetLineOffset(GetReal());
      break;
    case '4':
      *output << "\nEnter scaling factor" << endl;
      plot.SetLineScale(GetReal());
      break;
    case '5':
      *output << "\nEnter scaling factor" << endl;
      plot.SetVectorScale(GetReal());
      break;
    case '6':
      *output << "\nInvert horizontal axes? (y/n)" << endl;
      plot.SetInvertAxes(GetYesNo());
      break;
    case '7':
      SetColScheme();
      break;
    case '8':
      *output << "\nDimension to scale" << endl;
      axis = GetDimOpt(1,false);
      *output << "Enter scaling factor" << endl;
      plot.SetAxisScaling(axis,GetReal());
      break;
    case '9':
      *output << "\nEnter scaling factor" << endl;
      plot.SetDataScaling(GetReal());
      break;
    case 'a':
      *output << "\nPlot mesh? (y/n)" << endl;
      plot.SetDrawMesh(GetYesNo());
      break;
    case 'b':
      *output << "\nAltitude angle" << endl;
      alt = GetReal();
      *output << "Azimuthal angle" << endl;
      az = GetReal();
      plot.SetViewAngle(alt,az);
      break;
    case 'c':
      *output << "\nRotate horizontal axes (angle in degrees, multiple of 90)" 
	      << endl;
      plot.SetRotateAxes(GetReal());
      break;
    case 'd':
      *output << "\nLabel line contours? (y/n)" << endl;
      plot.SetPlotContLabel(GetYesNo());
      break;
    case 'e':
      *output << "\nEnter line width" << endl;
      input->getline(line,80);
      plot.SetLineWidth(atoi(line));
      break;
    case 'f':
      *output << "\nUse logarithmic scale for contours (y/n)" << endl;
      plot.SetLogContour(GetYesNo());
      break;
    case 'g':
      *output << "\nEnter line style" << endl;
      input->getline(line,80);
      plot.SetLineStyle(atoi(line));
      break;
    case 'h':
      *output << "\nEnter colour" << endl;
      input->getline(line,80);
      plot.SetColour(atoi(line));
      break;
    case 'i':
      *output << "\nEnter vector style" << endl;
      input->getline(line,80);
      *output << "\nFilled arrow? (y/n)" << endl;
      b1 = GetYesNo();
      *output << "\nArrow end at point? (y/n)" << endl;
      b2 = GetYesNo();
      plot.SetVectorStyle(atoi(line),b1,b2);
      break;
    case 'j':
      *output << "\nSet text scaling" << endl;
      plot.SetTextScaling(GetReal());
    case 'k':
      *output << "\nAspect ratio" << endl;
      plot.SetAspectRatio(GetReal());
      break;
    case 'm':
      break;
    default:
      cerr << "Unknown option - please try again" << endl;
      if (isCmdFile) return;
      err = true;
    }
  }
}

//
// Set the colour scheme to use
//
void Blasvis::SetColScheme() {
  char plotopt, line[80], col;

  *output << "    Set colour scheme\n\n" ;
  *output << " 0. Default\n";
  *output << " 1. Monochrome\n";
  *output << " 2. Colour\n";
  *output << " 3. Monochrome (zero white)\n";
  *output << " 4. Colour (zero white)\n";
  *output << " 5. Reversed monochrome\n";
  *output << " 6. Reversed colour\n";
  *output << " 7. Reversed monochrome (zero white)\n";
  *output << " 8. Reversed colour (zero white)\n";
  *output << " M. Return to main menu\n" << endl;

  plotopt = '\0';

  while (plotopt != 'm') {
    input->getline(line,10);

    plotopt = tolower(line[0]);

    col = (int) plotopt - 48;
    if ((col >= 0) && (col <= 8)) {
      plot.SetColScheme(col);
      return;
    }
    else if (plotopt != 'm') {
      cerr << "Unknown option - please try again" << endl;
      if (isCmdFile) return;
    }
  }
}


void Blasvis::StartPlot() {

  char temp[PATH_MAX],devfile[80];

  if (plot.IsOpen()) {
    cerr << "Plot already open - please close this first" << endl;
    return;
  }

  // Get file name if necessary
  if (plot.NeedDevFile()) {
    *output << "Enter plot file name" << endl;
    input->getline(temp,80);
    StringExpand(temp,devfile,80);
    plot.SetDevFile(devfile);
  }
  plot.InitPlot();
}

void Blasvis::NextPage() {

  if (!plot.IsOpen()) {
    cerr << "Plot not open" << endl;
    return;
  }

  plot.NextPlot();
}

void Blasvis::PlotData() {
  int ndim,vdim,gdim;
  char graphopt,cmax;
  bldata gval,gval2,gval3;
  DataSet *d1 = NULL, *d2 = NULL;
  DataSet *extra;
  char line[80];

  int nstore = 100;
  int nmax = 1000;
  bldata *xgrid, *ygrid, *zgrid, *sgrid, *hgrid;
  int npts;

  if (data1 == NULL) {
    cerr << "Must read data before plotting it" << endl;
    return;
  }


  if (!plot.IsOpen()) {
    cerr << "Must start a plot before plotting data" << endl;
    return;
  }


  ndim = data1->GetDims();

  switch (ndim) {
  case 1:
  case 2:
  case 3:
    *output << "    2-D or 3-D data set\n\n";
    *output << " 1. 1-D section\n";
    *output << " 2. 2-D line contour plot\n";
    *output << " 3. 2-D colour contour plot\n";
    *output << " 4. 2-D surface plot\n";
    *output << " 5. 2-D trajectory\n";
    if (data2 != NULL) {
      *output << " 6. 2-D vector plot\n";
      *output << " 7. Hodograph\n";
      *output << " 8. 2-D point vector plot\n";
      cmax = '8';
    }
    else {
      cmax = '5';
    }
    *output << " M. Return to main menu\n";

    do {
      input->getline(line,10);      
      graphopt = tolower(line[0]);
    } while (((graphopt < '1') || (graphopt > cmax)) && (graphopt != 'm'));
    

    if ((data2 != NULL) && (graphopt >= '1') && graphopt <= '4') {
      *output << "Enter vector component to plot (x,y,z)" << endl;
      vdim = GetDimOpt(1,false);
      if (vdim == 1) {
	d1 = data1;
      }
      else if (vdim == 2) {
	d1 = data2;
      }
      else{
	d1 = data3;
      }
    }
    else {
      vdim = 1;
      d1 = data1;
    }
    
    switch (graphopt) {
    case '1':
      *output << "\nEnter dimension to plot (x,y,z,hx,hy,ix,iy,s,r,sx,sy)" << endl;
      gdim = GetDimOpt(2,true);
      *output << "Enter value for other dimensions" << endl;
      gval = GetReal();
      gval2 = GetReal();
      // Need 3 values to start a trajectory
      if ((gdim == 6)||(gdim == 7)) {
	gval3 = GetReal();
	datafile->GetTraj(gval,gval2,gval3,-1.0,nstore,nmax, &xgrid, &ygrid, 
			  &zgrid, &sgrid, &hgrid, &npts, (bool)(gdim-6));
      }
      else if ((gdim == 9)||(gdim == 10)) {
	gval3 = GetReal();
	datafile->GetLinearTraj(gval,gval2,gval3,-1.0,nstore,nmax, &xgrid, &ygrid, 
			  &zgrid, &sgrid, &hgrid, &npts, (bool)(gdim-9));
      }

      if (plot.GetPlotLegend()) {
	*output << "Enter label for legend" << endl;
	input->getline(line,10);
      }
      extra = NULL;
      switch (gdim) {
      case 15:
      case 25:
	extra = temperature;
	break;
      case 16:
      case 26:
	extra = datafile->GetStreamFn();
	break;
      } 
      if ((gdim == 6)||(gdim == 7)||(gdim == 9) ||(gdim == 10)) {
	plot.LinePlot(d1, npts, xgrid, ygrid, zgrid, hgrid, sgrid, line);
      }
      else {
	plot.LinePlot(d1, gdim, gval, gval2, line, extra);
      }
      break;
    case '2':
    case '3':
    case '4':
      *output << "\nEnter dimension to plot through (x,y,z,h,i)" << endl;
      gdim = GetDimOpt(1,true);
      *output << "Enter value for dimension" << endl;
      gval = GetReal();
      if (graphopt == '2') {
	plot.LineContPlot(d1, gdim, gval, temperature);
      }
      else if (graphopt == '3') {
	plot.ColContPlot(d1, gdim, gval, temperature);
      }
      else {
	plot.SurfacePlot(d1, gdim, gval, temperature);
      }
      break;
    case '5':
      // Plot trajectory
      *output << "\nEnter dimension to plot through (x,y,z)" << endl;
      gdim = GetDimOpt(1,false);
      *output << "Enter starting point for trajectory" << endl;
      gval = GetReal();
      gval2 = GetReal();
      gval3 = GetReal();
      datafile->GetTraj(gval,gval2,gval3,-1.0,nstore,nmax, &xgrid, &ygrid, 
			&zgrid, &sgrid, &hgrid, &npts);
      switch (gdim) {
      case 1:
	plot.LinePlot(zgrid, npts, ygrid, 1, 0,0,0,0);      
	break;
      case 2:
	plot.LinePlot(zgrid, npts, xgrid, 1, 0,0,0,0);      
	break;
      case 3:
	plot.LinePlot(ygrid, npts, ygrid, 1, 0,0,0,0);      
	break;
      }      
      break;
    case '6':
      *output << "\nEnter dimension to plot through (x,y,z,h,i)" << endl;
      gdim = GetDimOpt(1,true);
      *output << "Enter value for dimension" << endl;
      gval = GetReal();
      if (gdim == 1) {
	d1 = data2;
	d2 = data3;
      }
      else if (gdim == 2) {
	d1 = data1;
	d2 = data3;
      }
      else if ((gdim == 3) || (gdim == 4) || (gdim == 5)) {
	d1 = data1;
	d2 = data2;
      }
      plot.VectorPlot(d1,d2,gdim,gval, temperature);
      break;
    case '7':
      *output << "\nEnter dimension to plot (x,y,z,h,i)" << endl;
      gdim = GetDimOpt(1,true);
      *output << "Enter value for other dimensions" << endl;
      gval = GetReal();
      gval2 = GetReal();
      if (plot.GetPlotLegend()) {
	*output << "Enter label for legend" << endl;
	input->getline(line,10);
      }
      if (gdim == 1) {
	d1 = data2;
	d2 = data3;
      }
      else if (gdim == 2) {
	d1 = data1;
	d2 = data3;
      }
      else if ((gdim == 3) || (gdim == 4) || (gdim == 5)) {
	d1 = data1;
	d2 = data2;
      }
      plot.Hodograph(d1, d2, gdim, gval, gval2, line, temperature);
      break;
    case '8':
      PlotSingleVectors(true);
      break;
    case 'm':
      return;
    }
    break;
  default:
    cerr << "Unknown dimension of data " << ndim << endl;
  }
  
}
 
void Blasvis::PlotTimeSeries() {

  char cmax, graphopt;
  int series;
  char line[80];

  if (tdata1 == NULL) {
    cerr << "Must read time series before plotting it" << endl;
    return;
  }

  if (!plot.IsOpen()) {
    cerr << "Must start a plot before plotting data" << endl;
    return;
  }

  *output << "    Time series\n\n";
  *output << " 1. Time series\n";
  if (tdata2 != NULL) {
    *output << " 2. Hodograph\n";
    cmax = '2';
  }
  else {
    cmax = '1';
  }
  *output << "M. Return to main menu\n" << endl;

  do {
    input->getline(line,10);
    graphopt = tolower(line[0]);
  } while (((graphopt < '1') || (graphopt > cmax)) && (graphopt != 'm'));
  

  switch (graphopt) {
  case '1':
    if (tdata2 != NULL) {
      *output << "Enter series to plot" << endl;
      while ( ((series = GetInt()) < 1) || (series > 2)) {      
	cerr << "Series must be either 1 or 2" << endl;
      }
    }
    else series = 1;
    if (plot.GetPlotLegend()) {
      *output << "Enter label for legend" << endl;
      input->getline(line,10);
    }
    if (series == 1) {
      plot.LinePlot(tdata1,line);
    }
    else {
      plot.LinePlot(tdata2,line);
    }
    break;
  case '2':
    if (plot.GetPlotLegend()) {
      *output << "Enter label for legend" << endl;
      input->getline(line,10);
    }
    plot.Hodograph(tdata1,tdata2,line);
    break;
  case 'm':
    return;
  }

}
 
void Blasvis::FinishPlot() {

  if (plot.IsOpen()) {
    plot.FinishPlot();
  }
  else {
    cerr << "No plot currently open" << endl;
  }
}

void Blasvis::OutputData() {
  char fname[80], exname[80];

  if (data1 == NULL) {
    cerr << "Must select data set before saving it to file" << endl;
    return;
  }

  *output << "Enter output file name (blank to not save)" << endl;
  input->getline(fname,80);
  
  if (fname[0] != '\0') {
    StringExpand(fname,exname,80);
    data1->DisplayData(exname);
  }
  
  if (data2 != NULL) {
    *output << "Enter 2nd output file name" << endl;
    input->getline(fname,80);
    if (fname[0] != '\0') {
      StringExpand(fname,exname,80);
      data2->DisplayData(exname);
    }
    *output << "Enter 3rd output file name" << endl;
    input->getline(fname,80);
    if (fname[0] != '\0') {
      StringExpand(fname,exname,80);
      data3->DisplayData(exname);
    }
  }
}

void Blasvis::OutputTimeSeries() {
  char fname[80], exname[80];

  if (tdata1 == NULL) {
    cerr << "Must select time series before saving it to file" << endl;
    return;
  }

  *output << "Enter output file name (blank to not save)" << endl;
  input->getline(fname,80);
  
  if (fname[0] != '\0') {
    StringExpand(fname,exname,80);
    tdata1->DisplayData(exname);
  }
  
  if (tdata2 != NULL) {
    *output << "Enter 2nd output file name" << endl;
    input->getline(fname,80);
    if (fname[0] != '\0') {
      StringExpand(fname,exname,80);
      tdata2->DisplayData(exname);
    }
  }
}

//
// Output statistics about data set
//
void Blasvis::OutputDataStats() {

  if (data1 == NULL) {
    cerr << "Must select data set before printing statistics" << endl;
    return;
  }

  data1->DisplaySummary();

  if (data2 != NULL)
    data2->DisplaySummary();

  if (data3 != NULL)
    data3->DisplaySummary();
  
}

//
// Return true for yes, false for no
//
bool Blasvis::GetYesNo() {
  char yn,line[10];  

  while (true) {
    input->getline(line,10);
    
    yn = tolower(line[0]);
    if (yn == 'y')
      return true;
    if (yn == 'n') 
      return false;
    
    cerr << "Must enter y or n" << endl;
    if (isCmdFile) return false;
  }

  return false;
}

bldata Blasvis::GetReal(istream *in, char delim) {
  char line[80];
  bldata val;
  int i;
  bool err;
  char *end;

  err = true;

  if (in == NULL) {
    in = input;
  }

  do {
    in->getline(line,80,delim);

    // Check if this is a variable
    if (line[0] == '%') {
      if (isdigit(line[1])) {
	i = (int) line[1] - 48;
	val = varvals[i];
	//	cout << "Variable " << i << " is " << varvals[i] << endl;
	err = false;
      }
    }
    else {
      val = strtod(line,&end);
      if ((val != 0) || (end != line)) {
      err = false;
      }
    } 
    if (err == true) {
      cout << "Invalid number or variable" << endl;
      // This stops an infinite loop for batch files
      if (in != &cin) {
	return 0;
      }
    }
  } while (err == true);
  
  return val;
}

int Blasvis::GetInt(istream *in, char delim) {
  char line[80];
  int val;
  int i;
  bool err;
  char *end;

  err = true;

  if (in == NULL) {
    in = input;
  }

  do {
    in->getline(line,80,delim);

    // Check if this is a variable
    if (line[0] == '%') {
      if (isdigit(line[1])) {
	i = (int) line[1] - 48;
	val = (int) varvals[i];
	//	cout << "Variable " << i << " is " << varvals[i] << endl;
	err = false;
      }
    }
    else {
      val = strtol(line,&end,10);
      if ((val != 0) || (end != line)) {
      err = false;
      }
    } 
    if (err == true) {
      cout << "Invalid number or variable" << endl;
      // This stops an infinite loop for batch files
      if (in != &cin) {
	return 0;
      }
    }
  } while (err == true);
  
  return val;
}

//
// Get the dimension to plot through (dim=1 i.e. a surface) 
// or to plot (dim=2 i.e. a line)
//
int Blasvis::GetDimOpt(int dim, bool ext) {
  char dimopt,dimopt2,line[10];
  int ret;

  ret = 0;

  while (true) {
    input->getline(line,10);
    dimopt = tolower(line[0]);
    
    switch (dimopt) {
    case 'x':
      return 1;
    case 'y':
      return 2;
    case 'z':
      return 3;
    case 'h':
      if (ext) ret = 4;     
      break;
    case 'i':
      // Isentropic surface
      if (ext) ret = 5;
      break;
    case 's':
      // Streamfunction surface
      if (ext) ret = 6;
      break;
    case 'r':
      // Streamfunction surface - backwards
      if (ext) ret = 7;
      break;
    case 'm':
      // Model surface
      if (ext) ret = 8;
      break;
    case 'n':
      // Streamline on model surface
      if (ext) ret = 9;
      break;
    case 'o':
      // Reverse streamline on model surface
      if (ext) ret = 10;
      break;
    }
    if (ret == 0) {
      cerr << "Invalid option" << endl;
      if (isCmdFile) return 0;
    }
    else {
      if (dim == 1) {
	return ret;
      }
      dimopt2 = tolower(line[1]);
      switch(dimopt2) {
      case 'x':
	ret +=10;
	return ret;
      case 'y':
	ret +=20;
	return ret;
      }
      if ((ret == 6) || (ret == 7) || (ret == 9) || (ret == 10)) return ret;
      cerr << "Invalid option" << endl;
      if (isCmdFile) return 0;
    }
  }
}

//
// Extract the run number from the file name
//
int Blasvis::GetRunNumber(char *name) {
  
  return strtol(BaseName(name)+3,(char**)NULL,10);

}

//
// Extract the time from the file name
//
int Blasvis::GetTime(char *name) {
  
  char *ptr;

  ptr = BaseName(name);
  while ((*ptr != '_') && (*ptr != '\0')) {
    ptr++;
  }
  if (*ptr=='\0') return 0;
  ptr++;  
  return strtol(ptr,(char**)NULL,10);
}

//
// Extract the base file name from a file name
//
char *Blasvis::BaseName(char *name) {
  
  char *ptr, *start;

  start = name;
  ptr = name;
  while (*ptr != '\0') {
    while ((*ptr != '/') && (*ptr != '\0')) {
      ptr++;
    }
    if (*ptr == '/') {
      start = ptr+1;
      ptr++;
    }
  }
  return start;
}

//
// Expand the source string to the destination string (with max length)
// to insert run number, time etc in titles.
//
int Blasvis::StringExpand(char *src, char *dest, int n) {
  
  char *sptr, *dptr;
  char temp[10];
  int nchar = 0, tlen;

  sptr = src;
  dptr = dest;

  while ((*sptr != '\0') && (nchar < n)) {
    // Just plain text then copy it
    if (*sptr != '%') {
      *dptr = *sptr;
      sptr++;
      dptr++;
      nchar++;
    }
    // Expand any escape characters
    else {
      sptr++;
      switch (*sptr) {
      case 'r':
	sprintf(temp,"%d",runno);
	tlen = strlen(temp);
	tlen = (tlen < n-nchar)?tlen:n-nchar;
	strncpy(dptr,temp,tlen);
	nchar += tlen;
	dptr += tlen;
	break;
      case 't':
	sprintf(temp,"%d",time);
	tlen = strlen(temp);
	tlen = (tlen < n-nchar)?tlen:n-nchar;
	strncpy(dptr,temp,tlen);
	nchar += tlen;
	dptr += tlen;
	break;
      case '%':
	*dptr = '%';
	dptr++;
	nchar++;
	break;
      default:
	cerr << "Invalid escape sequence ignored " << endl;
      }
      sptr++;
    }
  }

  // Null terminate string properly!
  if (nchar < n) {
    *dptr = '\0';
    nchar++;
  }

  return nchar;
}

//
// Check if flow separation occurs and locate it
//
void Blasvis::CheckSeparation() {
  bldata d;
  DataSet *taux;

  taux = datafile->GetData("TAUX");
  
  d = plot.Separation(taux);

  //  if (d > 0) {
  //    cout << d <<endl;
  //  }

}


//
// Get drag on hills. 
//
void Blasvis::GetDrag() {
  datafile->GetDrag();
}

//
// Work out derivative quantities from data
//
void Blasvis::DerivData() {

  char menuopt;
  char line[10];
  DataSet *data;
  DataVect *oldvect;
  int vdim;

  if (data1 == NULL) {
    cerr << "Must select data set first" << endl;
    return;
  }

  *output << "    Calculate derivatives\n\n";

  if (vect == NULL) {
    *output << " 1. Grad\n";
    *output << " M. Return to main menu\n";

    do {
      input->getline(line,10);      
      menuopt = tolower(line[0]);

      switch (menuopt) {
      case '1':
	vect = data1->Grad();
	data1 = vect->GetX();
	data2 = vect->GetY();
	data3 = vect->GetZ();
	break;
      case 'm':
	break;
      default:
	menuopt = '\0';
	cerr << "Invalid option - please try again" << endl;
	if (isCmdFile) return;
	break;
      }
    } while (menuopt == '\0');
    
  }
  else {
    *output << " 1. Div - NOT WORKING!\n";
    *output << " 2. Curl - NOT WORKING!\n";
    *output << " 3. Grad\n";
    *output << " M. Return to main menu\n";
    do {
      input->getline(line,10);      
      menuopt = tolower(line[0]);

      switch (menuopt) {
      case '1':
	//	data1 = vect->Div();
	data2 = NULL;
	data3 = NULL;
	if (!vect->IsCached() ) {
	  delete vect;
	}
	vect = NULL;
	break;
      case '2':
	oldvect = vect;
	vect = oldvect->Curl();
	data1 = vect->GetX();
	data2 = vect->GetY();
	data3 = vect->GetY();
	if (! oldvect->IsCached() ) {
	  delete oldvect;
	}
	break;
      case '3':
	*output << "Enter vector component to take deriv (x,y,z)" << endl;
	vdim = GetDimOpt(1,false);
	if (vdim == 1) {
	  data = data1;
	}
	else if (vdim == 2) {
	  data = data2;
	}
	else{
	  data = data3;
	}
	if (!vect->IsCached() ) {
	  delete vect;
	}
	vect = data->Grad();
	data1 = vect->GetX();
	data2 = vect->GetY();
	data3 = vect->GetZ();
	break;
      case 'm':
	break;
      default:
	menuopt = '\0';
	cerr << "Invalid option - please try again" << endl;
	if (isCmdFile) return;
	break;
      }
    } while (menuopt == '\0');
    
  }

  
}

//
// Work out normalised data
//
void Blasvis::NormData() {

  if (data1 == NULL) {
    cerr << "Must select data set first" << endl;
    return;
  }

  data1 = data1->Normalise();
}

//
// Set variable value
//
void Blasvis::SetVariable() {
  int ival;
  char line[80];
  
  *output << "Enter variable to set (0-9)" << endl;
  do {
    input->getline(line,80);
    ival = atol(line);
    if ( (ival < 0) || (ival > 9) ) {
      cerr << "Must select data set first" << endl;
    }
  } while ((ival < 0) || (ival > 9) );
  
  *output << "Enter value for variable" << endl;
  varvals[ival] = GetReal();
  //  cout << "Variable " << ival << " is " << varvals[ival] << endl;
  
}

//
// Get the values at a series of specified points
//
void Blasvis::GetValues() {
  char line[80], ifname[80], ofname[80];
  bool infile, outfile, model;
  ifstream fin;
  istream *in;
  ofstream fout;
  ostream *out;
  bldata x, y, z, val;  

  if (data1 == NULL) {
    cerr << "Must select data set first" << endl;
    return;
  }

  *output << "Enter file name to read points from (blank for keyboard input)" << endl;
  input->getline(line,80);
  if (line[0] != '\0') {
    StringExpand(line,ifname,80);
    infile = true;
    fin.open(ifname, ios::in);
    if (fin.good()) {
      in = &fin;
    }
    else {
      cerr << "Failure opening file " << ifname << endl;
      return;
    }
  }
  else {
    infile = false;
    in = input;
  }

  *output << "Enter file name for output (blank for screen)" << endl;
  input->getline(line,80);
  if (line[0] != '\0') {
    StringExpand(line,ofname,80);
    outfile = true;
    fout.open(ofname, ios::out);
    if (fout.good()) {
      out = &fout;
    }
    else {
      cerr << "Failure opening file " << ofname << endl;
      return;
    }    
  }
  else {
    outfile = false;
    out = output;
  }

  *output << "Use model coordinates (y/n)" << endl;
  model = GetYesNo();
  
  if (!infile) {
    *output << "Enter x, y and z values, q to finish" << endl;
  }
  while (in->get() != 'q') {
    in->unget();
    x = GetReal(in,' ');
    y = GetReal(in,' ');
    z = GetReal(in,'\n');
    val = data1->GetVal2(x,y,z,model);
    *out << x << " " << y << " " << z << " " << val;
    if (data2 != NULL) {
      val = data2->GetVal2(x,y,z,model);
      *out << " " << val;
      val = data3->GetVal2(x,y,z,model);
      *out << " " << val;
    }
    *out << endl;
  }
  
  if (infile) fin.close();
  if (outfile) fout.close();

}

//
// Get the values at a series of specified points
//
void Blasvis::PlotSingleVectors(bool ismodel) {
  char line[80], ifname[80];
  bool infile, model, vectin;
  ifstream fin;
  istream *in;
  int gdim;
  bldata x, y, z, val1, val2, xx, yy;  
  DataSet *d1, *d2;

  if (ismodel) {
    *output << "Input vectors as well as coordinates (y/n)" << endl;
    vectin = GetYesNo();
  }
  else {
    vectin = true;
  }

  if (!vectin) {
    if (data1 == NULL) {
      cerr << "Must select data set first" << endl;
      return;
    }
    
    if (data2 == NULL) {
      cerr << "Must select vector data set first" << endl;
      return;
    }
  }

  *output << "Enter file name to read points from (blank for keyboard input)" << endl;
  input->getline(line,80);
  if (line[0] != '\0') {
    StringExpand(line,ifname,80);
    infile = true;
    fin.open(ifname, ios::in);
    if (fin.good()) {
      in = &fin;
    }
    else {
      cerr << "Failure opening file " << ifname << endl;
      return;
    }
  }
  else {
    infile = false;
    in = input;
  }

  *output << "Use model coordinates (y/n)" << endl;
  model = GetYesNo();


  *output << "\nEnter dimension to plot (x,y,z,h,i)" << endl;
  gdim = GetDimOpt(1,true);
  if (gdim == 1) {
    d1 = data2;
    d2 = data3;
  }
  else if (gdim == 2) {
    d1 = data1;
    d2 = data3;
  }
  else if ((gdim == 3) || (gdim == 4) || (gdim == 5)) {
    d1 = data1;
    d2 = data2;
  }

  if (!infile) {
    if (vectin) 
      *output << "Enter x, y and z, v1 and v2 values, q to finish" << endl;
    else
      *output << "Enter x, y and z values, q to finish" << endl;
  }
  while (in->get() != 'q') {
    in->unget();
    x = GetReal(in,' ');
    y = GetReal(in,' ');
    if (vectin) {
      z = GetReal(in,' ');
      val1 = GetReal(in,' ');
      val2 = GetReal(in,'\n');
    }
    else {
      z = GetReal(in,'\n');
      val1 = d1->GetVal2(x,y,z,model);
      val2 = d2->GetVal2(x,y,z,model);
    }
    if (gdim == 1) {
      xx = y;
      yy = z;
    }
    else if (gdim == 2) {
      xx = x;
      yy = z;
    }
    else if ((gdim == 3) || (gdim == 4) || (gdim == 5)) {
      xx = x;
      yy = y;
    }
    plot.VectorPlot(val1, val2, xx, yy);
  }
  
  if (infile) fin.close();
  
}
 
//
// Plot other misc items on figure
// 
void Blasvis::PlotMisc() {
  char opt;
  char line[80];
  
  if (!plot.IsOpen()) {
    cerr << "Must start a plot before plotting" << endl;
    return;
  }
  
  *output << "    Plot misc items\n";
  *output << " 1. Point vectors\n";
  *output << " 2. Text label\n";
  *output << " M. Return to main menu\n" << endl;

  do {
    input->getline(line,10);      
    opt = tolower(line[0]);
  } while (((opt < '1') || (opt > '2')) && (opt != 'm'));

  
  switch(opt) {
  case '1':
    PlotSingleVectors(false);
    break;
  case '2':
    PlotLabels();
    break;
  case 'm':
    return;
  default:
    cerr << "Unknown option " << opt << endl;
    if (isCmdFile) return;
  }    
}

//
// Plot labels
//
void Blasvis::PlotLabels() {
  char line[80], ifname[80];
  bool infile;
  ifstream fin;
  istream *in;
  PLFLT x, y;

  *output << "Enter file name to read points from (blank for keyboard input)" << endl;
  input->getline(line,80);
  if (line[0] != '\0') {
    StringExpand(line,ifname,80);
    infile = true;
    fin.open(ifname, ios::in);
    if (fin.good()) {
      in = &fin;
    }
    else {
      cerr << "Failure opening file " << ifname << endl;
      return;
    }
  }
  else {
    infile = false;
    in = input;
  }
  while (true) {
    if (!infile) {
      *output << "Enter label (blank to finish)" << endl;
    }    
    in->getline(line,80);
    if (line[0] == '\0') break;
    if (!infile) {
      *output << "Enter x and y coordinates" << endl;
    }
    x = GetReal(in,' ');
    y = GetReal(in,'\n');
    plot.PlotLabel(line,x,y);
  }

}

int main (int argc, char **argv) {

  Blasvis blasvis;
  
  return blasvis.Run(argc,argv);
}

