//
// xblasvis - C++ version of graphics routines - X11 GUI 
//
// GUI : Main GUI window
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>
//

#include <qapp.h>
#include <qpopupmenu.h>
#include <qfiledialog.h>
#include <qmessagebox.h>

#include <cmath>

#include "gui.h"
#include "exportdlg.h"

#include "blasfile.h"

using namespace std;

int main(int argc, char *argv[]) {

  QApplication app(argc, argv);
  
  app.setFont(QFont("default", 12, QFont::Bold, FALSE));

  GUI gui(0,"BLASVIS++");
  app.setMainWidget(&gui);
  gui.show();
  return app.exec();
}

GUI::GUI(QWidget *parent, const char *name) : QMainWindow(parent,name) {

  // Set the minimum size of the window
  setMinimumSize(300,200);
  resize(300,200);


  // Create file popup menu
  fileMenu = new QPopupMenu();
  CHECK_PTR(fileMenu);
  fileMenu->insertItem("&Open data file", this , SLOT(openData()), CTRL+Key_O);
  swapEndianId = fileMenu->insertItem("Swap &byte order", this, SLOT(swapByteOrder()));
  swapEndian = FALSE;
  fileMenu->setCheckable(TRUE);
  fileMenu->setItemChecked(swapEndianId,swapEndian);  
  fileMenu->insertItem("Open &cmd file", this, SLOT(openCmd()), CTRL+Key_C);
  fileMenu->insertItem("&Export data", this, SLOT(saveData()), CTRL+Key_E);
  fileMenu->insertSeparator();
  fileMenu->insertItem("Close data file", this, SLOT(closeData()) );
  fileMenu->insertItem("E&xit", qApp, SLOT(quit()), CTRL+Key_Q);

  devMenu = new QPopupMenu();
  CHECK_PTR(devMenu);
  // N.B. These must be in the same order as DataPlot::devtype 
  // in order to work correctly
  devMenu->insertItem("&X window",-1);
  devMenu->insertItem("&Mono ps",-1);
  devMenu->insertItem("&Colour ps",-1);
  devMenu->insertItem("X&fig",-1);
  devMenu->insertItem("P&BM",-1);
  devMenu->insertItem("&PNG",-1);

  dev = DataPlot::xwin;
  devMenu->setCheckable(TRUE);
  connect(devMenu,SIGNAL(activated(int)),this,SLOT(setDevice(int)));
  // This doesn't work here - why???
  devMenu->setItemChecked(dev-1,TRUE);
  plot.SetDevice(dev);


  QPopupMenu *plotMenu = new QPopupMenu();
  CHECK_PTR(plotMenu);
  //  plotMenu->insertItem("Select &data", this, SLOT(selectData()), CTRL+Key_D);
  plotMenu->insertItem("Set &device", devMenu);
  plotMenu->insertItem("Plot &options", this, SLOT(setPlotOptions()), CTRL+Key_T);
  plotMenu->insertSeparator();
  //  plotMenu->insertItem("&Start plot", this, SLOT(startPlot()), CTRL+Key_T);
  //  plotMenu->insertItem("&Finish plot", this, SLOT(finishPlot()), CTRL+Key_F);
  //  plotMenu->insertItem("&Next page", this, SLOT(nextPage()), CTRL+Key_N);
  plotMenu->insertItem("Plot &1-D data", this, SLOT(plot1dData()), CTRL+Key_1);
  plotMenu->insertItem("Plot &2-D data", this, SLOT(plotData()), CTRL+Key_2);

  QPopupMenu *helpMenu = new QPopupMenu();
  CHECK_PTR(helpMenu);
  helpMenu->insertItem("&About", this, SLOT(about()));

  // Create menubar
  menu = new QMenuBar(this);
  CHECK_PTR(menu);
  menu->insertItem("&File",fileMenu);
  menu->insertItem("&Plot",plotMenu);
  menu->insertSeparator();
  menu->insertItem("&Help",helpMenu);
  menu->setSeparator(QMenuBar::InWindowsStyle);

  // Initially set pointers to NULL
  temperature = NULL;
  data1 = NULL;
  data2 = NULL;
  data3 = NULL;
  vect = NULL;
  model = blasius;

  plotDlg = NULL;
  plot1dDlg = NULL;
  optDlg = NULL;
  expDlg = NULL;
  datafile = NULL;
  
}

GUI::~GUI() {
  
}

// ********************
// * Open a data file *
// ********************
void GUI::openData() {
  
  QString fileName = QFileDialog::getOpenFileName();

  if ( fileName.isNull() ) {
    return;
  }

  // Close file if needed
  if (datafile != NULL) {
    closeData();
  }

  datafile = new BlasFile();
  CHECK_PTR(datafile);

  datafile->SwapEndian(swapEndian);  
  datafile->OpenFile(fileName);
  // runno = GetRunNumber(fname);
  runno = 1;
  time = (int) floor(datafile->GetParm("TIME"));
  plot.SetRunParms(runno,time);
  temperature = datafile->GetData("T");
  

}

//
// Close a data file
//
void GUI::closeData() {
  if (datafile) {
    datafile->CloseFile();
    delete datafile;
    datafile = NULL;

    // Check if plot dialog exists and delete it
    if (plotDlg) {
      delete plotDlg;
      plotDlg = NULL;
    }
    if (plot1dDlg) {
      delete plot1dDlg;
      plot1dDlg = NULL;
    }
    if (optDlg) {
      delete optDlg;
      optDlg = NULL;
    }
    if (expDlg) {
      delete expDlg;
      expDlg = NULL;
    }
  }
}

// ***********************
// * Open a command file *
// ***********************

void GUI::openCmd() {
  QMessageBox::information(this, "xblasvis++", "Command files currently not supported");
}

// *****************************
// * Save the current data set *
// *****************************
void GUI::saveData() {
  if (datafile == NULL) {
    QMessageBox::warning(this,"Warning!","No data file open");
    return;
  }

  if (expDlg == NULL) {
    expDlg = new ExportDlg(NULL,datafile, optDlg, "Export data");
  }
  
  expDlg->exec();
}

// ***************************
// * Select current data set *
// ***************************
void GUI::selectData() {
  QMessageBox::information(this, "xblasvis++", "Selecting data currently not supported");
}

// ********************
// * Set plot options *
// ********************
void GUI::setPlotOptions() {
  if (optDlg == NULL) {
    optDlg = new OptDlg(NULL,&plot, "Set options");
    CHECK_PTR(optDlg);
  }
  
  optDlg->exec();

}

// *************************
// * Plot current data set *
// *************************
void GUI::plotData() {

  if (datafile == NULL) {
    QMessageBox::warning(this,"Warning!","No data file open");
    return;
  }

  if (!optDlg) {
    optDlg = new OptDlg(NULL,&plot, "Set options");
    CHECK_PTR(optDlg);
  }

  if (!plotDlg) {
    plotDlg = new PlotDlg(this,datafile,&plot, optDlg,"Plot data");
    CHECK_PTR(plotDlg);
  }

  // Show dialog 
  if (plotDlg->exec()) {
    plotDlg->finishPlot();
  }

}

// *************************
// * Plot current data set *
// *************************
void GUI::plot1dData() {

  if (datafile == NULL) {
    QMessageBox::warning(this,"Warning!","No data file open");
    return;
  }

  if (!optDlg) {
    optDlg = new OptDlg(NULL,&plot, "Set options");
    CHECK_PTR(optDlg);
  }

  if (!plot1dDlg) {
    plot1dDlg = new Plot1dDlg(this,datafile,&plot, optDlg,"Plot data");
    CHECK_PTR(plot1dDlg);
  }

  // Show dialog 
  if (plot1dDlg->exec()) {
    plot1dDlg->finishPlot();
  }

}

// ******************************************
// * Display statistics on current data set *
// ******************************************
void GUI::displayStats() {
}

// ***********************
// * Help | about dialog *
// ***********************
void GUI::about() {
  QMessageBox::about(this,"About qblasvis","qblasvis\nGraphics for BLASIUS\n(C) 2001-2004 Andrew Ross <aross@env.leeds.ac.uk>");
}

// *************************************
// * Called when the device is changed *
// *************************************
void GUI::setDevice(int index) {

  devMenu->setItemChecked((int) dev -1,FALSE);
  devMenu->setItemChecked(index,TRUE);
  dev = (DataPlot::devtype) (index+1);

  plot.SetDevice(dev);
}

// ********************
// * Start a new plot *
// ********************
void GUI::startPlot() {

  const char *fileName;
  const char filter[7][30] = {"","","Postscript (*.ps,*.eps)","Postscript (*.ps,*.eps)","XFig (*.fig)","PBM (*.pbm)", "PNG (*.png)"};

  if (plot.IsOpen()) {
    QMessageBox::warning(this,"Warning!","Plot already open");
    return;
  }

  if (plot.NeedDevFile()) {
    fileName = QFileDialog::getSaveFileName(NULL,filter[dev]);
    plot.SetDevFile(fileName);
  }

  plot.InitPlot();

}

// ***************
// * Finish plot *
// ***************
void GUI::finishPlot() {

  if (!plot.IsOpen()) {
    QMessageBox::warning(this,"Warning!","Plot not open");
    return;
  }

  plot.FinishPlot();
  
}

// ********************
// * Start a new page *
// ********************
void GUI::nextPage() {

  if (!plot.IsOpen()) {
    QMessageBox::warning(this,"Warning!","Plot not open");
    return;
  }

  plot.NextPlot();

}

//
// Swap the byte order when reading in data
//
void GUI::swapByteOrder() {

  if (!datafile) {
    swapEndian = !swapEndian;
    fileMenu->setItemChecked(swapEndianId,swapEndian);
  }
  else {
    QMessageBox::warning(this,"Warning!","Cannot change byte order while file is open");
  }
}

#include "gui.moc"
