// 
// makesetup - create a BLASIUS setup.dat file from a model
// output file
//
// makesteup : Main class 
//
// (C) 2004 Andrew Ross <aross@env.leeds.ac.uk>


#include "blasfile.h"
#include "makesetup.h"
#include "dataset.h"
#include <iostream>
#include <cstdlib>
#include <cctype>
#include <cstring>
#include <cmath>
#include <limits.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

using namespace std;

// 
// Constructor
//
MakeSetup::MakeSetup() {
}

// 
// Destructor
//
MakeSetup::~MakeSetup() {
}

//
// Run the code
//
int MakeSetup::run(int argc,char **argv) {
  BlasFile *datafile;
  DataSet *dataset;
  bldata *data;
  char optchar;
  char dfile[PATH_MAX];
  int kkp, lo;
  bldata theta0, g;

  dfile[0] = '\0';

#ifdef HAVE_GETOPT_H  
  int opt_index = 0;
  static struct option long_options[] = {
	  {"version",0,NULL,'V'},
	  {"help",0,NULL,'h'}
  };
#endif

  #ifdef HAVE_GETOPT_H
  while ( (optchar = getopt_long(argc,argv, "hf:V",
				  long_options, &opt_index)) != EOF) {
#else
  while ( (optchar = getopt(argc,argv, "hf:V")) != EOF) {
#endif

    switch (optchar) {
    case 'f':
      strncpy(dfile,optarg,PATH_MAX);
      dfile[PATH_MAX-1] = '\0';
      break;
    case 'h':
    case '?':
      // Print help and exit   
      cout << "Usage: makesetup [OPTIONS]\n";
      cout << "Extract data from 1-d BLASIUS data file and use it to create a\nsetup.dat file for a 2-d / 3-d run.\n\n";
      cout << "  -f FILE        Select the data file FILE\n";
      cout << "  -h, --help     Print this help and exit\n";
      cout << "  -V, --version  Print version and exit\n\n";
      cout << "Please report bugs to Andrew Ross <aross@env.leeds.ac.uk>\n" << endl;
      return 0;
    case 'V':
      // Print out version and copyright and exit.
      cout << "makesetup " << version << "\n\n";
      cout << "Copyright (C) 2004 Andrew Ross\n";
      cout << "This is free software; see the source for copying conditions.  There is NO\n";
      cout << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << endl;
      return 0;
    }
  }

  if (strlen(dfile) == 0) {
    cerr << "Must specify a data file name\nSee makesetup -h for further details" << endl;
    return 1;
  }

  // Open data file
  datafile = new BlasFile();
  datafile->NoFortranRecords(true);
  datafile->OpenFile(dfile);

  // Set output format
  cout.precision(15);
  //cout.flags(ios::fixed | ios::scientific);
  //cout.width(20);

  kkp = (int) datafile->GetParm("KKP");
  lo = (int) datafile->GetParm("LORAYL");

  if (lo) {
    dataset = datafile->GetData("RAY");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
    dataset = datafile->GetData("RAYN");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
  }
  else {
    for (int k = 0; k<kkp; k++) {
      cout << 0.0 << " ";
    }
    cout << endl;
    for (int k = 0; k<kkp; k++) {
      cout << 0.0 << " ";
    }
    cout << endl;
  }
  // U
  dataset = datafile->GetData("U");
  data = dataset->Get1D(3);
  for (int k = 0; k<kkp; k++) {
    cout << data[k] << " ";
  }
  cout << endl;
  // V
  dataset = datafile->GetData("V");
  data = dataset->Get1D(3);
  for (int k = 0; k<kkp; k++) {
    cout << data[k] << " ";
  }
  cout << endl;

   lo = (int) datafile->GetParm("LOTEMP");
  if (lo > 0) {
    theta0 = datafile->GetParm("THETA0");
    g = datafile->GetParm("G");
    dataset = datafile->GetData("T");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << (data[k]-theta0)*g/theta0 << " ";
    }
    cout << endl;
  }
  lo = (int) datafile->GetParm("LOVAPR");
  if (lo > 0) {
    dataset = datafile->GetData("Q");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
  }
  lo = (int) datafile->GetParm("LOENGY");
  if (lo > 0 || lo <= -2) {
    dataset = datafile->GetData("E");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
  }
  if (lo <= -2) {
    dataset = datafile->GetData("B11");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
    dataset = datafile->GetData("B13");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
    dataset = datafile->GetData("B33");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
    dataset = datafile->GetData("B12");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
    dataset = datafile->GetData("B23");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
    dataset = datafile->GetData("EPSLN");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
  }
  lo = (int) datafile->GetParm("LOTRCR");
  if (lo > 0) {
    dataset = datafile->GetData("TRCR");
    data = dataset->Get1D(3);
    for (int k = 0; k<kkp; k++) {
      cout << data[k] << " ";
    }
    cout << endl;
  }
  lo = (int) datafile->GetParm("LOUBOT");
  if (lo == 0) {
    data[0] = datafile->GetParm("TAUXAV");
    data[1] = datafile->GetParm("TAUYAV");
    cout << sqrt(data[0]*data[0]+data[1]*data[1]) << endl;
  }

  return 0;

}

int main (int argc, char **argv) {
  MakeSetup makesetup;

  return makesetup.run(argc,argv);
}
