// 
// blasvis++ - C++ version of graphics routines
//
// Blasfile : Class to read the information in from the
//            Blasius file. Also from LEM file - both use
//            DaVie data format.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#include "blasfile.h"
#include <iostream>
#include <fstream>
#include <cctype>
#include <cstring>
#include <cmath>
#include <cstdlib>

using namespace std;

// Default constructor
BlasFile::BlasFile() : ReadFile()  {
}

// Constructor with a file name
BlasFile::BlasFile(char* fname)  : ReadFile(fname) {
  firstItem = NULL;
  lastItem = NULL;
  numitems = 0;
}

// destructor
BlasFile::~BlasFile() {
  CloseFile();
  delete[] nullgrid;
}

// 
// Do any initialisation associated with opening the file
//
void BlasFile::InitFile() {
  ReadHeader();
  ReadParms();
  LoadGrids();
}


//
// Close the current file
//
void BlasFile::CloseFile() {
  HeaderItem *hp1,*hp2;

  if (!file.is_open()) 
    return;

  file.close();

  // Delete linked list of header items
  hp1 = firstItem;

  while (hp1 != NULL) {
    hp2 = hp1->next;
    delete hp1;
    hp1 = hp2;
  }

  firstItem = NULL;
  lastItem = NULL;
  numitems = 0;

  // Delete any derived data sets
  if (pv) {
    delete pv;
    pv = NULL;
  }
  if (pvflux) {
    pvflux->Delete();
    delete pvflux;
    pvflux = NULL;
  }  
  if (gradt) {
    gradt->Delete();
    delete gradt;
    gradt = NULL;
  }  
  if (pvfluxalt) {
    pvfluxalt->Delete();
    delete pvfluxalt;
    pvflux = NULL;
  }
  if (bernoulli) {
    delete bernoulli;
    bernoulli = NULL;
  }
  if (windspeed) {
    delete windspeed;
    windspeed = NULL;
  }
  if (speedup) {
    delete speedup; 
    speedup = NULL;
  }
  if (deltaT) {
    delete deltaT;
    deltaT = NULL;
  }
  if (pcorr) {
    delete pcorr;
    pcorr = NULL;
  }
  if (timestep) {
    delete timestep;
    timestep = NULL;
  }
  if (bldepth) {
    delete bldepth;
    bldepth = NULL;
  }
  if (PVsrc) {
    delete PVsrc;
    PVsrc = NULL;
  }
  if (height) {
    delete height;
    height = NULL;
  }
  if (uw_res) {
    delete uw_res;
    uw_res = NULL;
  }
  if (uu_res) {
    delete uu_res;
    uu_res = NULL;
  }
  if (ww_res) {
    delete ww_res;
    ww_res = NULL;
  }

  delete[] xgrid;
  delete[] xngrid;
  delete[] ygrid;
  delete[] yngrid;
  delete[] zgrid;
  delete[] zngrid;
  delete[] orog;

  xgrid = NULL;
  xngrid = NULL;
  ygrid = NULL;
  yngrid = NULL;
  zgrid = NULL;
  zngrid = NULL;
  orog = NULL;
  
}

//
// Read in the header from file
//
void BlasFile::ReadHeader() {
  int nflds,ndim,dim[3],ccount,offset;
  char inchar;
  char clong[sizeof(long)];
  HeaderItem *ptr;
  long hdrlen;
  char dstring[12];
  int icount;
  char *nameptr;

  file.seekg(0,ios::beg);

  nflds = 0;
  ndim = 0;
  dim[0] = 1;
  dim[1] = 1;
  dim[2] = 1;

  // Counters for header records
  hdrrecs = 1;
  crec = 0;

  // cout << "Reading header..." << endl;

  // Read in record length
  if (!norec) {
    file.read(clong,sizeof(long));
  }
  if (!file) {
    cerr << "Error reading header 1" << endl;
  }
  // cout << clong << endl;
  ccount = 0;

  // Read in from the beginning of the file
  inchar = ReadChar();
  if (!file) {
    cerr << "Error reading header 2" << endl;
  }

  // Keep going until we reach the end of the header
  while (inchar != ':') {

    // Comment - just ignore this
    if (inchar == '!') {
      inchar = ReadChar();
      while (inchar != '!') inchar = ReadChar();
      inchar = ReadChar();
      if (!file) {
	cerr << "Error reading header 3" << endl;
      }
    }
    else if (inchar == '&') {
      // Start of new data block - reset the variables
      ndim = 0;
      dim[0] = 1;
      dim[1] = 1;
      dim[2] = 1;
      ccount = 0;
      inchar = ReadChar();
      if (!file) {
	cerr << "Error reading header 4" << endl;
      }
    }
    else if (isdigit(inchar)) {
      if (ndim != 0) cerr << "Error in header format" << endl;;
      dstring[0] = inchar;
      do  {
	if (ndim > 0) {
	  dstring[0] = ReadChar();
	}
	icount = 0;
	while (isdigit(dstring[icount])) {
	  icount++;	
	  if (icount == 12) {
	    dstring[11] = '\0';
	    cerr << "Digit too long in header " << dstring << endl;
	    return;
	  }
	  dstring[icount] = ReadChar();
	}
	inchar = dstring[icount];
	dstring[icount] = '\0';
	dim[ndim] = (int) atol(dstring);
	//cout << "Dim = " << dim[ndim] << endl;
	ndim++;
      } while (inchar == ',');
    }
    else if (isalpha(inchar) || (inchar == ' ') || (inchar == '_') ) {
      if (lastItem == NULL) {
	lastItem = new HeaderItem;
	firstItem = lastItem;
	lastItem->offset = 0;
      }
      else {
	lastItem->next = new HeaderItem;
	lastItem->next->offset = lastItem->offset + 
	  lastItem->dim[0]*lastItem->dim[1]*lastItem->dim[2];
	lastItem = lastItem->next;
      }
      lastItem->ndim = ndim;
      lastItem->dim[0] = dim[0];
      lastItem->dim[1] = dim[1];
      lastItem->dim[2] = dim[2];
      lastItem->next = NULL;
      lastItem->data = NULL;
      nflds++;
      lastItem->id = nflds;
      ccount = 0;
      do {
	lastItem->name[ccount] = inchar;
	ccount++;
	inchar = ReadChar();
	if (!file) {
	  cerr << "Error reading header 5" << endl;
	}
      } while ((isalnum(inchar)|| (inchar == ' ') || (inchar == '_'))&&(ccount<hnlen-1));
      lastItem->name[ccount] = '\0';
      // Check if LES stats and correct dimensions accordingly
      if ((ccount >= 2) && (strncasecmp(&lastItem->name[ccount-2],"AV",2)==0)) {
	if (lastItem->dim[0] > 1) {
	  if (lastItem->dim[1] == 1) {
	    lastItem->dim[2] = lastItem->dim[0];
	    lastItem->dim[0] = 1;
	  }
	  else {
	    lastItem->dim[2] = lastItem->dim[1];
	    lastItem->dim[1] = 1;
	  }
	  lastItem->ndim = 3;
	}
      }
      if ( ((strncasecmp(lastItem->name,"UI",2)==0) || 
         (strncasecmp(lastItem->name,"VI",2)==0) ||
         (strncasecmp(lastItem->name,"TI",2)==0) ||
         (strncasecmp(lastItem->name,"EI",2)==0) ||
         (strncasecmp(lastItem->name,"QI",2)==0)) &&
	   (strlen(lastItem->name) == 2)) {
	lastItem->dim[2] = lastItem->dim[0];
	lastItem->dim[0] = 1;
	lastItem->ndim = 3;
      }
      if ( (strncasecmp(lastItem->name,"TRCRI",5)==0) ||
	   (strncasecmp(lastItem->name,"B11I",4)==0) ||
	   (strncasecmp(lastItem->name,"B13I",4)==0) ||
	   (strncasecmp(lastItem->name,"B33I",4)==0) ||
	   (strncasecmp(lastItem->name,"B12I",4)==0) ||
	   (strncasecmp(lastItem->name,"B23I",4)==0) ||
	   (strncasecmp(lastItem->name,"ESNI",4)==0) ) {
	lastItem->dim[2] = lastItem->dim[0];
	lastItem->dim[0] = 1;
	lastItem->ndim = 3;
      }
      if ( ((strncasecmp(lastItem->name,"RAY",3)==0) &&
	   (strlen(lastItem->name) == 3)) || 
	   ((strncasecmp(lastItem->name,"RAYN",4)==0) &&
	    (strlen(lastItem->name) == 4)) ) {
	lastItem->dim[2] = lastItem->dim[0];
	lastItem->dim[0] = 1;
	lastItem->ndim = 3;
      }
      //cout << lastItem->name << endl;
      if (inchar == ',') inchar = ReadChar();
    }
    else {
      cerr << "Error reading header" << endl;
      inchar = ':';     
    }

  }

  // If no records then skip whitespace to end of header.
  // Otherwise header ends at the end of the record.
  if (norec) {
    while ( ReadChar() == ' ' ) {
    }
    // Back up one character
    UnreadChar();
    hdrlen = crec;
  }
  else {
    hdrlen = hdrrecs*irecl;
  }

  // Now update the offset values to actual file positions
  // and convert all header names to upper case

  ptr = firstItem;
  ccount = 0;

  while (ptr != NULL) {
    offset = ptr->offset*sizeof(float) + hdrlen;
    if (!norec) {
      offset += (offset/irecl)*2*reclen+reclen;
    }
    ptr->offset = offset;
    nameptr = (ptr->name);
    while (*nameptr != '\0') {
      *nameptr = toupper(*nameptr);
      nameptr++;
    }
    //    cout << ptr->name << endl;
    ptr = ptr->next;
    ccount++;
  }
  // cout << "Number of items : " << ccount << endl;

  
  //  cout << "Done..." << endl;
  
}


//
// Read in the 0-D parameters and store their values
//
void BlasFile::ReadParms() {
  HeaderItem *ptr;
  int count = 0;

  ptr = firstItem;

  while (ptr != NULL) {
    // Check if this is a 0-D parameter and if so read it in.
    if ((ptr->ndim == 1)  && (ptr->dim[0] == 1)) {
      if (!file) {
	cerr << "Error before seek" << endl;
	return;
      }
      file.seekg(ptr->offset,ios::beg);
      if (!file) {
	cerr << "Error seeking" << endl;
	return;
      }
      ptr->value = (bldata) BinRead<float>();
      ptr->ndim = 0;
      //cout << ptr->name << " = " << ptr->value << endl;
      count++;
    }
    ptr = ptr->next;
  }

  //cout  << "Number of parameters : " << count << endl;

}

// 
// Display the 0-D parameters
//
void BlasFile::DisplayParms() {
  HeaderItem *ptr;
  int count = 0;

  ptr = firstItem;

  while (ptr != NULL) {
    // Check if this is a 0-D parameter and if so read it in.
    if (ptr->ndim == 0) {
      cout << ptr->name << " = " << ptr->value << endl;
      count++;
      if (count == 23) {
        cout << "Press return to continue" << endl;
        while (cin.get() != '\n');
        count = 0;
      }
    }
    ptr = ptr->next;
  }
}

//
// Get the specified dataset. If it is not cached then load it from file.
//
DataSet * BlasFile::GetData(const char name[hnlen]) {

  HeaderItem * ptr;
  bldata *** dptr;
  int nx,ny,nz;
  long offset;

  ptr = firstItem;

  while ((ptr != NULL) && (strncasecmp(name,ptr->name,hnlen)!=0)) {
    ptr = ptr->next;
  }

  if (ptr == NULL) {
    cerr << "Data " << name << " not found." << endl;
    return NULL;
  }

  if (ptr->data == NULL) {

    if (!file) {
      cout << "Failed before we even started!" << endl;
      return NULL;
    }
    nx = ptr->dim[0];
    ny = ptr->dim[1];
    nz = ptr->dim[2];    
    // cout << name << " " << nx << " " << ny << " " << nz << endl;
    dptr = Alloc3DArray<bldata>(nx,ny,nz);
    offset = ptr->offset;
    file.seekg(offset,ios::beg);
    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  dptr[i][j][k] = (bldata) BinRead<float>();
	  //	  cout << i << "," << j << "," << k << " " << dptr[i][j][k] << endl;
	  if (file.fail()) {
	    cout << "Warning failed reading file at " << i << "," << j << "," << k << endl;
	  }
	  //	  cout << hex << offset << dec << endl;
	  offset = offset + sizeof(float);
	  // Check if we have reached the end of a record and update
	  // the file pointer accordingly.
	  if (!norec) {
	    if ((offset + reclen) % (irecl+2*reclen) == 0) {
	      offset = offset + 2*reclen;
	      file.seekg(2*reclen,ios::cur);
	    }
	  }
	}
      }
    }

    // If this is one of the heat fluxes then clear the highest points.
    if ( (strncasecmp(name,"HTFLX",5) == 0) || 
	 (strncasecmp(name,"TAU",3) == 0) ) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  dptr[i][j][nz-1] = 0;
	}
      }
    }   

    // Create a new DataSet to contain the data.
    ptr->data = new DataSet(dptr,ptr->ndim,nx,ny,nz);

    // Add in the grids
    if ( (strncmp(name,"U",hnlen) == 0) || 
	 (strncasecmp(name,"HTFLXX",hnlen) == 0) ||
	 (strncasecmp(name,"U2AV",hnlen) == 0) ||
	 (strncasecmp(name,"U3AV",hnlen) == 0) ||
	 (strncasecmp(name,"UAV",hnlen) == 0) ) {
      ptr->data->SetGrids(xgrid, yngrid, zngrid, nxg, nyg, nzg, H, orogx);
    }
    else if ( (strncasecmp(name,"V",hnlen) == 0) || 
	      (strncasecmp(name,"HTFLXY",hnlen) == 0) || 
	      (strncasecmp(name,"V2AV",hnlen) == 0) || 
	      (strncasecmp(name,"V3AV",hnlen) == 0) || 
	      (strncasecmp(name,"VAV",hnlen) == 0) ) {
      ptr->data->SetGrids(xngrid, ygrid, zngrid, nxg, nyg, nzg, H, orogy);
    }
    else if ( (strncasecmp(name,"W",hnlen) == 0) || 
	      (strncasecmp(name,"VIS",hnlen) == 0) || 
	      (strncasecmp(name,"HTFLXZ",hnlen) == 0) || 
	      (strncasecmp(name,"W2AV",hnlen) == 0) || 
	      (strncasecmp(name,"W3AV",hnlen) == 0) || 
	      (strncasecmp(name,"WAV",hnlen) == 0) ) {
      ptr->data->SetGrids(xngrid, yngrid, zgrid, nxg, nyg, nzg, H, orog);
    }
    else if ( (strncasecmp(name,"VORTX",hnlen) == 0) ||  
	      (strncasecmp(name,"TAU23",hnlen) == 0) ||  
	      (strncasecmp(name,"T23AV",hnlen) == 0) ||
	      (strncasecmp(name,"VWAV",hnlen) == 0) ||
	      (strncasecmp(name,"B23",hnlen) == 0) ) {
      ptr->data->SetGrids(xngrid, ygrid, zgrid, nxg, nyg, nzg, H, orogy);
    }
    else if ( (strncasecmp(name,"VORTY",hnlen) == 0) ||  
	      (strncasecmp(name,"TAU13",hnlen) == 0) ||  
	      (strncasecmp(name,"T13AV",hnlen) == 0) ||
	      (strncasecmp(name,"UWAV",hnlen) == 0) ||
	      (strncasecmp(name,"B13",hnlen) == 0) ) {
      ptr->data->SetGrids(xgrid, yngrid, zgrid, nxg, nyg, nzg, H, orogx);
    }
    else if ( (strncasecmp(name,"VORTZ",hnlen) == 0) ||  
	      (strncasecmp(name,"TAU12",hnlen) == 0) ||  
	      (strncasecmp(name,"T12AV",hnlen) == 0) ||
	      (strncasecmp(name,"UVAV",hnlen) == 0) ||
	      (strncasecmp(name,"B12",hnlen) == 0) ) {
      ptr->data->SetGrids(xgrid, ygrid, zngrid, nxg, nyg, nzg, H, orogxy);
    }
    else if (strncasecmp(name,"GHALFN",hnlen) == 0) {
      ptr->data->SetGrids(xngrid, yngrid, nullgrid, nxg, nyg, 1, H, orog);
    }
    else if ( (strncasecmp(name,"X",hnlen)==0) ||
	      (strncasecmp(name,"XN",hnlen)==0) ||
	      (strncasecmp(name,"Y",hnlen)==0) ||
	      (strncasecmp(name,"YN",hnlen)==0) ||
	      (strncasecmp(name,"Z",hnlen)==0) ||
	      (strncasecmp(name,"ZN",hnlen)==0) ) {
      ptr->data->SetGrids(nullgrid, nullgrid, nullgrid, 1, 1, 1, H, (bldata **) NULL);
    }
    else {
      // This will not work correctly for those 1-D cases where 
      ptr->data->SetGrids(xngrid, yngrid, zngrid, nxg, nyg, nzg, H, orog);
    }

    // Perform conversion if this is the temperature
    if (strncasecmp(name,"T",hnlen) == 0) {
      ptr->data->ConvTemp(GetParm("THETA0"),GetParm("G"),false);
    }
    if ( (strncasecmp(name,"HTFLXX",hnlen) == 0) || 
	 (strncasecmp(name,"HTFLXY",hnlen) == 0) || 
	 (strncasecmp(name,"HTFLXZ",hnlen) == 0) ) {
      ptr->data->ConvTemp(GetParm("THETA0"),GetParm("G"),true);
    }

    // Set the name of the data set
    ptr->data->SetName(name);
    
  }  

  return ptr->data;
  
}

//
// Get the specified dataset. If it is not cached then load it from file.
//
TimeSeries * BlasFile::GetTimeSeries(const char name[hnlen]) {

  HeaderItem *ptr;
  bldata *dptr, *tptr;
  int npts;
  long offset;

  ptr = firstItem;

  while ((ptr != NULL) && (strncasecmp(name,ptr->name,hnlen)!=0)) {
    ptr = ptr->next;
  }

  if (ptr == NULL) {
    cerr << "Data " << name << " not found." << endl;
    return NULL;
  }

  if (ptr->tdata == NULL) {

    if (!file) {
      cout << "Failed before we even started!" << endl;
      return NULL;
    }
    npts = ptr->dim[0];
    if ((ptr->dim[1] != 2) || (ptr->dim[2] != 1)) {
      cerr << "Data is not a time series" << endl;
      return NULL;
    }
    //cout << name << " " << npts << endl;
    tptr = new bldata[npts];
    dptr = new bldata[npts];
    offset = ptr->offset;
    file.seekg(offset,ios::beg);
    ReadArray(dptr,npts,ptr,offset);
    ReadArray(tptr,npts,ptr,offset);

    // Create new timeseries
    ptr->tdata = new TimeSeries(dptr, tptr, npts);

    // Set the name of the data set
    ptr->tdata->SetName(name);
    
  }  

  return ptr->tdata;
  
}

//
// Read an array of data from the file
//
void
BlasFile::ReadArray(bldata *data, int n, HeaderItem *ptr, long &offset) {

  for (int i=0;i<n;i++) {
    data[i] = (bldata) BinRead<float>();
    if (file.fail()) {
      cerr << "Warning failed reading file at " << i << endl;
    }
    //cout << i << " " << data[i] << endl;
    offset = offset + sizeof(float);
    // Check if we have reached the end of a record and update
    // the file pointer accordingly.
    if (!norec) {
      if ((offset + reclen) % (irecl+2*reclen) == 0) {
	offset = offset + 2*reclen;
	file.seekg(2*reclen,ios::cur);
      }
    }
  }

}

//
// Get the specified parameter value
//
bldata BlasFile::GetParm(const char name[hnlen]) {

  HeaderItem * ptr;

  ptr = firstItem;

  while ((ptr != NULL) && (strncasecmp(name,ptr->name,hnlen)!=0)) {
    ptr = ptr->next;
  }

  if ((ptr == NULL) || (ptr->ndim != 0)) {
    cerr << "Parameter " << name << " not found." << endl;
    return 0;
  }

  return ptr->value;
}

//
// Get the corrected pressure field
//
DataSet *BlasFile::GetPressureCorr() {
  DataSet *pressure, *ghalf;
  int ndim, dimx, dimy, dimz;
  bldata *** p, ***g, ***data, htimav;
  
  if (pcorr == NULL) {
    pressure = GetData("P");
    ghalf = GetData("GHALFN");
    if (!pressure || !ghalf) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    htimav = GetParm("HTIMAV");    

    p = pressure->GetData();
    g = ghalf->GetData();

    ndim = pressure->GetDims();
    dimx = pressure->GetDim(1);
    dimy = pressure->GetDim(2);
    dimz = pressure->GetDim(3);

    // Create an empty data set
    pcorr = new DataSet(ndim, dimx, dimy, dimz);

    data = pcorr->GetData();
    
    // Calculate the corrected pressure
    for (int i = 0;i<dimx;i++) {
      for (int j = 0;j<dimy;j++) {
	for (int k = 0;k<dimz;k++) {
	  data[i][j][k] = p[i][j][k] + htimav*g[i][j][0]*zngrid[k];
	}
      }
    }
    pcorr->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);   
    pcorr->SetName("Corrected pressure");
  }
  
  return pcorr;  
}

//
// Get the speedup over a hill i.e. u / u_0
//
DataSet *BlasFile::GetSpeedup() {
  DataSet *u, *v, *w, *ui, *vi;
  int ndim, dimx, dimy, dimz;
  bldata *** data, ***ud, ***vd, ***wd, ***uid, ***vid, **h;
  
  bldata uu, vv, ww, uvi, u1, v1, z;

  //  return ReadFile::GetSpeedup();

  if (speedup == NULL) {
    u = GetData("U");
    v = GetData("V");
    w = GetData("W");
    ui = GetData("UI");
    vi = GetData("VI");
    if (!u || !v || !w || !ui || !vi) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();
    uid = ui->GetData();
    vid = vi->GetData();

    ndim = u->GetDims();
    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    h = orog->Get2D(3);

    // Create an empty data set
    speedup = new DataSet(ndim, dimx, dimy, dimz);

    data = speedup->GetData();

    for (int k = 0;k<dimz;k++) {      
      uvi = sqrt(uid[0][0][k]*uid[0][0][k] + vid[0][0][k]*vid[0][0][k]);
      for (int j = 0;j<dimy;j++) {
	for (int i = 0;i<dimx;i++) {
	  uu = ud[i][j][k];
	  z = zngrid[k]*(1-h[i][j]/H);
	  u1 = ui->GetVal(0,0,z);
	  v1 = vi->GetVal(0,0,z);
	  // uvi2 = sqrt(u1*u1+v1*v1);
	  if ((dimx > 1) && (i<dimx-1)) {
	    uu = (uu+ud[i+1][j][k])*0.5;
	  }
	  vv = vd[i][j][k];
	  if ((dimy > 1)&&(j<dimy-1)) {
	    vv = (vv+vd[i][j+1][k])*0.5;
	  }
	  ww = wd[i][j][k];
	  if ((dimz > 1)&&(k<dimz-1)) {
	    ww = (ww+wd[i][j][k+1])*0.5;
	  }
	  uu = uu;
	  vv = vv;
	  data[i][j][k] = sqrt(uu*uu + vv*vv + ww*ww) - uvi;
	  data[i][j][k] /= uvi;
	  //	  cout << i << " " << k << " " << z << " " << u1 << " " << v1 << endl;
	}
      }
    }

    speedup->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);
    speedup->SetName("Speedup");
  }

  return speedup;
  
}

//
// Load in the coordinate grids
//
void BlasFile::LoadGrids() {

  DataSet *gr;
  bldata ***ox, ***oy, ***oxy, ***oro;
  int odim;

  gr = GetData("X");
  if (gr == NULL) {
    cerr << "Unable to proceed without X grid" << endl;
    CloseFile();
    return;
  }
  xgrid = gr->Get1D(1);
  delete gr;
  gr = GetData("XN");
  if (gr == NULL) {
    cerr << "Unable to proceed without XN" << endl;
    CloseFile();
    return;
  }  
  xngrid = gr->Get1D(1);
  nxg = gr->GetDim(1);
  delete gr;
//   for (int i=0;i<nxg;i++) {
//     cout << xgrid[i] << " " << xngrid[i] << "\n";
//   }
//   cout << endl;


  gr = GetData("Y");
  if (gr == NULL) {
    cerr << "Unable to proceed without Y grid" << endl;
    CloseFile();
    return;
  }
  ygrid = gr->Get1D(1);
  delete gr;
  gr = GetData("YN");
  if (gr == NULL) {
    cerr << "Unable to proceed without YN grid" << endl;
    CloseFile();
    return;
  }
  yngrid = gr->Get1D(1);
  nyg = gr->GetDim(1);
  delete gr;
  //  for (int i=0;i<nyg;i++) {
  //    cout << ygrid[i] << " " << yngrid[i] << "\n";
  //  }
  //  cout << endl;

  gr = GetData("Z");
  if (gr == NULL) {
    cerr << "Unable to proceed without Z grid" << endl;
    CloseFile();
    return;
  }
  zgrid = gr->Get1D(1);
  delete gr;
  gr = GetData("ZN");
  if (gr == NULL) {
    cerr << "Unable to proceed without ZN grid" << endl;
    CloseFile();
    return;
  }
  zngrid = gr->Get1D(1);
  nzg = gr->GetDim(1);
  delete gr;
  //  for (int i=0;i<nzg;i++) {
  //    cout << zgrid[i] << " " << zngrid[i] << "\n";
  //  }
  //  cout << endl;

  H = zgrid[nzg-2];

  // If LEM there is no orography
  if (model == lem) {
    odim = 1;
    if ((nxg > 1)&&(nyg > 1)) odim++;
    orog = new DataSet(odim,nxg,nyg,1);
    oro = orog->GetData();
    orogx = new DataSet(odim,nxg,nyg,1);
    ox = orogx->GetData();
    orogy = new DataSet(odim,nxg,nyg,1);
    oy = orogy->GetData();
    orogxy = new DataSet(odim,nxg,nyg,1);
    oxy = orogxy->GetData();

    for (int i=0;i<nxg;i++) {
      for (int j=0;j<nyg;j++) {
	oro[i][j][0] = 0.0;
	ox[i][j][0] = 0.0;
	oy[i][j][0] = 0.0;
	oxy[i][j][0] = 0.0;
      }
    }
    orog->SetGrids(xngrid,yngrid,nullgrid,nxg,nyg,1,H,orog);
    orog->SetName("ZSN");
    orogx->SetGrids(xgrid,yngrid,nullgrid,nxg,nyg,1,H,orog);
    orogx->SetName("ZSX");
    orogy->SetGrids(xngrid,ygrid,nullgrid,nxg,nyg,1,H,orog);
    orogy->SetName("ZSY");
    orogxy->SetGrids(xgrid,ygrid,nullgrid,nxg,nyg,1,H,orog);
    orogxy->SetName("ZSXY");
  }
  else {
    orog = GetData("ZSN");
    if (orog == NULL) {
      cerr << "Unable to proceed without Z grid" << endl;
      CloseFile();
      return;
    }
      
    if (nxg == 1) {
      orog->SwapData();
    }
    orog->SetGrids(xngrid,yngrid,nullgrid,nxg,nyg,1,H,orog);
    
    orogx = GetData("ZSX");
    if (nxg == 1) {
      orogx->SwapData();
    }
    orogx->SetGrids(xgrid,yngrid,nullgrid,nxg,nyg,1,H,orogx);
    
    orogy = GetData("ZSY");
    if (nxg == 1) {
      orogy->SwapData();
    }
    orogy->SetGrids(xngrid,ygrid,nullgrid,nxg,nyg,1,H,orogy);

    orogxy = new DataSet(orogy->GetDims(),nxg,nyg,1);
    
    ox = orogx->GetData();
    oy = orogy->GetData();
    oxy = orogxy->GetData();
    
    for (int i=0;i<nxg-1;i++) {
      for (int j=0;j<nyg-1;j++) {
	oxy[i][j][0] = 0.25*(oy[i][j][0]+oy[i][j+1][0]+ox[i][j][0]+ox[i+1][j][0]);
      }
      oxy[i][nyg-1][0] = 0.5*(ox[i][nyg-1][0]+ox[i+1][nyg-1][0]);
    }
    for (int j=0;j<nyg-1;j++) {
      oxy[nxg-1][j][0] = 0.5*(oy[nxg-1][j][0]+oy[nxg-1][j+1][0]);
    }
    oxy[nxg-1][nyg-1][0] = 0.5*(ox[nxg-1][nyg-1][0]+oy[nxg-1][nyg-1][0]);
    orogxy->SetGrids(xgrid,ygrid,nullgrid,nxg,nyg,1,H,orogxy);
    orogxy->SetName("ZSXY");
  }
  
}


//
// Get the surface stress components, rotate and turn into a 
// vector
//
DataVect *BlasFile::GetTau() {
  DataVect *tau;
  DataSet *taux, *tauy, *cosalp, *sinalp, *tau1, *tau2;
  bldata ***tauxd, ***tauyd, ***cosalpd, ***sinalpd, ***data1, ***data2;
  int ndim, dimx, dimy, dimz;

  taux = GetData("TAUX");
  tauy = GetData("TAUY");
  cosalp = GetData("COSALP");
  sinalp = GetData("SINALP");

  if (!taux || !tauy || !cosalp || !sinalp) {
    cerr << "Unable to read all required data" << endl;
    return NULL;
  }

  tauxd = taux->GetData();
  tauyd = tauy->GetData();
  cosalpd = cosalp->GetData();
  sinalpd = sinalp->GetData();

  ndim = taux->GetDims();
  dimx = taux->GetDim(1);
  dimy = taux->GetDim(2);
  dimz = taux->GetDim(3);

  tau1 = new DataSet(ndim, dimx, dimy, dimz);
  tau2 = new DataSet(ndim, dimx, dimy, dimz);

  data1 = tau1->GetData();
  data2 = tau2->GetData();

  for (int k=0; k<dimz; k++) {
    for (int j=0; j<dimy; j++) {
      for (int i=0; i<dimx; i++) {
	data1[i][j][k] = tauxd[i][j][k]*cosalpd[i][j][k] - 
	  tauyd[i][j][k]*sinalpd[i][j][k];
	data2[i][j][k] = tauyd[i][j][k]*cosalpd[i][j][k] +
	  tauxd[i][j][k]*sinalpd[i][j][k];
      }
    }
  }

  tau1->SetGrids(xngrid, yngrid, zngrid, dimx, dimy, dimz, H, orog);
  tau2->SetGrids(xngrid, yngrid, zngrid, dimx, dimy, dimz, H, orog);
  
  tau1->SetName("Tau_x");
  tau2->SetName("Tau_y");

  tau = new DataVect(tau1, tau2, NULL);

  return tau;
}


//
// Get the source term for PV
//
DataSet *BlasFile::GetPVSource() {
  
  DataSet *ds;
  bldata ***tau11, ***tau12, ***tau13, ***tau22, ***tau23, ***tau33, 
    ***theta, ***hflx, ***hfly, ***hflz, ***rho; 
  bldata ***Fx, ***Fy, ***Fz, ***curlfx, ***curlfy, ***curlfz;
  bldata ***gradthx,***gradthy,***gradthz;
  bldata ***divh, ***delhx, ***delhy, ***delhz;
  bldata ***vortx, ***vorty, ***vortz;
  bldata ***vis, ***vish, ***rigrad;
  bldata **orx, **ory, **oro;
  bldata ***src;
  bldata theta0, g, fcor, busdyr, rpr;
  bldata Cx, Cy, Cxu, Cyu, Cxv, Cyv, Cxuv, Cyuv;
  bldata ghalf, ghalfx, ghalfy, ghalfxy;
  bldata rpr1, rpr2, rpr3, rpr4, rpr5, rpr6;
  int i1, i2, j1, j2;


  if (PVsrc != NULL) 
    return PVsrc;

  // Read in stresses and heat fluxes
  // Note these are all multiplied by the metric term G^(1/2)

  ds = GetData("TAU11");
  if (ds == NULL) {
    cerr << "Unable to read TAU11" << endl;
    return NULL;
  }
  tau11 = ds->GetData();

  ds = GetData("TAU12");
  if (ds == NULL) {
    cerr << "Unable to read TAU12" << endl;
    return NULL;
  }
  tau12 = ds->GetData();

  ds = GetData("TAU13");
  if (ds == NULL) {
    cerr << "Unable to read TAU13" << endl;
    return NULL;
  }
  tau13 = ds->GetData();

  ds = GetData("TAU22");
  if (ds == NULL) {
    cerr << "Unable to read TAU22" << endl;
    return NULL;
  }
  tau22 = ds->GetData();

  ds = GetData("TAU23");
  if (ds == NULL) {
    cerr << "Unable to read TAU23" << endl;
    return NULL;
  }
  tau23 = ds->GetData();

  ds = GetData("TAU33");
  if (ds == NULL) {
    cerr << "Unable to read TAU33" << endl;
    return NULL;
  }
  tau33 = ds->GetData();

  ds = GetData("T");
  if (ds == NULL) {
    cerr << "Unable to read T" << endl;
    return NULL;
  }
  theta = ds->GetData();

  ds = GetData("VIS");
  if (ds == NULL) {
    cerr << "Unable to read VIS" << endl;
    return NULL;
  }
  vis = ds->GetData();

  ds = GetData("VISH");
  if (ds == NULL) {
    vish = vis;
  }
  else {
    vish = ds->GetData();
  }

  ds = GetData("RIGRAD");
  if (ds == NULL) {
    cerr << "Unable to read RIGRAD" << endl;
    return NULL;
  }
  rigrad = ds->GetData();


   ds = GetData("HTFLXX");
   if (ds == NULL) {
     cerr << "Unable to read HTFLXX" << endl;
     return NULL;
   }
   hflx = ds->GetData();

   ds = GetData("HTFLXY");
   if (ds == NULL) {
     cerr << "Unable to read HTFLXY" << endl;
     return NULL;
   }
   hfly = ds->GetData();

   ds = GetData("HTFLXZ");
   if (ds == NULL) {
     cerr << "Unable to read HTFLXZ" << endl;
     return NULL;
   }
   hflz = ds->GetData();

  ds = GetData("VORTX");
  if (ds == NULL) {
    cerr << "Unable to read VORTX" << endl;
    return NULL;
  }
  vortx = ds->GetData();

  ds = GetData("VORTY");
  if (ds == NULL) {
    cerr << "Unable to read VORTY" << endl;
    return NULL;
  }
  vorty = ds->GetData();

  ds = GetData("VORTZ");
  if (ds == NULL) {
    cerr << "Unable to read VORTZ" << endl;
    return NULL;
  }
  vortz = ds->GetData();

  ds = GetData("RHO");
  if (ds == NULL) {
    cerr << "Unable to read RHO" << endl;
    return NULL;
  }
  rho = ds->GetData();

  theta0 = GetParm("THETA0");
  g = GetParm("G");
  //  H = GetParm("H");
  fcor = GetParm("FCOR");
  rpr = 1.0/GetParm("PR");
  busdyr = GetParm("BUSDYR");

  // Rescale T to get theta - what!!! This has already been done.
  //  for (int k=0; k<nzg; k++) {
  //    for (int j=0; j<nyg; j++) {
  //      for (int i=0; i<nxg; i++) {
  //	theta[i][j][k] = theta[i][j][k]*theta0/g + theta0;
  //      }
  //    }
  //  }

  // Get orography
  oro = orog->Get2D(3);
  orx = orogx->Get2D(3);
  ory = orogy->Get2D(3);

  // All fields are now correctly read in.

  // Create arrays for F
  Fx = Alloc3DArray<bldata>(nxg,nyg,nzg);
  Fy = Alloc3DArray<bldata>(nxg,nyg,nzg);
  Fz = Alloc3DArray<bldata>(nxg,nyg,nzg);

  // Create arrays for heat flux hfl
  //  hflx = Alloc3DArray<bldata>(nxg,nyg,nzg);
  //  hfly = Alloc3DArray<bldata>(nxg,nyg,nzg);
  //  hflz = Alloc3DArray<bldata>(nxg,nyg,nzg);

  // Create arrays for curl F
  curlfx = Alloc3DArray<bldata>(nxg,nyg,nzg);
  curlfy = Alloc3DArray<bldata>(nxg,nyg,nzg);
  curlfz = Alloc3DArray<bldata>(nxg,nyg,nzg);

  divh = Alloc3DArray<bldata>(nxg,nyg,nzg);

  // Create the data set for the output
  PVsrc = new DataSet(3, nxg, nyg, nzg);
  src = PVsrc->GetData();

  // Zero all data sets
  for (int k=0; k<nzg; k++) {
    for (int j=0; j<nyg; j++) {
      for (int i=0; i<nxg; i++) {
	Fx[i][j][k] = 0;
	Fy[i][j][k] = 0;
	Fz[i][j][k] = 0;
	//	hflx[i][j][k] = 0;
	//	hfly[i][j][k] = 0;
	//	hflz[i][j][k] = 0;
	curlfx[i][j][k] = 0;
	curlfy[i][j][k] = 0;
	curlfz[i][j][k] = 0;
	divh[i][j][k] = 0;
	src[i][j][k] = 0;
      }
    }
  }

  //
  // First calculate dissipation term, F, on RHS of the momentum equation
  //

  // F_i = rho^(-1) * d(tau_ij) / d(x_j)
  // Fx is on a U point, etc...
  // Actually this is G^(1/2) * F_i
  for (int j=0; j<nyg; j++) {
    for (int i=0; i<nxg; i++) {
      if (i > 0)
	Cx = -(orx[i][j]-orx[i-1][j])/(xgrid[i]-xgrid[i-1]);
      else
	Cx = 0;
      if (j > 0)	
	Cy = -(ory[i][j]-ory[i][j-1])/(ygrid[j]-ygrid[j-1]);
      else 
	Cy = 0;

      if (i < nxg-1)
	Cxu = -(oro[i+1][j]-oro[i][j])/(xngrid[i+1]-xngrid[i]);
      else 
	Cxu = 0;

      if ((i > 0) && (j < nyg-1))
	Cxv = -0.5*(orx[i][j]+orx[i][j+1] - orx[i-1][j]-orx[i-1][j+1])/(xgrid[i]-xgrid[i-1]);
      else 
	Cxv = 0;
      
      if ((j > 0) && (i < nxg-1))
	Cyu = -0.5*(ory[i][j]+ory[i+1][j] - ory[i][j-1]-orx[i+1][j-1])/(ygrid[j]-ygrid[j-1]);
      else 
	Cyu = 0;

      if (j < nyg-1)
	Cyv = -(oro[i][j+1]-oro[i][j])/(yngrid[j+1]-yngrid[j]);
      else 
	Cyv = 0;

      i1 = i-1;
      i2 = i+1;
      if (i == 0) i1 = i;
      if (i == nxg-1) i2 = i;
      j1 = j-1;
      j2 = j+1;
      if (j == 0) j1 = j;
      if (j == nyg-1) j2 = j;

      for (int k=1; k<nzg-1; k++) {
	if (i < nxg-1) {
	  Fx[i][j][k] += (tau11[i+1][j][k] - tau11[i][j][k]) / 
	    (xngrid[i+1]-xngrid[i])
	    + 0.5*((tau11[i][j][k+1]+tau11[i+1][j][k+1])*(1 - zngrid[k+1]/H) - 
		   (tau11[i][j][k-1]+tau11[i+1][j][k-1])*(1 - zngrid[k-1]/H))
	    *Cxu / (zngrid[k+1]-zngrid[k-1]);
	}
	if (j > 0) {
	  Fx[i][j][k] += (tau12[i][j][k] - tau12[i][j-1][k]) / 
	    (ygrid[j]-ygrid[j-1])
	    + 0.5*((tau12[i][j][k+1]+tau12[i][j-1][k+1])*(1 - zngrid[k+1]/H) - 
		   (tau12[i][j][k-1]+tau12[i][j-1][k-1])*(1 - zngrid[k-1]/H))
	    *Cyu / (zngrid[k+1]-zngrid[k-1]);
	}
	Fx[i][j][k] += (tau13[i][j][k]-tau13[i][j][k-1])/(zgrid[k]-zgrid[k-1]);
	Fx[i][j][k] /= 0.5*(rho[i2][j][k]+rho[i][j][k]);
	
	if (j < nyg-1) {
	  Fy[i][j][k] = +(tau22[i][j+1][k] - tau22[i][j][k]) / 
	    (yngrid[j+1]-yngrid[j])
	    + 0.5*((tau22[i][j][k+1]+tau22[i][j+1][k+1])*(1 - zngrid[k+1]/H) - 
		   (tau22[i][j][k-1]+tau22[i][j+1][k-1])*(1 - zngrid[k-1]/H))
	    *Cyv / (zngrid[k+1]-zngrid[k-1]);
	}
	if (i > 0) {
	  Fy[i][j][k] += (tau12[i][j][k] - tau12[i-1][j][k]) / 
	    (xgrid[i]-xgrid[i-1])
	    + 0.5*((tau12[i][j][k+1]+tau12[i-1][j][k+1])*(1 - zngrid[k+1]/H) - 
		   (tau12[i][j][k-1]+tau12[i-1][j][k-1])*(1 - zngrid[k-1]/H))
	    *Cxv / (zngrid[k+1]-zngrid[k-1]);
	}
	Fy[i][j][k] += (tau23[i][j][k]-tau23[i][j][k-1])/(zgrid[k]-zgrid[k-1]);
	Fy[i][j][k] /= 0.5*(rho[i][j2][k]+rho[i][j][k]);	
	
	if (i>0) {
	  Fz[i][j][k] = +(tau13[i][j][k] - tau13[i-1][j][k])/(xgrid[i]-xgrid[i-1])
	    + 0.5*( (tau13[i][j][k+1]+tau13[i-1][j][k+1])*(1-zgrid[k+1]/H)
		    - (tau13[i][j][k-1]+tau13[i-1][j][k-1])*(1-zgrid[k-1]/H))
	    * Cx / (zgrid[k+1]-zgrid[k-1]);
	}
	if (j>0) {
	  Fz[i][j][k] += (tau23[i][j][k] - tau23[i][j-1][k])/(ygrid[j]-ygrid[j-1])
	    + 0.5*( (tau23[i][j][k+1]+tau23[i][j-1][k+1])*(1-zgrid[k+1]/H)
		    - (tau23[i][j][k-1]+tau23[i][j-1][k-1])*(1-zgrid[k-1]/H))
	    * Cy / (zgrid[k+1]-zgrid[k-1]);
	}
	Fz[i][j][k] += (tau33[i][j][k+1]-tau33[i][j][k])/(zngrid[k+1]-zngrid[k]);
	Fz[i][j][k] /= 0.5*(rho[i][j][k+1]+rho[i][j][k]);
	
      }
      
      // Do boundary values - assume d/dz = 0 on boundary ***
      
      if (i < nxg-1) {
	Fx[i][j][0] += (tau11[i+1][j][0] - tau11[i][j][0]) / 
	  (xngrid[i+1]-xngrid[i]);
	Fx[i][j][nzg-1] += (tau11[i+1][j][nzg-1] - tau11[i][j][nzg-1]) / 
	  (xngrid[i+1]-xngrid[i]);
      }

      if (j > 0) {
	Fx[i][j][0] += (tau12[i][j][0] - tau12[i][j-1][0]) / 
	  (ygrid[j]-ygrid[j-1]);
	Fx[i][j][nzg-1] += (tau12[i][j][nzg-1] - tau12[i][j-1][nzg-1]) / 
	  (ygrid[j]-ygrid[j-1]);
      }
      Fx[i][j][nzg-1] += (tau13[i][j][nzg-1]-tau13[i][j][nzg-2])
	/(zgrid[nzg-1]-zgrid[nzg-2]);
      Fx[i][j][0] /= 0.5*(rho[i2][j][0]+rho[i][j][0]);
      Fx[i][j][nzg-1] /= 0.5*(rho[i2][j][nzg-1]+rho[i][j][nzg-1]);
      
      if (j < nyg-1) {
	Fy[i][j][0] += (tau22[i][j+1][0] - tau22[i][j][0]) / 
	  (yngrid[j+1]-yngrid[j]);
	Fy[i][j][nzg-1] += (tau22[i][j+1][nzg-1] - tau22[i][j][nzg-1]) / 
	  (yngrid[j+1]-yngrid[j]);
      }
      if (i > 0) {
	Fy[i][j][0] += (tau12[i][j][0] - tau12[i-1][j][0]) / 
	  (xgrid[i]-xgrid[i-1]);
	Fy[i][j][nzg-1] += (tau12[i][j][nzg-1] - tau12[i-1][j][nzg-1]) / 
	  (xgrid[i]-xgrid[i-1]);
      }
      Fy[i][j][nzg-1] += (tau23[i][j][nzg-1]-tau23[i][j][nzg-2])
	/(zgrid[nzg-1]-zgrid[nzg-2]);
      Fy[i][j][0] /= 0.5*(rho[i][j2][0]+rho[i][j][0]);      
      Fy[i][j][nzg-1] /= 0.5*(rho[i][j2][nzg-1]+rho[i][j][nzg-1]);
      
	
      if (i > 0) {
	Fz[i][j][0] += (tau13[i][j][0] - tau13[i-1][j][0])
	  /(xgrid[i]-xgrid[i-1]);
	Fz[i][j][nzg-1] += (tau13[i][j][nzg-1] - tau13[i-1][j][nzg-1])
	  /(xgrid[i]-xgrid[i-1]);
      }
      if (j > 0) { 
	Fz[i][j][0] += (tau23[i][j][0] - tau23[i][j-1][0])
	  /(ygrid[j]-ygrid[j-1]);
	Fz[i][j][nzg-1] += (tau23[i][j][nzg-1] - tau23[i][j-1][nzg-1])
	  /(ygrid[j]-ygrid[j-1]);
      }
      Fz[i][j][0] += (tau33[i][j][1]-tau33[i][j][0])/(zngrid[1]-zngrid[0]);
      Fz[i][j][0] /= 0.5*(rho[i][j][1]+rho[i][j][0]);	
      Fz[i][j][nzg-1] /= rho[i][j][nzg-1];      
	
    }
  }

  // Keep d/dx and d/dy = 0 at boundary. Is this OK??? ***
  // Actually for nothing for now - don't care there.
  /*
  for (int k = 0; k < nzg; k++) {
    for (int j = 0; j < nyg; j++) {
    }
  }
  */

  // *** What do we do about boundary conditions? ***

  // Next want to calculate curl(F)
  // curl(F)_x is on a VW point etc
  for (int j=0; j<nyg; j++) {
    for (int i=0; i<nxg; i++) { 
      if (i < nxg-1) {
	Cxuv = -(ory[i+1][j]-ory[i][j])/(xngrid[i+1]-xngrid[i]);
	Cxu = -(oro[i+1][j]-oro[i][j])/(xngrid[i+1]-xngrid[i]);
      }
      else {
	Cxuv = 0;
	Cxu = 0;
      }
      if (j < nyg-1) {
	Cyuv = -(orx[i][j+1]-orx[i][j])/(yngrid[j+1]-yngrid[j]);
	Cyv = -(oro[i][j+1]-oro[i][j])/(yngrid[j+1]-yngrid[j]);
      }
      else {
	Cyuv = 0;
	Cyv = 0;
      }

      ghalfy = (1-ory[i][j]/H);
      ghalfx = (1-orx[i][j]/H);
      if ((i < nxg-1) && (j < nyg-1)) 
	ghalfxy = (1-0.25*(orx[i][j]+orx[i][j+1]+ory[i][j]+ory[i+1][j])/H);
      else 
	ghalfxy = 0;

      // Calculate G^(1/2)*curl(F) then divide by G^(1/2)
      for (int k=1; k<nzg-1; k++) {
	if (j < nyg-1) {
	  curlfx[i][j][k] = (Fz[i][j+1][k]-Fz[i][j][k])/(yngrid[j+1]-yngrid[j])
	    + 0.5*((Fz[i][j][k+1]+Fz[i][j+1][k+1])*(1 - zgrid[k+1]/H) - 
		   (Fz[i][j][k-1]+Fz[i][j+1][k-1])*(1 - zgrid[k-1]/H))
	    *Cyv / (zgrid[k+1]-zgrid[k-1])
	    - (Fy[i][j][k+1]-Fy[i][j][k])/(zngrid[k+1]-zngrid[k]);
	  curlfx[i][j][k] /= ghalfy;
	}

	if (i < nxg-1) {
	  curlfy[i][j][k] = -(Fz[i+1][j][k]-Fz[i][j][k])/(xngrid[i+1]-xngrid[i])
	    - 0.5*((Fz[i][j][k+1]+Fz[i+1][j][k+1])*(1 - zgrid[k+1]/H) - 
		   (Fz[i][j][k-1]+Fz[i+1][j][k-1])*(1 - zgrid[k-1]/H))
	    *Cxu / (zgrid[k+1]-zgrid[k-1])
	    + (Fx[i][j][k+1]-Fx[i][j][k])/(zngrid[k+1]-zngrid[k]);
	  curlfy[i][j][k] /= ghalfx;
	}
	
	if ( (i < nxg-1) && (j < nyg-1) ) {
	  curlfz[i][j][k] = (Fy[i+1][j][k]-Fy[i][j][k])/(xngrid[i+1]-xngrid[i])
	    + 0.5*((Fy[i][j][k+1]+Fy[i+1][j][k+1])*(1 - zngrid[k+1]/H) - 
		   (Fy[i][j][k-1]+Fy[i+1][j][k-1])*(1 - zngrid[k-1]/H))
	    *Cxuv / (zngrid[k+1]-zngrid[k-1])
	    - (Fx[i][j+1][k]-Fx[i][j][k])/(yngrid[j+1]-yngrid[j])
	    - 0.5*((Fx[i][j][k+1]+Fx[i][j+1][k+1])*(1 - zngrid[k+1]/H) - 
		   (Fx[i][j][k-1]+Fx[i][j+1][k-1])*(1 - zngrid[k-1]/H))
	    *Cyuv / (zngrid[k+1]-zngrid[k-1]);
	  curlfz[i][j][k] /= ghalfxy;
	}
      }

      if (j < nyg-1) {
	curlfx[i][j][0] = (Fz[i][j+1][0]-Fz[i][j][0])/(yngrid[j+1]-yngrid[j])
	  - (Fy[i][j][1]-Fy[i][j][0])/(zngrid[1]-zngrid[0]);
	curlfx[i][j][0] /= ghalfy;      

      curlfx[i][j][nzg-1] = (Fz[i][j+1][nzg-1]-Fz[i][j][nzg-1])
	/(yngrid[j+1]-yngrid[j]);
      curlfx[i][j][nzg-1] /= ghalfy;
      }

      if (i < nxg-1) {
	curlfy[i][j][0] = -(Fz[i+1][j][0]-Fz[i][j][0])/(xngrid[i+1]-xngrid[i])
	  + (Fx[i][j][1]-Fx[i][j][0])/(zngrid[1]-zngrid[0]);
	curlfy[i][j][0] /= ghalfx;
	
	curlfy[i][j][nzg-1] = -(Fz[i+1][j][nzg-1]-Fz[i][j][nzg-1])
	  /(xngrid[i+1]-xngrid[i]);
	curlfy[i][j][nzg-1] /= ghalfx;
      }
	
      if ((i < nxg-1) && (j < nyg-1)) {
	curlfz[i][j][0] = (Fy[i+1][j][0]-Fy[i][j][0])/(xngrid[i+1]-xngrid[i])
	  - (Fx[i][j+1][0]-Fx[i][j][0])/(yngrid[j+1]-yngrid[j]);
	curlfz[i][j][0] /= ghalfxy;
	
	curlfz[i][j][nzg-1] = (Fy[i+1][j][nzg-1]-Fy[i][j][nzg-1])
	  /(xngrid[i+1]-xngrid[i])
	- (Fx[i][j+1][nzg-1]-Fx[i][j][nzg-1])/(yngrid[j+1]-yngrid[j]);
	curlfz[i][j][nzg-1] /= ghalfxy;
      }
    }
  }  

  // Recycle arrays now
  gradthx = Fx;
  gradthy = Fy;
  gradthz = Fz;

  // Make sure gradth? is blank
  for (int k=0; k<nzg;k++) {
    for (int j=0; j<nyg;j++) {
      for (int i=0; i<nxg;i++) {
	gradthx[i][j][k] = 0;
	gradthy[i][j][k] = 0;
	gradthz[i][j][k] = 0;
      }
    }
  }


  // Now calculate grad(theta) 
  // grad(theta)_x is on u points etc.
  for (int j=0; j<nyg; j++) {
    for (int i=0; i<nxg; i++) {
      if (i < nxg-1) {
	Cxu = -(oro[i+1][j]-oro[i][j])/(xngrid[i+1]-xngrid[i]);
      }
      else {
	Cxu = 0;
      }
      if (j < nyg-1) {
	Cyv = -(oro[i][j+1]-oro[i][j])/(yngrid[j+1]-yngrid[j]);
      }
      else {
	Cyv = 0;
      }
      ghalfx = (1-orx[i][j]/H);
      ghalfy = (1-ory[i][j]/H);
      ghalf = (1-oro[i][j]/H);
      for (int k=1; k<nzg-1; k++) {
	if (i < nxg-1) {
	  gradthx[i][j][k] = (theta[i+1][j][k]-theta[i][j][k])
	    /(xngrid[i+1]-xngrid[i])
	    + Cxu * (1-zgrid[k]/H)/ghalfx
	    * 0.5*( (theta[i][j][k+1]+theta[i+1][j][k+1]) -
		    (theta[i][j][k-1]+theta[i+1][j][k-1]) ) 
	    / (zngrid[k+1]-zngrid[k-1]);
	}

      	if (j < nyg-1) {
	  gradthy[i][j][k] = (theta[i][j+1][k]-theta[i][j][k])
	    /(yngrid[j+1]-yngrid[j])
	    + Cyv * (1-zgrid[k]/H)/ghalfy
	    * 0.5*( (theta[i][j][k+1]+theta[i][j+1][k+1]) -
		    (theta[i][j][k-1]+theta[i][j+1][k-1]) ) 
	    / (zngrid[k+1]-zngrid[k-1]);
	}
	  
	gradthz[i][j][k] = (theta[i][j][k+1]-theta[i][j][k])
	  / (zngrid[k+1]-zngrid[k]) / ghalf;
      }

      if (i < nxg-1) {
	gradthx[i][j][0] = (theta[i+1][j][0]-theta[i][j][0])
	  /(xngrid[i+1]-xngrid[i]);
	
	gradthx[i][j][nzg-1] = (theta[i+1][j][nzg-1]-theta[i][j][nzg-1])
	  /(xngrid[i+1]-xngrid[i]);
      }
	
      if (j < nyg-1) {
	gradthy[i][j][0] = (theta[i][j+1][0]-theta[i][j][0])
	  /(yngrid[j+1]-yngrid[j]);
	
	gradthy[i][j][nzg-1] = (theta[i][j+1][nzg-1]-theta[i][j][nzg-1])
	  /(yngrid[j+1]-yngrid[j]);
      }

      gradthz[i][j][0] = (theta[i][j][1]-theta[i][j][0])
	  / (zngrid[1]-zngrid[0]) / ghalf;

      gradthz[i][j][nzg-1] = 0;
    }
  }

  // Work out the theta_dot term on p points

  // Recycle array
  delhx = Fx;
  delhy = Fy;
  delhz = Fz;

  // First work out heat flux using theta rather than 
  // output heat flux (which is for T).
  if (false) {
  for (int j=1; j<nyg-1; j++) {
    for (int i=1; i<nxg-1; i++) {      

      Cxu = -(oro[i+1][j]-oro[i][j])/(xngrid[i+1]-xngrid[i]);
      Cyv = -(oro[i][j+1]-oro[i][j])/(yngrid[j+1]-yngrid[j]);
      ghalf = (1-oro[i][j]/H);

      for (int k=1; k<nzg-1; k++) {
	// Strictly need to check Richardson number here!
	// and include prandtl number effects
	rpr1 = rpr;
	rpr2 = rpr;
	rpr3 = rpr;
	rpr4 = rpr;
	rpr5 = rpr;
	rpr6 = rpr;

	if (rigrad[i][j][k] < 0.0) {
	  rpr1 = rpr*pow((1.0-busdyr*rpr*rigrad[i][j][k]),0.25);
	}
	if (rigrad[i+1][j][k] < 0.0) {
	  rpr2 = rpr*pow((1.0-busdyr*rpr*rigrad[i+1][j][k]),0.25);
	}
	if (rigrad[i][j][k-1] < 0.0) {
	  rpr3 = rpr*pow((1.0-busdyr*rpr*rigrad[i][j][k-1]),0.25);
	}
	if (rigrad[i+1][j][k-1] < 0.0) {
	  rpr4 = rpr*pow((1.0-busdyr*rpr*rigrad[i+1][j][k-1]),0.25);
	}
	if (rigrad[i][j+1][k] < 0.0) {
	  rpr5 = rpr*pow((1.0-busdyr*rpr*rigrad[i][j+1][k]),0.25);
	}
	if (rigrad[i][j+1][k-1] < 0.0) {
	  rpr6 = rpr*pow((1.0-busdyr*rpr*rigrad[i][j+1][k-1]),0.25);
	}

	hflx[i][j][k] = 0.125*(rho[i+1][j][k]+rho[i][j][k])
	  *(rpr1*vish[i][j][k]+rpr2*vish[i+1][j][k]+rpr3*vish[i][j][k-1]
	    +rpr4*vish[i+1][j][k-1])
	  *(((1-oro[i+1][j]/H)*theta[i+1][j][k]-(1-oro[i][j]/H)*theta[i][j][k])
	    / (xngrid[i+1]-xngrid[i])
	    + 0.5*((theta[i][j][k+1]+theta[i+1][j][k+1])*(1 - zngrid[k+1]/H) - 
		   (theta[i][j][k-1]+theta[i+1][j][k-1])*(1 - zngrid[k-1]/H))
	    *Cxu / (zngrid[k+1]-zngrid[k-1]));
	
	hfly[i][j][k] = 0.125*(rho[i][j+1][k]+rho[i][j][k])
	  *(rpr1*vish[i][j][k]+rpr5*vish[i][j+1][k]+rpr3*vish[i][j][k-1]
	    +rpr6*vish[i][j+1][k-1])
	  *(((1-oro[i][j+1]/H)*theta[i][j+1][k]-(1-oro[i][j]/H)*theta[i][j][k])
	    / (yngrid[j+1]-yngrid[j])
	    + 0.5*((theta[i][j][k+1]+theta[i][j+1][k+1])*(1 - zngrid[k+1]/H) - 
		   (theta[i][j][k-1]+theta[i][j+1][k-1])*(1 - zngrid[k-1]/H))
	    *Cyv / (zngrid[k+1]-zngrid[k-1]));
	
	hflz[i][j][k] = 0.5*(rho[i][j][k+1]+rho[i][j][k])*vis[i][j][k]
	  *(theta[i][j][k+1]-theta[i][j][k])/(zngrid[k+1]-zngrid[k]);
      }
    }
  }
  }

  // Calculate div(H) on p points
  for (int j=0; j<nyg; j++) {
    for (int i=0; i<nxg; i++) {

      if (i > 0)
	Cx = -(orx[i][j]-orx[i-1][j])/(xgrid[i]-xgrid[i-1]);
      else
	Cx = 0;
      if (j > 0)
	Cy = -(ory[i][j]-ory[i][j-1])/(ygrid[j]-ygrid[j-1]);
      else 
	Cy = 0;
      ghalf = (1-oro[i][j]/H);

      for (int k=1; k<nzg-1; k++) {
	divh[i][j][k] = 0;
	if (i > 0) {
	  divh[i][j][k] += (hflx[i][j][k]-hflx[i-1][j][k])/(xgrid[i]-xgrid[i-1])
	    + 0.5*((hflx[i][j][k+1]+hflx[i-1][j][k+1])*(1 - zngrid[k+1]/H) - 
		   (hflx[i][j][k-1]+hflx[i-1][j][k-1])*(1 - zngrid[k-1]/H))
	    *Cx / (zngrid[k+1]-zngrid[k-1]);
	}
	if (j > 0) {
	    divh[i][j][k] += (hfly[i][j][k]-hfly[i][j-1][k])/(ygrid[j]-ygrid[j-1])
	      + 0.5*((hfly[i][j][k+1]+hfly[i][j-1][k+1])*(1 - zngrid[k+1]/H) - 
		     (hfly[i][j][k-1]+hfly[i][j-1][k-1])*(1 - zngrid[k-1]/H))
	      *Cy / (zngrid[k+1]-zngrid[k-1]);
	}       
        divh[i][j][k] += (hflz[i][j][k]-hflx[i][j][k-1])/(zgrid[k]-zgrid[k-1]);
	//	divh[i][j][k] /= ghalf;
      }

      // Do upper and lower boundaries
      divh[i][j][0] = 0;
      divh[i][j][nzg-1] = 0;
      if (i > 0) {
	divh[i][j][0] += (hflx[i][j][0]-hflx[i-1][j][0])/(xgrid[i]-xgrid[i-1]);
	divh[i][j][nzg-1] += (hflx[i][j][nzg-1]-hflx[i-1][j][nzg-1])
	  /(xgrid[i]-xgrid[i-1]);
      }
      if (j > 0) {
	divh[i][j][0] += (hfly[i][j][0]-hfly[i][j-1][0])/(ygrid[j]-ygrid[j-1]);
	divh[i][j][nzg-1] += (hfly[i][j][nzg-1]-hfly[i][j-1][nzg-1])
	  /(ygrid[j]-ygrid[j-1]);
      }
      divh[i][j][nzg-1] += (hflz[i][j][nzg-1]-hflx[i][j][nzg-2])
	/(zgrid[nzg-1]-zgrid[nzg-2]);
    }
  }

  // Calculate grad(div(H)) 
  // x component on u points etc
  for (int j=0; j<nyg; j++) {
    for (int i=0; i<nxg; i++) {
      if (i < nzg-1) {
	Cxu = -(oro[i+1][j]-oro[i][j])/(xngrid[i+1]-xngrid[i]);
      }
      else {
	Cxu = 0;
      }
      if (j < nyg-1) {
	Cyv = -(oro[i][j+1]-oro[i][j])/(yngrid[j+1]-yngrid[j]);
      }
      else {
	Cyv = 0;
      }
      ghalfx = (1-orx[i][j]/H);
      ghalfy = (1-ory[i][j]/H);
      ghalf = (1-oro[i][j]/H);
      delhx[i][j][0] = 0;
      delhy[i][j][0] = 0;
      delhz[i][j][0] = 0;
      delhx[i][j][nzg-1] = 0;
      delhy[i][j][nzg-1] = 0;
      delhz[i][j][nzg-1] = 0;
      for (int k=1; k<nzg-1; k++) {
	delhx[i][j][k] = 0;
	delhy[i][j][k] = 0;

	if (i < nxg-1) {
	  delhx[i][j][k] = (divh[i+1][j][k]-divh[i][j][k])
	    /(xngrid[i+1]-xngrid[i])
	    + Cxu * (1-zgrid[k]/H)/ghalfx
	    * 0.5*( (divh[i][j][k+1]+divh[i+1][j][k+1]) -
		    (divh[i][j][k-1]+divh[i+1][j][k-1]) ) 
	    / (zngrid[k+1]-zngrid[k-1]);
	}
	
	if (j < nyg - 1) {
	  delhy[i][j][k] = (divh[i][j+1][k]-divh[i][j][k])
	    /(yngrid[j+1]-yngrid[j])
	    + Cyv * (1-zgrid[k]/H)/ghalfy
	    * 0.5*( (divh[i][j][k+1]+divh[i][j+1][k+1]) -
		    (divh[i][j][k-1]+divh[i][j+1][k-1]) ) 
	    / (zngrid[k+1]-zngrid[k-1]);
	}

	delhz[i][j][k] = (divh[i][j][k+1]-divh[i][j][k])
	  / (zngrid[k+1]-zngrid[k]) / ghalf;
      }
    }
  }
  

  // Now calculate the source term interpolated onto p points.
  for (int k=1; k<nzg-1; k++) {
    for (int j=0; j<nyg; j++) {
      for (int i=0; i<nxg; i++) {
	i1 = i-1;
	j1 = j-1;
	if (i == 0) i1 = i;
	if (j == 0) j1 = j;
	// contribution from momentum equation
	src[i][j][k] = 0.125*(gradthx[i1][j][k]+gradthx[i][j][k])
	  *(curlfx[i][j1][k-1]+curlfx[i][j][k-1]+curlfx[i][j1][k]
	    +curlfx[i][j][k])
	  + 0.125*(gradthy[i][j1][k]+gradthy[i][j][k])
	  *(curlfy[i1][j][k-1]+curlfy[i][j][k-1]+curlfy[i1][j][k]
	    +curlfy[i][j][k])
	  + 0.125*(gradthz[i][j][k-1]+gradthz[i][j][k])
	  *(curlfz[i1][j1][k]+curlfz[i][j1][k]+curlfz[i1][j][k]
	    +curlfz[i][j][k]);

	// Add on contribution from theta equation.
 	src[i][j][k] += 0.125*(delhx[i1][j][k]+delhx[i][j][k])
 	  *(vortx[i][j1][k-1]+vortx[i][j][k-1]+vortx[i][j1][k]
 	    +vortx[i][j][k])
 	  + 0.125*(delhy[i][j1][k]+delhy[i][j][k])
 	  *(vorty[i1][j][k-1]+vorty[i][j][k-1]+vorty[i1][j][k]
 	    +vorty[i][j][k])
 	  + 0.125*(delhz[i][j][k-1]+delhz[i][j][k])
 	  *(vortz[i1][j1][k]+vortz[i][j1][k]+vortz[i1][j][k]
 	    +vortz[i][j][k] + 4.0*fcor);
		
	// Finally divide by rho
	src[i][j][k] /= rho[i][j][k];
      }
    }
  }

  PVsrc->SetGrids(xngrid,yngrid,zngrid,nxg,nyg,nzg,H,orog);
  PVsrc->SetName("PV source");
  
  
  Free3DArray(Fx);
  Free3DArray(Fy);
  Free3DArray(Fz);

  Free3DArray(curlfx);
  Free3DArray(curlfy);
  Free3DArray(curlfz);

  Free3DArray(divh);
  
  return PVsrc;
}

//
// Get the source term for PV
//
DataSet *BlasFile::GetLESuw() {
  
  DataSet *ds;
  bldata ***uwav, ***uav, ***wav, ***data;
  int ndim, dimx, dimy, dimz;
  bldata u_av, w_av;


  if (uw_res != NULL) 
    return uw_res;

  // Read in LES average fields

  ds = GetData("UAV");
  if (ds == NULL) {
    cerr << "Unable to read UAV" << endl;
    return NULL;
  }
  uav = ds->GetData();

  ds = GetData("WAV");
  if (ds == NULL) {
    cerr << "Unable to read WAV" << endl;
    return NULL;
  }
  wav = ds->GetData();

  ds = GetData("UWAV");
  if (ds == NULL) {
    cerr << "Unable to read UWAV" << endl;
    return NULL;
  }
  uwav = ds->GetData();

  ndim = ds->GetDims();
  dimx = ds->GetDim(1);
  dimy = ds->GetDim(2);
  dimz = ds->GetDim(3);

  // Create a new empty data set
  uw_res = new DataSet(ndim, dimx, dimy, dimz);  
  data = uw_res->GetData();

  for (int i=0; i<dimx; i++) {
    for (int j=0; j<dimy; j++) {
      for (int k=0; k<dimz; k++) {
	if (i < dimx-1) {
	  w_av = 0.5*(wav[i][j][k]+wav[i+1][j][k]);
	}
	else {
	  w_av = wav[i][j][k];
	}
	if (k < dimz-1) {
	  u_av = 0.5*(uav[i][j][k]+uav[i][j][k+1]);
	}
	else {
	  u_av = uav[i][j][k];
	}
	data[i][j][k] = uwav[i][j][k] - u_av*w_av;
      }
    }
  }

  if (dimx > 1) 
    uw_res->SetGrids(xgrid,nullgrid,zgrid, dimx, dimy, dimz, H, orog);
  else 
    uw_res->SetGrids(nullgrid,nullgrid,zgrid, dimx, dimy, dimz, H, orog);
  uw_res->SetName("u'w' resolved");
  
  return uw_res;

}

//
// Get the LES resolved part of u'u'
//
DataSet *BlasFile::GetLESuu() {
  
  DataSet *ds;
  bldata ***uuav, ***uav, ***data;
  int ndim, dimx, dimy, dimz;

  if (uu_res != NULL) 
    return uu_res;

  // Read in LES average fields

  ds = GetData("UAV");
  if (ds == NULL) {
    cerr << "Unable to read UAV" << endl;
    return NULL;
  }
  uav = ds->GetData();

  ds = GetData("U2AV");
  if (ds == NULL) {
    cerr << "Unable to read U2AV" << endl;
    return NULL;
  }
  uuav = ds->GetData();

  ndim = ds->GetDims();
  dimx = ds->GetDim(1);
  dimy = ds->GetDim(2);
  dimz = ds->GetDim(3);

  // Create a new empty data set
  uu_res = new DataSet(ndim, dimx, dimy, dimz);  
  data = uu_res->GetData();

  for (int i=0; i<dimx; i++) {
    for (int j=0; j<dimy; j++) {
      for (int k=0; k<dimz; k++) {
	data[i][j][k] = uuav[i][j][k] - uav[i][j][k]*uav[i][j][k];
      }
    }
  }

  if (dimx > 1) 
    uu_res->SetGrids(xgrid,nullgrid,zngrid, dimx, dimy, dimz, H, orog);
  else 
    uu_res->SetGrids(nullgrid,nullgrid,zngrid, dimx, dimy, dimz, H, orog);
  uu_res->SetName("u'u' resolved");
  
  return uu_res;

}
//
// Get the LES resolved part of v'v'
//
DataSet *BlasFile::GetLESvv() {
  
  DataSet *ds;
  bldata ***vvav, ***vav, ***data;
  int ndim, dimx, dimy, dimz;

  if (vv_res != NULL) 
    return vv_res;

  // Read in LES average fields

  ds = GetData("VAV");
  if (ds == NULL) {
    cerr << "Unable to read VAV" << endl;
    return NULL;
  }
  vav = ds->GetData();

  ds = GetData("V2AV");
  if (ds == NULL) {
    cerr << "Unable to read V2AV" << endl;
    return NULL;
  }
  vvav = ds->GetData();

  ndim = ds->GetDims();
  dimx = ds->GetDim(1);
  dimy = ds->GetDim(2);
  dimz = ds->GetDim(3);

  // Create a new empty data set
  vv_res = new DataSet(ndim, dimx, dimy, dimz);  
  data = vv_res->GetData();

  for (int i=0; i<dimx; i++) {
    for (int j=0; j<dimy; j++) {
      for (int k=0; k<dimz; k++) {
	data[i][j][k] = vvav[i][j][k] - vav[i][j][k]*vav[i][j][k];
      }
    }
  }

  if (dimx > 1) 
    vv_res->SetGrids(xngrid,nullgrid,zngrid, dimx, dimy, dimz, H, orog);
  else 
    vv_res->SetGrids(nullgrid,nullgrid,zngrid, dimx, dimy, dimz, H, orog);
  vv_res->SetName("v'v' resolved");
  
  return vv_res;

}

//
// Get the les resolved w'w'
//
DataSet *BlasFile::GetLESww() {
  
  DataSet *ds;
  bldata ***wwav, ***wav, ***data;
  int ndim, dimx, dimy, dimz;

  if (ww_res != NULL) 
    return ww_res;

  // Read in LES average fields

  ds = GetData("WAV");
  if (ds == NULL) {
    cerr << "Unable to read WAV" << endl;
    return NULL;
  }
  wav = ds->GetData();

  ds = GetData("W2AV");
  if (ds == NULL) {
    cerr << "Unable to read W2AV" << endl;
    return NULL;
  }
  wwav = ds->GetData();

  ndim = ds->GetDims();
  dimx = ds->GetDim(1);
  dimy = ds->GetDim(2);
  dimz = ds->GetDim(3);

  // Create a new empty data set
  ww_res = new DataSet(ndim, dimx, dimy, dimz);  
  data = ww_res->GetData();

  for (int i=0; i<dimx; i++) {
    for (int j=0; j<dimy; j++) {
      for (int k=0; k<dimz; k++) {
	data[i][j][k] = wwav[i][j][k] - wav[i][j][k]*wav[i][j][k];
      }
    }
  }

  if (dimx > 1) 
    ww_res->SetGrids(xngrid,nullgrid,zgrid, dimx, dimy, dimz, H, orog);
  else 
    ww_res->SetGrids(nullgrid,nullgrid,zgrid, dimx, dimy, dimz, H, orog);
  ww_res->SetName("w'w' resolved");
  
  return ww_res;

}

//
// Get the LES resolved tke
//
DataSet * BlasFile::GetLESenergy() {

  bldata ***uuav, ***vvav, ***wwav, ***data;
  int i1, i2, ndim, dimx, dimy, dimz;

  if (e_res != NULL) 
    return e_res;

  GetLESuu();
  GetLESvv();
  GetLESww();

  if (!uu_res || !vv_res || !ww_res) {
    cerr << "Unable to calculate resolved turbulence" << endl;
    return NULL;
  }


  uuav = uu_res->GetData();
  vvav = vv_res->GetData();
  wwav = ww_res->GetData();

  ndim = uu_res->GetDims();
  dimx = uu_res->GetDim(1);
  dimy = uu_res->GetDim(2);
  dimz = uu_res->GetDim(3);

  e_res = new DataSet(ndim, dimx, dimy, dimz);
  data = e_res->GetData();

  // We know dimy == 1 for LES
  // Check if dimx == 1
  if (dimx == 1) {
    for (int k = 0; k<dimz-1; k++) {
      data[0][0][k] = 0.5*( uuav[0][0][k]+vvav[0][0][k]+
			    0.5*(wwav[0][0][k]+wwav[0][0][k+1]) );
    }
    data[0][0][dimz-1] = 0;
  }
  else {
    for (int i=0;i<dimx; i++) {
      i1 = i;
      if (i < dimx-1) {
	i2 = i+1;
      }
      else {
	i2 = i;
      }
      for (int k = 0; k<dimz-1; k++) {
	data[i][0][k] = 0.5*(0.5*(uuav[i1][0][k]+uuav[i2][0][k])
			     + vvav[i][0][k]
			     + 0.5*(wwav[i][0][k]+wwav[i][0][k+1]));
      }
      data[i][0][dimz-1] = 0;
    } 
  }

  if (dimx > 1) 
    e_res->SetGrids(xngrid,nullgrid,zngrid, dimx, dimy, dimz, H, orog);
  else 
    e_res->SetGrids(nullgrid,nullgrid,zngrid, dimx, dimy, dimz, H, orog);
  e_res->SetName("Resolved tke");
  
  return e_res;


}

