// 
// blasvis++ - C++ version of graphics routines
//
// DataPlot : Class to plot a specified data set(s).
//            Handles various types of plot.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#include "dataplot.h"
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <fstream>

using namespace std;

//
// Default constructor for the dataplot class.
//
DataPlot::DataPlot() {
  pls = NULL;

#ifdef USE_PLPLOT
  dev = xwin;
#else
  dev = file;
#endif
  devfile[0] = '\0';
  dolegend = true;
  plothill = PLOTINTERSECT;
  transform = true;
  invert = false;
  logcontour = false;
  nix = 100;
  niy = 100;

  // Set for auto axes by default
  //  SetRange(0.0, 0.0, 0.0, 0.0);
  autoXRange = true;
  autoYRange = true;
  autoZRange = true;
  autoTRange = true;
  
  //  SetContours(0.0,0.0,20);
  autoContourRange = true;
  nlev = 20;
  
  // Set default titles
  SetTitles("%d. Run %r, time %ts.","%x","%y","%d");
  // Set % as the escape character
  escstr.SetEscChar('%');

  offset = 0.0;
  scale = 1.0;
  rscale = 0.0;
  linewidth = 2;
  linestyle = -1;
  vectorstyle = -1;
  vectorfill = false;
  vectorcentre = false;
  linecol = 1;
  textsize = 1.0;
  aspect = 0.0;

  SetColScheme(0);
  
  xscale = 1.0;
  yscale = 1.0;
  zscale = 1.0;
  dscale = 1.0;

  xlog = false;
  ylog = false;
  zlog = false;
  dlog = false;

  mesh = true;
  alt = 60.0;
  az = 30.0;
  theta = 0.0;

  contlab=false;

}

//
// Initialise plot stream
//
void DataPlot::InitPlot() {
#ifdef USE_PLPLOT
  int cs;

  // Define colours for cts colour scheme
  PLFLT pos[7] = {0.0, 0.001, 0.25, 0.5, 0.75, 0.999, 1.0};
  PLFLT red[7] = {1.0, 0.001, 0.25, 0.5, 1.00, 1.000, 0.0};
  PLFLT green[7] = {1.0, 0.998, 0.50, 0.5, 0.50, 0.998, 0.0};
  PLFLT blue[7] = {1.0, 1.000, 1.00, 0.5, 0.25, 0.001, 0.0};
  PLINT rev[7] = {0, 0, 0, 0, 0, 0, 0};

  // Define grayscale colour scheme
  PLFLT posgray[5] = {0.0, 0.499, 0.5, 0.501, 1.0};
  PLFLT redgray[5] = {0.0, 0.499, 0.5, 0.501, 1.0};
  PLFLT greengray[5] = {0.0, 0.499, 0.5, 0.501, 1.0};
  PLFLT bluegray[5] = {0.0, 0.499, 0.5, 0.501, 1.0};
  PLINT revgray[5] = {0,0,0,0,0};

  // Define colours for fixed colour set
  PLINT ired[20] = {255, 0, 255, 0, 0, 255, 255, 0, 153, 0, 
		    0, 153, 153, 0, 255, 153, 153, 153, 204, 102};
  PLINT igreen[20] = {255, 0, 0, 255, 0, 255, 0, 255, 0, 153, 
		      0, 153, 0, 153, 153, 255, 153, 153, 204, 102};
  PLINT iblue[20] = {255, 0, 0, 0, 255, 0, 255, 255, 0, 0, 153, 
		     0, 153, 153, 153, 153, 255, 153, 204, 102};
#endif

  NextPlot();

  // Set device type
  switch (dev) {
    // No setting up needed for data file
  case file:
    pls = NULL;
    fout.open(devfile);
    break;
#ifdef USE_PLPLOT
  case xwin:
    pls = new plstream(1,1,255,255,255,"xwin");
    break;
  case ps:
    // Use Adobe fonts for postscript drivers
    pls = new plstream(1,1,255,255,255,"ps", devfile);
    pls->SetOpt("drvopt","text");
    break;
  case psc:
    pls = new plstream(1,1,255,255,255,"psc", devfile);
    pls->SetOpt("drvopt","text");
    break;
  case xfig:
    pls = new plstream(1,1,255,255,255,"xfig", devfile);
    break;
  case pbm:
    pls = new plstream(1,1,255,255,255,"pbm", devfile);
    break;
  case png:
    pls = new plstream(1,1,255,255,255,"png", devfile);
    // Add Freetype support for png files
    pls->SetOpt("drvopt","text");
    break;
  case jpeg:
    pls = new plstream(1,1,255,255,255,"jpeg", devfile);
    // Add Freetype support for jpeg files
    pls->SetOpt("drvopt","text");
    break;
  default:
    pls = new plstream(1,1,255,255,255);
    break;
#else
  default:
    cerr << "Error - unknown device specified" << endl;
    pls = NULL;
    break;
#endif
  }

#ifdef USE_PLPLOT

  //  if (aspect > 0.0) {
  //    pls->sdidev(PL_NOTSET,aspect,PL_NOTSET,PL_NOTSET);    
  //  }
  
  if (pls != NULL) {
    pls->init();
  }
  // Don't need colours etc for file.
  if (dev != file) {
    // Set colour scheme
    cs = colscheme;
    
    // Default - if b&w postscript use greyscale, else use colour
    if (cs == 0) {
      if (dev == ps) {
	// BW
	cs = 1;
      }
      else {
	//Colour
	cs = 2;
      }
    }
    
    // Use colour or monochrome scheme
    if (cs%2 == 1)  {
      pls->scmap1l(1,5,posgray,redgray,greengray,bluegray,revgray);
    }
    else {
      pls->scmap1l(1,7,pos,red,green,blue,rev);
    }
    
    pls->scmap0(ired,igreen,iblue,20);

    pls->psty(0);
    pls->col0(1);
  }    
#endif
}

//
// Move onto the next pages of a plot.
//
void DataPlot::NextPlot() {
  numplots = 0;
  ncont = 0;
  numlines = 0;
  maxlines = 20;
}


//
// Finish the current plot
//
void DataPlot::FinishPlot() {
#ifdef USE_PLPLOT
  if (pls) {
    delete pls;
  }
#endif
  if (fout.is_open()) {
    fout.close();
  }

  pls = NULL;
}

//
// Draw a colour contour plot
//
void DataPlot::ColContPlot(DataSet *ds, int sdim, PLFLT sval, DataSet *temp) {
  int nx,ny;
  PLFLT ** data;
  PLFLT *clev;
  PLFLT *xgrid, *ygrid;
  PLFLT **xigrid, **yigrid, **zigrid, **higrid;
  PLFLT xmi, xma, ymi, yma, zmi, zma;
  PLFLT ldmin, ldmax;

  if (numplots == 0) 
    AutoRange(ds, sdim);

  xmi = xmin; 
  xma = xmax;
  ymi = ymin; 
  yma = ymax;
  zmi = zmin; 
  zma = zmax;

  switch(sdim) {
  case 1:
    xmi = sval; 
    xma = sval;
    break;
  case 2:
    ymi = sval;
    yma = sval;
    break;
  case 3:
  case 4:
  case 5:
    zmi = sval;
    zma = sval;
    break;
  }

  // Get interpolated section of data.
  nx = nix;
  ny = niy;
  ds->CreateInterpGrid(nx, ny, (PLFLT) xmi, (PLFLT) xma, (PLFLT) ymi, 
		       (PLFLT) yma, (PLFLT) zmi, (PLFLT) zma, 
		       sdim, &xgrid, &ygrid,
		       &xigrid, &yigrid, &zigrid, &higrid, 
		       xscale, yscale, zscale, theta, invert, transform,temp);
  ds->GetInterpSection(&data, xigrid, yigrid, zigrid, higrid, dscale, nx,ny, 
		       invert, transform);
  // Debugging
  //for (int j=0;j<ny;j++) {
  //  cout << xigrid[nx/2][j] << " " << zigrid[nx/2][j] << " " << higrid[nx/2][j] << " " << data[nx/2][j] <<  endl;
  //}

  if (data == NULL) {
    return;
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	fout << xigrid[i][j] << "\t" << yigrid[i][j] << "\t" << zigrid[i][j] 
	     << "\t" << data[i][j] << endl;
      }
    }
    return;
  }

  // Set up contour levels
  if (autoContourRange) {
    if (logcontour) {
      dmin = GetMinNonzero(data,nx,ny);
      dmax = GetMaxNonzero(data,nx,ny);
    }
    else {
      dmin = GetMin(data,nx,ny);
      dmax = GetMax(data,nx,ny);
    }
  }

  clev = new PLFLT[nlev];

  if (logcontour && (dmax*dmin > 0) ) {
    ldmax = log(fabs(dmax));
    ldmin = log(fabs(dmin));
    for (int i=0;i<nlev;i++) {
      clev[i] = copysign(exp(((PLFLT) i / (PLFLT) (nlev - 1))*(ldmax-ldmin) + ldmin),dmax);
    }
  }
  else {
    if (logcontour) {
      cerr << "Must have contour values all the same sign for logarithmic plot" << endl;
    }
    for (int i=0;i<nlev;i++) {
      clev[i] = ((PLFLT) i / (PLFLT) (nlev - 1))*(dmax-dmin) + dmin;
    }
  }

  // Check contours are ok before plotting
  if (dmax > dmin) {
    ColContPlot(data,nx,ny,xgrid,ygrid,clev,nlev);
    
    if (numplots == 1) {
      PlotTitles(ds->GetName(), sdim, sval);
   
      if (transform) 
	PlotHill(sdim, xigrid, yigrid, zigrid, higrid, xgrid, ygrid, 
		 nx, ny);
    }
  }
  else {
    cerr << "Contour range must be positive - " << dmin << " to " << dmax << endl;
  }
  
  delete[] clev;
  delete[] data[0];
  delete[] data;
  delete[] xgrid;
  delete[] ygrid;
  delete[] xigrid[0];
  delete[] xigrid;
  delete[] yigrid[0];
  delete[] yigrid;
  delete[] zigrid[0];
  delete[] zigrid;
  delete[] higrid[0];
  delete[] higrid;
}

//
// Draw a colour contour plot from a data array
//
void DataPlot::ColContPlot(PLFLT **data, int nx,int ny, PLFLT *xgrid, 
			   PLFLT *ygrid, PLFLT *clev, int nlev, 
			   bool datalog, bool xlogaxis, bool ylogaxis ) {

#ifdef USE_PLPLOT
  PLFLT tmp1,tmp2,wymin,wymax,dwy,chrhgt,xdiff,ydiff,colour,zdiff;
  PLFLT lxmin, lxmax, lymin, lymax;
  PLFLT xbox[4] = {0.0, 0.0, 1.0, 1.0};
  PLFLT ybox[4] = {0.0, 1.0, 1.0, 0.0};
  int ilab;
  char label[11];
  char xopts[10],yopts[10];

  if (pls == NULL) {
    cerr << "Must initialise plot first" << endl;
    return;
  }

  // If this is the first plot then initialise the page
  // and the axes etc.

  // Need these for the legend as well.
  lxmin = (xgrid[0]<xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lxmax = (xgrid[0]>xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lymin = (ygrid[0]<ygrid[ny-1])?ygrid[0]:ygrid[ny-1];
  lymax = (ygrid[0]>ygrid[ny-1])?ygrid[0]:ygrid[ny-1];
  xdiff = (lxmax-lxmin)*0.01;
  ydiff = (lymax-lymin)*0.01;
  if (numplots == 0) {
    pls->adv(0);
    
    if (dolegend) {
      pls->vpas(0.15,0.7,0.15,0.8,aspect);
      pls->gspa(tmp1,tmp2,wymin,wymax);
      //pls->gchr(tmp1,chrhgt);
      //chrhgt = chrhgt/(0.65*(wymax-wymin))*1.5;
      pls->wind(lxmin-xdiff,lxmax+xdiff,lymin-ydiff,lymax+ydiff);
      if (xlogaxis) {
	strncpy(xopts,"bcnstl",10);
      }
      else {
	strncpy(xopts,"bcnst",10);
      }
      if (ylogaxis) {
	strncpy(yopts, "bcnstvl", 10);
      }
      else {
	strncpy(yopts, "bcnstv",10);
      }
      pls->box(xopts,0.0,0,yopts,0.0,0);
    }
    else {
      // Put in log stuff
      pls->vpas(0.15,0.9,0.15,0.8,aspect);      
      pls->wind(lxmin-xdiff,lxmax+xdiff,lymin-ydiff,lymax+ydiff);
      pls->box("bcinst",0.0,0,"bcinstv",0.0,0);
      //pls->env(lxmin-xdiff,lxmax+xdiff,lymin-ydiff,lymax+ydiff,0,0);
    }
  }

  zdiff = (clev[nlev-1] - clev[0])*0.0001;

  // Check the data to avoid "holes".
  for (int i=0;i<nx;i++) {
    for (int j=0;j<ny;j++) {
      if (data[i][j] > clev[nlev-1])
	data[i][j] = clev[nlev-1] - zdiff;
      if (data[i][j] < clev[0])
	data[i][j] = clev[0] + zdiff;
      if (datalog)
	data[i][j] = log10(data[i][j]);      
    }
  }

  // Plot each set of colour contours.
  for (int i=0;i<nlev-1;i++) {
    colour = ((PLFLT ) i)/((PLFLT) (nlev-2));
    // Check if reversed colour scheme
    if (colreverse) {
      colour = 1.0 - colour;
    }
    // Check if zero is white    
    if (whitezero) {
      colour = (colour+0.1)/1.1;
      if (clev[i]*clev[i+1] <= 0.0) {
	colour = 0.0;
      }
    }
#ifdef PLSTREAM_TR_NOT_STATIC
    pls->shade((PLFLT **)data,nx,ny,NULL,lxmin,lxmax,lymin,lymax,clev[i],clev[i+1],1,
	       colour,2,0,0,0,0,plfill,1,NULL,NULL);
#else
    pls->shade((PLFLT **)data,nx,ny,NULL,lxmin,lxmax,lymin,lymax,clev[i],clev[i+1],1,
	       colour,2,0,0,0,0,plstream::fill,1,NULL,NULL);
#endif
  }
  
  pls->psty(0);

  // Plot the legend if required (only on first plot)
  if ((numplots == 0) && dolegend) {
    dwy = (0.8-0.15)/(nlev-1);
    wymin = 0.15;
    // This is a bit of a botch - I'd like something better
    ilab = nlev / 16 + 1;
    if (ilab == 3) ilab = 4;
    for (int i=0;i<nlev-1;i++) {
      colour = ((PLFLT) i)/(nlev-2);
      // Check if reversed colour scheme
      if (colreverse) {
	colour = 1.0 - colour;
      }
      // Check if white is zero
      if (whitezero) {
	colour = (colour+0.1)/1.1;
	if (clev[i]*clev[i+1] <= 0.0) {
	  colour = 0.0;
	}
      }
      pls->col(colour);
      pls->vpor(0.75,0.8,wymin,wymin+dwy);
      pls->wind(0.0,1.0,0.0,1.0);
      pls->fill(4,xbox,ybox);
      pls->col0(1);
      pls->box("bci",0.0,0,"bci",0.0,0);
      if (i%ilab == 0) {
	sprintf(label,"%#.3G",clev[i]);
	pls->mtex("rv",1.0,0.0,0.0,label);
      }
      wymin += dwy;
    }
    pls->col0(1);
    if ((nlev-1)%ilab == 0) {
      sprintf(label,"%#.3G",clev[nlev-1]);
      pls->mtex("rv",1.0,1.0,0.0,label);      
    }
    pls->vpas(0.15,0.7,0.15,0.8,aspect);
    pls->wind(lxmin-xdiff,lxmax+xdiff,lymin-ydiff,lymax+ydiff);
  }

  numplots++;
#endif
}

//
// Draw a line contour plot
//
void DataPlot::LineContPlot(DataSet *ds, int sdim, PLFLT sval, DataSet *temp) {
  int nx,ny;
  PLFLT **data;
  PLFLT *clev;
  PLFLT *xgrid, *ygrid;
  PLFLT **xigrid, **yigrid, **zigrid, **higrid;
  PLFLT xmi, xma, ymi, yma, zmi, zma;

  if (numplots == 0)
    AutoRange(ds,sdim);

  xmi = xmin; 
  xma = xmax;
  ymi = ymin; 
  yma = ymax;
  zmi = zmin; 
  zma = zmax;

  switch(sdim) {
  case 1:
    xmi = sval; 
    xma = sval;
    break;
  case 2:
    ymi = sval;
    yma = sval;
    break;
  case 3:
  case 4:
  case 5:
    zmi = sval;
    zma = sval;
  }

  // Get interpolated section of data.
  nx = nix;
  ny = niy;
  ds->CreateInterpGrid(nx, ny, (PLFLT) xmi, (PLFLT) xma, (PLFLT) ymi, 
		       (PLFLT) yma, (PLFLT) zmi, (PLFLT) zma, 
		       sdim, &xgrid, &ygrid,
		       &xigrid, &yigrid, &zigrid, &higrid,
		       xscale, yscale, zscale, theta, invert, transform,temp);
  ds->GetInterpSection(&data, xigrid, yigrid, zigrid, higrid, dscale, nx ,ny, 
		       invert, transform);

  if (data == NULL) {
    return;
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	fout << xigrid[i][j] << "\t" << yigrid[i][j]  << "\t" << zigrid[i][j] 
	     << "\t" << data[i][j] << endl;
      }
    }
    return;
  }

  // Set up contour levels
  if (autoContourRange) {
    dmin = GetMin(data,nx,ny);
    dmax = GetMax(data,nx,ny);
  }

  clev = new PLFLT[nlev];

  for (int i=0;i<nlev;i++) {
    clev[i] = ((PLFLT) i / (PLFLT) (nlev - 1))*(dmax-dmin) + dmin;
  }

  // Why was this here? 
  // To try and work round contour plotting ***bug in plplot***
  //for (int j=1;j<ny;j++) {
  //  data[0][j] = data[0][j-1];
  //}

  LineContPlot(data,nx,ny,xgrid,ygrid,clev,nlev);

  if (numplots == 1) {
    PlotTitles(ds->GetName(), sdim, sval);
    
    if (transform) {
      PlotHill(sdim, xigrid, yigrid, zigrid, higrid, xgrid, ygrid, 
	       nx, ny);
    }
  }

  delete[] clev;
  delete[] data[0];
  delete[] data;
  delete[] xgrid;
  delete[] ygrid;
  delete[] xigrid[0];
  delete[] xigrid;
  delete[] yigrid[0];
  delete[] yigrid;
  delete[] zigrid[0];
  delete[] zigrid;
  delete[] higrid[0];
  delete[] higrid;

}

//
// Draw a line contour plot from a data array
//
void DataPlot::LineContPlot(PLFLT **data,int nx,int ny,PLFLT *xgrid, PLFLT *ygrid, 
			    PLFLT *clev, int nlev) {

#ifdef USE_PLPLOT
  PLcGrid trgrid;
  PLFLT lxmin, lxmax, lymin, lymax;

  if (pls == NULL) {
    cerr << "Must initialise plot first" << endl;
    return;
  }

  lxmin = (xgrid[0]<xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lxmax = (xgrid[0]>xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lymin = (ygrid[0]<ygrid[ny-1])?ygrid[0]:ygrid[ny-1];
  lymax = (ygrid[0]>ygrid[ny-1])?ygrid[0]:ygrid[ny-1];

  trgrid.xg = xgrid;
  trgrid.yg = ygrid;
  trgrid.nx = nx;
  trgrid.ny = ny;

  pls->col0(1);

  // If this is the first plot then initialise the page
  // and the axes etc.
  if (numplots == 0) {
    pls->adv(0);
    pls->vpas(0.15,0.9,0.15,0.8,aspect);
    pls->wind(lxmin,lxmax,lymin,lymax);
    pls->box("bcinst",0.0,0,"bcinstv",0.0,0);
    //pls->env(lxmin,lxmax,lymin,lymax,0,0);
  }


  ncont++;
  //  if (ncont == -2) {
  //    int mark = 4000;
  //    int space = 8000;
  //    pls->styl(1,&mark,&space);
  //  }
  //  else {
  //cout << ncont << " " << linewidth << endl;
  if (linestyle == -1) {
    pls->lsty(ncont);
  }
  else {
    pls->lsty(linestyle);
  }
  //  }
  if (linecol == -1) {
    pls->col0(1);
  }
  else {
    pls->col0(linecol);
  }
  pls->wid(linewidth);
  if (contlab) {
    // pls->setcontlabelparam(0.005,0.9,0.5,1);
    pls->setcontlabelparam(0.015,1.0,0.6,1);
    pls->setcontlabelformat(4,4);
  }
  else {
    pls->setcontlabelparam(0.0006,0.3,0.1,0);
  }
#ifdef PLSTREAM_TR_NOT_STATIC
  pls->cont((PLFLT **)data,nx,ny,1,nx,1,ny,(PLFLT *)clev,nlev,pltr1,(void *) &trgrid);
#else
  pls->cont((PLFLT **)data,nx,ny,1,nx,1,ny,(PLFLT *)clev,nlev,plstream::tr1,(void *) &trgrid);
#endif
  
  pls->lsty(1);
  pls->wid(3);

  numplots++;
#endif
}

//
// Draw a vector plot
//
void DataPlot::VectorPlot(DataSet *ds1, DataSet *ds2, int sdim, PLFLT sval, DataSet *temp) {
  int nx,ny;
  PLFLT **data1, **data2;
  PLFLT d1, d2, costh, sinth;
  PLFLT *xgrid, *ygrid;
  PLFLT **xigrid, **yigrid, **zigrid, **higrid;
  PLFLT xmi, xma, ymi, yma, zmi, zma;

  if (numplots == 0)
    AutoRange(ds1,sdim);

  xmi = xmin; 
  xma = xmax;
  ymi = ymin; 
  yma = ymax;
  zmi = zmin; 
  zma = zmax;

  switch(sdim) {
  case 1:
    xmi = sval; 
    xma = sval;
    break;
  case 2:
    ymi = sval;
    yma = sval;
    break;
  case 3:
  case 4:
  case 5:
    zmi = sval;
    zma = sval;
  }

  // Get interpolated section of data.
  nx = nix;
  ny = niy;
  ds1->CreateInterpGrid(nx, ny, (PLFLT) xmi, (PLFLT) xma, (PLFLT) ymi, 
		       (PLFLT) yma, (PLFLT) zmi, (PLFLT) zma, 
			sdim, &xgrid, &ygrid,
		       &xigrid, &yigrid, &zigrid, &higrid,
		       xscale, yscale, zscale, theta, invert, transform,temp);
  ds1->GetInterpSection(&data1, xigrid, yigrid, zigrid, higrid, dscale, nx,ny,
			invert, transform);

  if (data1 == NULL) {
    return;
  }

  ds2->GetInterpSection(&data2, xigrid, yigrid, zigrid, higrid, dscale, nx,ny,
			invert, transform);

  if (data2 == NULL) {
    delete[] data1[0];
    delete[] data1;
    return;
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	fout << xigrid[i][j] << "\t" << yigrid[i][j] << "\t" << zigrid[i][j] 
	     << "\t" << data1[i][j]  << "\t" << data2[i][j] << endl;
      }
    }
    return;
  }
  
  // Rotate vectors if necessary 
  if (theta != 0.0) {
    costh = cos(theta/180.0*M_PI);
    sinth = sin(theta/180.0*M_PI);
    for (int i=0;i<nx;i++) {
      for (int j=0;j<ny;j++) {
	d1 = data1[i][j];
	d2 = data2[i][j];
	data1[i][j] = costh*d1-sinth*d2;
	data2[i][j] = costh*d2+sinth*d1;
      }
    }
  }

  VectorPlot(data1,data2,nx,ny,xgrid,ygrid);

  if (numplots == 1) {
    PlotTitles(ds1->GetName(), sdim, sval);

    PlotHill(sdim, xigrid, yigrid, zigrid, higrid, xgrid, ygrid, 
	     nx, ny);
  }
  
  delete[] data1[0];
  delete[] data1;
  delete[] data2[0];
  delete[] data2;
  delete[] xgrid;
  delete[] ygrid;
  delete[] xigrid[0];
  delete[] xigrid;
  delete[] yigrid[0];
  delete[] yigrid;
  delete[] zigrid[0];
  delete[] zigrid;
  delete[] higrid[0];
  delete[] higrid;
}

//
// Draw a vector plot
//
void DataPlot::VectorPlot(PLFLT v1, PLFLT v2, PLFLT x, PLFLT y) {
#ifdef USE_PLPLOT
  PLFLT **d1, **d2, costh, sinth;
  PLcGrid2 cgrid2;

  if (numplots == 0)
      cerr << "Cannot start a plot with single vectors" << endl; 

  d1 = AllocArray<PLFLT>(1,1);
  d2 = AllocArray<PLFLT>(1,1);

  d1[0][0] = v1;
  d2[0][0] = v2;


  // Rotate vectors if necessary 
  if (theta != 0.0) {
    costh = cos(theta/180.0*M_PI);
    sinth = sin(theta/180.0*M_PI);
    d1[0][0] = costh*v1-sinth*v2;
    d2[0][0] = costh*v2+sinth*v1;
  }

  if (pls == NULL) {
    cerr << "Must initialise plot first" << endl;
    return;
  }

  if (linestyle == -1) {
    pls->lsty(1);
  }
  else {
    pls->lsty(linestyle);
  }
  pls->wid(linewidth);
  if (linecol == -1) {    
    pls->col0(1);
  }
  else {
    pls->col0(linecol);
  }

  cgrid2.xg = AllocArray<PLFLT>(1,1);
  cgrid2.yg = AllocArray<PLFLT>(1,1);

  cgrid2.nx = 1;
  cgrid2.ny = 1;

  cgrid2.xg[0][0] = x;
  cgrid2.yg[0][0] = y;

#ifdef MISSING_PLSTREAM_VECT
  cerr << "Vector plot support is missing from this version of plplot" << endl;
#else
  SetVectorStyle();
#ifdef PLSTREAM_TR_NOT_STATIC
  pls->vect(d1,d2,1,1,rscale,pltr2,(void *)&cgrid2);
#else
  pls->vect(d1,d2,1,1,rscale,plstream::tr2,(void *)&cgrid2);
#endif
#endif
  
  pls->lsty(1);
  pls->wid(3);

  FreeArray<PLFLT>(cgrid2.xg);
  FreeArray<PLFLT>(cgrid2.yg);

  FreeArray<PLFLT>(d1);
  FreeArray<PLFLT>(d2);
#endif
}

//
// Draw a line contour plot from a data array
//
void DataPlot::VectorPlot(PLFLT **data1,PLFLT **data2,int nx,int ny,
			  PLFLT *xgrid,PLFLT *ygrid) {

#ifdef USE_PLPLOT
  PLFLT lxmin, lxmax, lymin, lymax;
  PLcGrid2 cgrid2;

  if (pls == NULL) {
    cerr << "Must initialise plot first" << endl;
    return;
  }

  pls->col0(1);

  lxmin = (xgrid[0]<xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lxmax = (xgrid[0]>xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lymin = (ygrid[0]<ygrid[ny-1])?ygrid[0]:ygrid[ny-1];
  lymax = (ygrid[0]>ygrid[ny-1])?ygrid[0]:ygrid[ny-1];


  // If this is the first plot then initialise the page
  // and the axes etc.
  if (numplots == 0) {
    //    pls->adv(0);
    
    pls->env(lxmin,lxmax,lymin,lymax,0,0);
    //    pls->wind(xmin,xmax,ymin,ymax);
  }

  if (linestyle == -1) {
    pls->lsty(1);
  }
  else {
    pls->lsty(linestyle);
  }
  pls->wid(linewidth);
  if (linecol == -1) {    
    pls->col0(1);
  }
  else {
    pls->col0(linecol);
  }

  cgrid2.xg = AllocArray<PLFLT>(nx,ny);
  cgrid2.yg = AllocArray<PLFLT>(nx,ny);

  cgrid2.nx = nx;
  cgrid2.ny = ny;

  for (int i=0;i<nx;i++) {
    for (int j=0;j<ny;j++) {
      cgrid2.xg[i][j] = xgrid[i];
      cgrid2.yg[i][j] = ygrid[j];
    }
  }

#ifdef MISSING_PLSTREAM_VECT
  cerr << "Vector plot support is missing from this version of plplot" << endl;
#else
  SetVectorStyle();
#ifdef PLSTREAM_TR_NOT_STATIC
  pls->vect(data1,data2,nx,ny,rscale,pltr2,(void *)&cgrid2);
#else
  pls->vect(data1,data2,nx,ny,rscale,plstream::tr2,(void *)&cgrid2);
#endif
#endif

  pls->lsty(1);
  pls->wid(3);

  numplots++;

  FreeArray<PLFLT>(cgrid2.xg);
  FreeArray<PLFLT>(cgrid2.yg);
#endif
}

//
// Draw a surface plot
//
void DataPlot::SurfacePlot(DataSet *ds, int sdim, PLFLT sval, DataSet *temp) {
  int nx,ny;
  PLFLT ** data;
  PLFLT *xgrid, *ygrid;
  PLFLT **xigrid, **yigrid, **zigrid, **higrid;
  PLFLT xmi, xma, ymi, yma, zmi, zma;

  if (numplots == 0)
    AutoRange(ds,sdim);

  xmi = xmin; 
  xma = xmax;
  ymi = ymin; 
  yma = ymax;
  zmi = zmin; 
  zma = zmax;

  switch(sdim) {
  case 1:
    xmi = sval; 
    xma = sval;
    break;
  case 2:
    ymi = sval;
    yma = sval;
    break;
  case 3:
  case 4:
  case 5:
    zmi = sval;
    zma = sval;
  }

  // Get interpolated section of data.
  nx = nix;
  ny = niy;
  ds->CreateInterpGrid(nx, ny, (PLFLT) xmi, (PLFLT) xma, (PLFLT) ymi, 
		       (PLFLT) yma, (PLFLT) zmi, (PLFLT) zma, 
			sdim, &xgrid, &ygrid,
		       &xigrid, &yigrid, &zigrid, &higrid,
		       xscale, yscale, zscale, theta, invert, transform,temp);
  ds->GetInterpSection(&data, xigrid, yigrid, zigrid, higrid, dscale, nx,ny, 
		       invert, transform);

  if (data == NULL) {
    return;
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	fout << xigrid[i][j] << "\t" << yigrid[i][j] << "\t" << zigrid[i][j] 
	     << "\t" << data[i][j] << endl;
      }
    }
    return;
  }

  // Set up Z range
  if (autoContourRange && (numplots == 0) ) {
    dmin = GetMin(data,nx,ny);
    dmax = GetMax(data,nx,ny);
  }

  SurfacePlot(data,nx,ny,xgrid,ygrid);
  if (numplots == 1) {
    //    PlotTitles3D(ds->GetName(), sdim, sval);
  }
  
  delete[] data[0];
  delete[] data;
  delete[] xgrid;
  delete[] ygrid;
  delete[] xigrid[0];
  delete[] xigrid;
  delete[] yigrid[0];
  delete[] yigrid;
  delete[] zigrid[0];
  delete[] zigrid;
  delete[] higrid[0];
  delete[] higrid;

}

//
// Draw a surface plot from a data array
//
void DataPlot::SurfacePlot(PLFLT **data, int nx,int ny, PLFLT *xgrid, 
			   PLFLT *ygrid) {
#ifdef USE_PLPLOT
  PLFLT lxmin, lxmax, lymin, lymax, ldmin, ldmax;

  if (pls == NULL) {
    cerr << "Must initialise plot first" << endl;
    return;
  }

  lxmin = (xgrid[0]<xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lxmax = (xgrid[0]>xgrid[nx-1])?xgrid[0]:xgrid[nx-1];
  lymin = (ygrid[0]<ygrid[ny-1])?ygrid[0]:ygrid[ny-1];
  lymax = (ygrid[0]>ygrid[ny-1])?ygrid[0]:ygrid[ny-1];
  ldmax = dmax;
  ldmin = dmin;


  // If this is the first plot then initialise the page
  // and the axes etc.
  if (numplots == 0) {
    pls->adv(0);
    pls->col0(1);
    pls->vpas(0.0,1.0,0.0,0.9,aspect);
    pls->wind(-1.0, 1.0, -0.9, 1.1);
    pls->env(-1.0,1.0,-0.9,1.1,1,-2);
    pls->w3d(1.0, 1.0, 1.0, lxmin, lxmax, lymin, lymax, ldmin, ldmax, alt, az);
    //    pls->env(-2.0,2.0,-1.9,2.1,1,-2);
    //    pls->w3d(4.0, 1.67, 0.5, lxmin, lxmax, lymin, lymax, lzmin, lzmax, alt, az);

  }
  if (mesh) {
    pls->mesh(xgrid, ygrid, data, nx, ny, 3);
  }
  else {
    pls->plot3d(xgrid, ygrid, data, nx, ny, 3, 0);
    //pls->plotsh3d(xgrid, ygrid, data, nx, ny, 1);
  }
  
  numplots++;
#endif
}


//
// Plot a 1-D curve
//
void DataPlot::LinePlot(DataSet * ds, int sdim, PLFLT sval1, PLFLT sval2, 
			char *label, DataSet *temp) {

  PLFLT *data, *grid;
  PLFLT *xigrid, *yigrid, *zigrid, *higrid;
  DataSet *extra = NULL;
  int nx, col;

  PLFLT xmi, xma, ymi, yma, zmi, zma;

  if (numplots == 0) 
    AutoRange(ds, sdim);

  xmi = xmin; 
  xma = xmax;
  ymi = ymin; 
  yma = ymax;
  zmi = zmin; 
  zma = zmax;

  switch (sdim%10) {
  case 5:
    extra = temp;
    break;
  case 6:
    cerr << "Integrating along streamfunctions not yet implemented" << endl;
    return;
  }

  switch(sdim) {
  case 1:
  case 14:
  case 15:
  case 16:
    ymi = sval1;
    yma = sval1;
    zmi = sval2;
    zma = sval2;
    break;
  case 2:
  case 24:
  case 25:
  case 26:
    xmi = sval1;
    xma = sval1;
    zmi = sval2;
    zma = sval2;
    break;
  case 3:
    xmi = sval1;
    xma = sval1;
    ymi = sval2;
    yma = sval2;
    break;
  }

  // Get interpolated section of data.
  nx = nix;
  ds->CreateInterpGrid1D(nx, (PLFLT) xmi, (PLFLT) xma, (PLFLT) ymi, 
		       (PLFLT) yma, (PLFLT) zmi, (PLFLT) zma, 
		       sdim%10, &grid,
		       &xigrid, &yigrid, &zigrid, &higrid, 
		       xscale, yscale, zscale, theta, invert, transform,extra);
  ds->GetInterpLine(&data, xigrid, yigrid, zigrid, higrid, dscale, nx,
		       invert, transform);

  if (data == NULL) {
    return;
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int i=0;i<nx;i++) {
      fout << xigrid[i] << "\t" << yigrid[i] << "\t" << zigrid[i] << "\t" 
	   << data[i] << endl;    
    }
    return;
  }

  if (numplots == 0) {

    xmi = grid[0];
    xma = grid[0];
    for (int i=1;i<nx;i++) {
      xmi = (grid[i]<xmi)?grid[i]:xmi;
      xma = (grid[i]>xma)?grid[i]:xma;
    }
    
    if (autoContourRange) {
      ymi = data[0];
      yma = data[0];
      for (int i=1;i<nx;i++) {
	ymi = (data[i]<ymi)?data[i]:ymi;
	yma = (data[i]>yma)?data[i]:yma;
      }
    }
    else {
      ymi = dmin;
      yma = dmax;
    }
    
  }

  if (linecol == -1) {
    col = numlines+1;
  }
  else {
    col = linecol;
  }

  if (sdim == 3) {
    LinePlot(grid,nx,data, col, ymi, yma, xmi, xma);
    //    for (int i=0;i<nx;i++) {
    //      cout << grid[i] << " " << data[i] << endl;
    //    }
  }
  else {
    LinePlot(data,nx,grid, col, xmi, xma, ymi, yma);
  }
  if (numplots == 1)
    PlotTitles(ds->GetName(), 0, 0.0);

  LinePlotLegend(label, numlines, col);

  delete[] data;
  delete[] grid;
}


//
// Plot a 1-D time series
//
void DataPlot::LinePlot(TimeSeries * ds, char *label) {

  PLFLT *data, *time;
  int nx, col;

  PLFLT dmi, dma;

  if ((numplots == 0) && (autoTRange)) {
    tmin = ds->GetMinTime();
    tmax = ds->GetMaxTime();
  }

  // Get interpolated section of data.
  nx = ds->GetNumPoints();
  time = new bldata[nx];
  data = new bldata[nx];

  if ((time == NULL) || (data == NULL)) {
    return;
  }

  for (int i=0;i<nx;i++) {
    time[i] = tmin + (tmax-tmin)*((bldata)i)/((bldata)(nx-1));
    data[i] = ds->GetVal(time[i]);
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int i=0;i<nx;i++) {
      fout << time[i] << "\t" << data[i] << endl;    
    }
    return;
  }

  if (numplots == 0) {

    if (autoContourRange) {
      dmi = data[0];
      dma = data[0];
      for (int i=1;i<nx;i++) {
	dmi = (data[i]<dmi)?data[i]:dmi;
	dma = (data[i]>dma)?data[i]:dma;
      }
    }
    else {
      dmi = dmin;
      dma = dmax;
    }
    
  }

  if (linecol == -1) {
    col = numlines+1;
  }
  else {
    col = linecol;
  }

  LinePlot(data,nx,time, col, tmin, tmax, dmi, dma);
  if (numplots == 1)
    PlotTitles(ds->GetName(), 0, 0.0);

  LinePlotLegend(label, numlines, col);

  delete[] data;
  delete[] time;
}

//
// Plot a 1-D curve along a given trajectory
//
void DataPlot::LinePlot(DataSet * ds, int npts, PLFLT *xgrid, PLFLT *ygrid, 
			PLFLT *zgrid, PLFLT *hgrid, PLFLT *grid, 
			char *label) {

  PLFLT *data;

  PLFLT xmi, xma, ymi, yma;

  int col;

  //  cout << "Interpolation started..." << endl;

  // Get interpolated section of data.
  ds->GetInterpLine(&data, xgrid, ygrid, zgrid, hgrid, dscale, npts,
		    invert, transform);

  //  cout << "Interpolation done!" << endl;

  if (data == NULL) {
    return;
  }

  // If saving data to file then do so and return
  if (fout.is_open()) {
    for (int i=0;i<npts;i++) {
      fout << grid[i] << "\t" << xgrid[i] << "\t" << ygrid[i] << "\t" << 
	zgrid[i] << "\t"  << hgrid[i] << "\t" << data[i] << endl;
    }  
    return;
  }

  if (numplots == 0) {

    xmi = grid[0];
    xma = grid[0];
    for (int i=1;i<npts;i++) {
      xmi = (grid[i]<xmi)?grid[i]:xmi;
      xma = (grid[i]>xma)?grid[i]:xma;
    }
    
    if (autoContourRange) {
      ymi = data[0];
      yma = data[0];
      for (int i=1;i<npts;i++) {
	ymi = (data[i]<ymi)?data[i]:ymi;
	yma = (data[i]>yma)?data[i]:yma;
      }
    }
    else {
      ymi = dmin;
      yma = dmax;
    }
    
  }

  if (linecol = -1) {
    col = numlines+1;
  }
  else {
    col = linecol;
  }
  LinePlot(data,npts,grid, col, xmi, xma, ymi, yma);

  cout << "Plot done" << endl;

  if (numplots == 1)
    PlotTitles(ds->GetName(), 0, 0.0);

  LinePlotLegend(label, numlines, col);

  delete[] data;
}

void DataPlot::LinePlot(PLFLT *data, int npts, PLFLT *grid, int colour, PLFLT lxmin, PLFLT lxmax, PLFLT lymin, PLFLT lymax) {

#ifdef USE_PLPLOT
  if (pls == NULL) {
    cerr << "Must initialise plot first" << endl;
    return;
  }

  if (lxmin >= lxmax) {
    lxmin = GetMin(grid,npts);
    lxmax = GetMax(grid,npts);
  }
  
  if (lymin >= lymax) {
    lymin = GetMin(data,npts);
    lymax = GetMax(data,npts);
  }

  // If this is the first plot then initialise the page
  // and the axes etc.
  if (numplots == 0) {
    pls->adv(0);
    if (dolegend) {
      pls->vpas(0.15,0.7,0.15,0.8,aspect);
      pls->mtex("rv",1.0,0.954,0.0,"Legend");
      //    pls->env(xmin,xmax,ymin,ymax,0,0);
    }
    else {
      pls->vpas(0.15,0.9,0.15,0.9,aspect);
    }
    pls->wind(lxmin,lxmax,lymin,lymax);
    pls->box("bcnsti",0.0,0,"bcnstvi",0.0,0);
    //    pls->lab(xtitle,ytitle,title);

  }

  // Set line colour, style and width
  pls->col0(colour);
  if (linestyle == -1) {
    pls->lsty(1);
  }
  else {
    pls->lsty(linestyle);
  }
  pls->wid(linewidth);

  pls->line(npts,grid,data);

  pls->col0(1);
  pls->wid(3);

  numplots++;
  numlines++;
#endif
}

//
// Plot a hodograph from a pair of data sets
//
void DataPlot::Hodograph(DataSet *ds1, DataSet *ds2, int sdim, PLFLT sval1, 
			 PLFLT sval2, char *label, DataSet *temp) {
  int nx, col;
  PLFLT *data1, *data2;

  // For this case use dmin and dmax to get the range of values.
  if (autoContourRange && (numplots == 0)) {
    dmin = ds1->GetGridMin(sdim);
    dmax = ds1->GetGridMax(sdim);
  }

  nx = nix;
  
  ds1->GetInterpLine_old(&data1, sdim, sval1, sval2, nx, dmin, dmax,
		    0, scale, transform, temp);
  ds2->GetInterpLine_old(&data2, sdim, sval1, sval2, nx, dmin, dmax,
		    0, scale, transform, temp);

  if (numplots == 0) {
    if (autoXRange) {
      xmin = data1[0];
      xmax = data1[0];
      for (int i=1;i<nx;i++) {
	xmin = (data1[i]<xmin)?data1[i]:xmin;
	xmax = (data1[i]>xmax)?data1[i]:xmax;
      }
    }
    if (autoYRange) {
      ymin = data2[0];
      ymax = data2[0];
      for (int i=1;i<nx;i++) {
	ymin = (data2[i]<ymin)?data2[i]:ymin;
	ymax = (data2[i]>ymax)?data2[i]:ymax;
      }
    }
  }
  
  if (linecol == -1) {
    col = numlines+1;
  }
  else {
    col = linecol;
  }
  LinePlot(data2, nx, data1, col, xmin, xmax, ymin, ymax);
  if (numplots == 1)
    PlotTitles(ds1->GetName(), 0, 0.0);

  LinePlotLegend(label, numlines, col);

  delete[] data1;
  delete[] data2;

}

//
// Plot a hodograph from a pair of time series
//
void DataPlot::Hodograph(TimeSeries *ds1, TimeSeries *ds2, char *label) {
  int nx, col;
  PLFLT *data1, *data2, time;

  // For this case use dmin and dmax to get the range of values.
  if (autoContourRange && (numplots == 0)) {
    dmin = ds1->GetMinTime();
    dmax = ds1->GetMaxTime();
  }

  nx = ds1->GetNumPoints();
  data1 = new bldata[nx];
  data2 = new bldata[nx];

  for (int i=0;i<nx;i++) {
    time = ds1->GetTime(i);
    data1[i] = ds1->GetVal(i);
    data2[i] = ds2->GetVal(time);
  }
  
  if (numplots == 0) {
    if (autoXRange) {
      xmin = data1[0];
      xmax = data1[0];
      for (int i=1;i<nx;i++) {
	xmin = (data1[i]<xmin)?data1[i]:xmin;
	xmax = (data1[i]>xmax)?data1[i]:xmax;
      }
    }
    if (autoYRange) {
      ymin = data2[0];
      ymax = data2[0];
      for (int i=1;i<nx;i++) {
	ymin = (data2[i]<ymin)?data2[i]:ymin;
	ymax = (data2[i]>ymax)?data2[i]:ymax;
      }
    }
  }
  
  if (linecol == -1) {
    col = numlines+1;
  }
  else {
    col = linecol;
  }
  LinePlot(data2, nx, data1, col, xmin, xmax, ymin, ymax);
  if (numplots == 1)
    PlotTitles(ds1->GetName(), 0, 0.0);

  LinePlotLegend(label, numlines, col);

  delete[] data1;
  delete[] data2;

}

void DataPlot::LinePlotLegend(char *label, int pos, int colour ) {

#ifdef USE_PLPLOT
  PLFLT tmp1, tmp2, wymin, wymax, def, sca, chrhgt;

  pls->col0(colour);
  pls->wid(1);

  // Plot legend entry if necessary
  if (dolegend) {
    pls->gspa(tmp1,tmp2,wymin,wymax);
    pls->gchr(def,sca);
    pls->schr(0,textsize);
    pls->gchr(tmp1,chrhgt);
    chrhgt *= 1.5/(0.65*(wymax-wymin));
    pls->col0(1);
    pls->mtex("rv",1.0,0.95-pos*chrhgt,0.0,label);
    pls->col0(colour);
    pls->wid(2);
    pls->mtex("rv",10.0,0.95-pos*chrhgt,0.0," -\0");
    pls->schr(def,sca);
  }

  pls->col0(1);
  pls->wid(3);
#endif
}


// 
// Plot the orography on the current axes
//
void DataPlot::PlotHill1D(PLFLT *orog, PLFLT *xgrid, int nx, PLFLT offset) {
  
#ifdef USE_PLPLOT
  PLFLT *data, *x2, *o2;
  PLFLT xmi, xma, ymi, yma, diff;
  
  pls->lsty(1);
  pls->col0(1);

  if (offset != 0.0) {
    data = new PLFLT[nx];
    for (int i=0;i<nx;i++) {
      data[i] = orog[i] + offset;
    }
    pls->line(nx,xgrid,data);
    delete[] data;
  }
  else {
    // Fill in area under hill to blank out rogue contours 
    // *** gvpw() not yet in mainstream C++ bindings
    pls->gvpw(xmi,xma,ymi,yma);
    //plgvpw(&xmi,&xma,&ymi,&yma);
    diff = (yma-ymi)*0.01/1.02;
    x2 = new PLFLT[nx+2];
    o2 = new PLFLT[nx+2];
    for (int i=0;i<nx;i++) {
      x2[i+1] = xgrid[i];
      o2[i+1] = orog[i];
    }
    x2[0] = x2[1];
    o2[0] = ymi+diff;
    x2[nx+1] = x2[nx];
    o2[nx+1] = ymi+diff;
    pls->col0(0);
    pls->fill(nx+2,x2,o2);
    // Now draw the hill
    pls->col0(1);
    pls->line(nx,xgrid,orog);
    delete[] x2;
    delete[] o2;
  }
#endif
}


// 
// Plot the orography on the current axes
//
void DataPlot::PlotHill2D(PLFLT sval, PLFLT**orog, PLFLT *xgrid, 
			  PLFLT *ygrid, int nx, int ny) {
#ifdef USE_PLPLOT
  PLFLT clev;
  PLcGrid trgrid;

  trgrid.xg = xgrid;
  trgrid.yg = ygrid;
  trgrid.nx = nx;
  trgrid.ny = ny;

  pls->col0(1);
  pls->lsty(1);
  clev = sval;
  pls->setcontlabelparam(0.0006,0.3,0.1,0);
#ifdef PLSTREAM_TR_NOT_STATIC
  pls->cont((PLFLT **)orog,nx,ny,1,nx,1,ny,&clev,1, pltr1, (void *) &trgrid);
#else
  pls->cont((PLFLT **)orog,nx,ny,1,nx,1,ny,&clev,1, plstream::tr1, (void *) &trgrid);
#endif
#endif
}

//
// Set range for plot using specified values
//

void DataPlot::SetRange(PLFLT xmi,PLFLT xma,PLFLT ymi,PLFLT yma,
			PLFLT zmi,PLFLT zma) {
  xmin = xmi;
  xmax = xma;
  ymin = ymi;
  ymax = yma;
  zmin = zmi;
  zmax = zma;

  if (xmin == xmax) {
    autoXRange = true;   
  }
  else {
    autoXRange = false;   
  }

  if (ymin == ymax) {
    autoYRange = true;   
  }
  else {
    autoYRange = false;   
  }
  if (zmin == zmax) {
    autoZRange = true;   
  }
  else {
    autoZRange = false;   
  }
}
 

//
// Set range for contour / surface plots using specified values
//

void DataPlot::SetContours(PLFLT dmi,PLFLT dma,int l) {
  dmin = dmi;
  dmax = dma;
  if (dmin == dmax) {
    autoContourRange = true;
  }
  else {
    autoContourRange = false;
  }
  nlev = l;
} 

//
// Automatically set data range using dataset
//
void DataPlot::AutoRange(DataSet *ds, int sdim) {
    
  if (autoXRange) {
    xmin = ds->GetGridMin(1);
    xmax = ds->GetGridMax(1);
  }

  if (autoYRange) {
    ymin = ds->GetGridMin(2);
    ymax = ds->GetGridMax(2);
    //    cout << ymin << " " << ymax;
  }

  if (autoZRange) {
    zmin = ds->GetGridMin(3);
    zmax = ds->GetGridMax(3);
    //    cout << zmin << " " << zmax;
  }
}

//
// Set the titles for the plot
//
void DataPlot::SetTitles(char *t, char *xt, char *yt, char *zt) {
  strncpy(title,t,80);
  strncpy(xtitle,xt,40);
  strncpy(ytitle,yt,40);  
  strncpy(ztitle,zt,40);  
}

//
// Set the run number and time to plot on graphs
//
void DataPlot::SetRunParms(int r, int t) {
  char temp[20];

  // Set run number
  sprintf(temp,"%d",r);
  escstr.SetEscKey('r',temp);

  // Set (int) time
  sprintf(temp,"%d",t);
  escstr.SetEscKey('t',temp);
}

//
// Plot the hill on the current axes
//
void DataPlot::SetPlotHill(int ph) {
  plothill = ph;
}

//
// Set the plotting device
//
void DataPlot::SetDevice(devtype d) {
#ifndef USE_PLPLOT
  if (dev != file) {
    cerr << "Not a valid device" << endl;
    return;
  }
#endif
  dev = d;
  
}

//
// See if an output file is needed
//
bool DataPlot::NeedDevFile() {
  if (dev == xwin) {
    return false;
  }

  return true;
}

//
// Set the name of the device output file
// 
void DataPlot::SetDevFile(const char *f) {
  strncpy(devfile,f,80);
}

//
// Plot a legend
//
void DataPlot::SetPlotLegend(bool dl) {
  dolegend = dl;
}

//
// Get if plotting a legend
//
bool DataPlot::GetPlotLegend() {
  return dolegend;
}

//
// Transform data in the vertical
//
void DataPlot::SetTransform(bool t) {
  transform = t;
}

//
// Set the size of the interpolation grid
//
void DataPlot::SetInterpolation(int ix, int iy) {
  nix = ix;
  niy = iy;
}

//
// Set the offset for plotting lines
//
void DataPlot::SetLineOffset(PLFLT o) {
  offset = o;
}

//
// Set the scaling factor for plotting lines
//
void DataPlot::SetLineScale(PLFLT s) {
  scale = s;
}

//
// Set the line width for plotting lines
//
void DataPlot::SetLineWidth(int w) {
  linewidth = w;
}

//
// Set the scaling factor for plotting vectors
//
void DataPlot::SetVectorScale(PLFLT s) {
  rscale = s;
}

//
// Invert horizontal axes?
//
void DataPlot::SetInvertAxes(bool i) {
  invert = i;
}

//
// Set colour scheme
//
void DataPlot::SetColScheme(int cs) {
  colscheme = cs;

  if (cs == 3 || cs == 4 || cs == 7 || cs == 8) {
    whitezero = true;
  }
  else {
    whitezero = false;
  }
  if ((cs >= 5) && (cs <= 8)) {
    colreverse = true;
  }
  else {
    colreverse = false;
  }
  if ((cs == 2) || (cs == 4) || (cs == 6) || (cs == 8)) {
    iscol = true;
  }
  else {
    iscol = false;
  }

}

//
// Set the scaling factors for the axes
//
void DataPlot::SetAxisScaling(int axis, PLFLT scale) {
  switch (axis) {
  case 1:
    xscale = scale;
    break;
  case 2:
    yscale = scale;
    break;
  case 3:
    zscale = scale;
    break;
  default:
    cerr << "Invalid axis" << endl;
  }
}

//
// Set the scaling factors for the data
//
void DataPlot::SetDataScaling(PLFLT scale) {
  dscale = scale;
}

//
// Plot a mesh or solid surface for 3D plots
//
void DataPlot::SetDrawMesh(bool m) {
  mesh = m;
}

//
// Set view angles for 3D surface plot
//
void DataPlot::SetViewAngle(PLFLT a, PLFLT b) {
  alt = a;
  az = b;
}

//
// Rotate the horizontal axes by angle theta in degrees
//
void DataPlot::SetRotateAxes(PLFLT t) {
  theta = t;
}

//
// Label contours?
//
void DataPlot::SetPlotContLabel(bool l) {
  contlab = l;
}

//
// Is there a plot already open?
//
bool DataPlot::IsOpen() {
  return ((pls != NULL) || fout.is_open());
}

//
// Expand and escape sequences and plot titles
//
void DataPlot::PlotTitles(char *name, int sdim, PLFLT sval) {
#ifdef USE_PLPLOT
  char *t, *xt, *yt;
  PLFLT def, sca;

  GetExpTitles(name, sdim, sval);

  escstr.SetString(title);
  t = escstr.GetExpandedString();

  escstr.SetString(xtitle);
  xt = escstr.GetExpandedString();

  escstr.SetString(ytitle);
  yt = escstr.GetExpandedString();

  pls->gchr(def,sca);
  pls->schr(0,1.5);
  pls->lab(xt,yt,t);
  pls->schr(def,sca);

  delete[] t;
  delete[] xt;
  delete[] yt;
#endif
}

//
// Expand and escape sequences and plot titles
//
void DataPlot::PlotTitles3D(char *name, int sdim, PLFLT sval) {
#ifdef USE_PLPLOT
  char *t, *xt, *yt, *zt;
  
  GetExpTitles(name, sdim, sval);
  
  escstr.SetString(title);
  t = escstr.GetExpandedString();
  
  escstr.SetString(xtitle);
  xt = escstr.GetExpandedString();
    
  escstr.SetString(ytitle);
  yt = escstr.GetExpandedString();

  escstr.SetString(ztitle);
  zt = escstr.GetExpandedString();
    
  pls->box3("bnstu",xt,0.0,0,"bnstu",yt,0.0,0,"bcdmnstuv",zt,0.0,0);
  pls->mtex("t", 1.0, 0.5, 0.5, t);

  delete[] t;
  delete[] xt;
  delete[] yt;
  delete[] zt;
#endif
}

//
// Set up the string expander for the titles
//
void DataPlot::GetExpTitles(char *name, int sdim, PLFLT sval) {
  char xaxis[5];
  char yaxis[5];
  char zaxis[5];
  char valstr[20];

  sprintf(valstr,"%g",sval);

  switch (sdim) {
  case 1:
    sprintf(xaxis,"y");
    sprintf(yaxis,"z");
    sprintf(zaxis,"x");
    break;
  case 2:
    sprintf(xaxis,"x");
    sprintf(yaxis,"z");
    sprintf(zaxis,"y");
    break;
  case 3:
  case 4:
  case 5:
    sprintf(xaxis,"x");
    sprintf(yaxis,"y");
    if (sdim == 3) 
      sprintf(zaxis,"z");
    else if (sdim == 4)
      sprintf(zaxis,"h");
    else // if (sdim == 5)
      sprintf(zaxis,"i");      
    break;
  default:
    xaxis[0] = '\0';
    yaxis[0] = '\0';
    zaxis[0] = '\0';
  }

  escstr.SetEscKey('d',name);
  escstr.SetEscKey('v',valstr);
  escstr.SetEscKey('x',xaxis);
  escstr.SetEscKey('y',yaxis);
  escstr.SetEscKey('z',zaxis);

}
 
//
// Plot the hill using the generated interpolation grid
// 
void DataPlot::PlotHill(int sdim, PLFLT **xigrid, PLFLT **yigrid, 
			PLFLT **zigrid, PLFLT **higrid, 
			PLFLT *xgrid, PLFLT *ygrid, int n1, int n2) {
  int orogtype; 
  PLFLT *orog1;
  PLFLT **orog2;

  if ((sdim == 1) || (sdim ==2)) {
    orogtype = 1;   
  }
  else {
    orogtype = 2;
  }
  
  // Check if we are actually plotting the hill
  if (plothill & PLOTINTERSECT) {
    switch (orogtype) {
      // Vertical section
    case 1:
      orog1 = new PLFLT[n1];
      for (int i=0; i<n1; i++) {
	orog1[i] = higrid[i][0]*zscale;
      }
      PlotHill1D(orog1, xgrid, n1,0.0);
      break;
      // Horizontal section
    case 2:
      orog2 = AllocArray<PLFLT>(n1,n2);
      for (int i=0; i<n1; i++) {
	for (int j=0; j<n2; j++) {
	  orog2[i][j] = zigrid[i][j] - higrid[i][j];
	}
      }
      PlotHill2D(1.0E-3,orog2, xgrid, ygrid, n1, n2);
      break;
    }
  }

    // Check if we are plotting the base.
  if ((plothill & PLOTBASE) && (orogtype == 2)) {
    PlotHill2D(1.0E-3, higrid, xgrid, ygrid, n1, n2);
  }

   
}

//
// Check for flow separation over a hill
//
PLFLT DataPlot::Separation(DataSet *taux) {
  PLFLT *data, **orog, y, *grid, wid = 0;
  int nx;

  if (taux == NULL)
    return -2.;

  nx = taux->GetDim(2);
  data = taux->Get1D(2);
  orog = taux->GetOrog();
  grid = taux->GetGrid(2);

  for (int i=1;i<nx;i++) {
    if (orog[0][i-1] < 0) {
      if (wid == 0) {
	wid = grid[i] - (grid[i]-grid[i-1])/(orog[0][i-1]-orog[0][i]);
      }
      // Check if stress changes sign
      if (data[i]*data[i-1] < 0.) {
	y = grid[i-1] + data[i-1]/(data[i-1]-data[i])*(grid[i]-grid[i-1]);
	cout << -y << "\t" << y/wid << endl;
	return -y;
      }
    }
  }

  cout << "0.0\t0.0" << endl;
  return -1.;
}

//
// Use a logarithmic scale for contours
//
void DataPlot::SetLogContour(bool l) {
  logcontour = l;
}


//
// Set line style for contours, lines etc.
// 
void DataPlot::SetLineStyle(int ls) {
  if (ls < -1) {
    cerr << "Invalid line style " << ls << endl;
    return;
  }
  linestyle = ls;
}

//
// Set line style for contours, lines etc.
// 
void DataPlot::SetVectorStyle(int vs, bool fill, bool centre) {

  if ((vs < -1) || (vs >= narr)) {
    cerr << "Invalid vector style " << vs << endl;
    return;
  }
  vectorstyle = vs;
  vectorfill = fill;
  vectorcentre = centre;
}

//
// Set vector style using defined parameters
//
void DataPlot::SetVectorStyle() {
  int iarrow[narr] = {6,6,11};
  PLFLT arrow_x[narr][12] = { {-0.5, 0.5, 0.3, 0.5, 0.3, 0.5},
			      {-0.5, 0.3, 0.3, 0.5, 0.3, 0.3},
			      {-0.5, 0.5, 0.3, 0.5, 0.3, 0.5, 0.3, 0.1, 
			       0.3, 0.1, 0.3} };
  PLFLT arrow_y[narr][12] = { {0.0, 0.0, 0.2, 0.0, -0.2, 0.0},
			      {0.0, 0.0,   0.2, 0.0, -0.2, 0.0}, 
			      {0.0, 0.0, 0.2, 0.0, -0.2, 0.0, 0.0, 0.2, 0.0, -0.2, 0.0} };

  PLFLT *axp, *ayp;
  PLFLT ax[12], ay[12];

  if (vectorstyle == -1) return;

#ifndef MISSING_PLSTREAM_VECT
  if (vectorcentre) {
    for (int i=0;i<iarrow[vectorstyle];i++) {
      ax[i] = arrow_x[vectorstyle][i]+0.5;
    }
    axp = ax;
    ayp = arrow_y[vectorstyle];
    
  }
  else {
    axp = arrow_x[vectorstyle];
    ayp = arrow_y[vectorstyle];
  }
#ifdef USE_PLPLOT
  pls->svect(axp,ayp,iarrow[vectorstyle],vectorfill);
#else
  cout << "Must have plplot support enabled to set vector style" << endl;
#endif
#endif
}

//
// Set colour used for lines, vectors etc
//
void DataPlot::SetColour(int c) {
  linecol = c;
}


//
// Plot a given label on the figure.
// 
void DataPlot::PlotLabel(char *text, PLFLT x, PLFLT y, PLFLT just, 
			 PLFLT angle) {
  PLFLT def, sca;
  PLFLT a = angle/180.0*M_PI;
  PLFLT xmin, xmax, ymin, ymax;

#ifdef USE_PLPLOT
  pls->gvpw(xmin,xmax,ymin,ymax);

  if ((xmin<=x) && (xmax>=x) && (ymin<=y) && (ymax>=y)) {
    
    pls->gchr(def,sca);
    pls->schr(0,textsize);
    pls->ptex(x,y,cos(a),sin(a),just,text);
    pls->schr(def,sca);
  }
#else
  cout << "Must have plplot support enabled to plot labels" << endl;
#endif

}

//
// Set text scaling factor for labels
//
void DataPlot::SetTextScaling(PLFLT ts) {
  if (ts < 0) {
    cerr << "Invalid text scaling " << ts << endl;
    return;
  }
  
  textsize = ts;
}

//
// Set aspect ratio for plot
//
void DataPlot::SetAspectRatio(PLFLT a) {
  if (a < 0) {
    cerr << "Invalid aspect ratio " << a << endl;
    return;
  }
  
  aspect = a;
}

