// -*-c++-*-
// 
// blasvis++ - C++ version of graphics routines
//
// DataPlot : Class to plot a specified data set(s).
//            Handles various types of plot.
// 
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#ifndef _DATAPLOT_H
#define _DATAPLOT_H

#include "../config.h"

#ifdef USE_PLPLOT
#include <plplot/plstream.h>
#endif
#include <fstream>
#include "blastype.h"
#include "dataset.h"
#include "timeseries.h"
#include "escstring.h"
#include "arrayutil.h"

class DataPlot : public ArrayUtil {

public:
  // Enumerator for device types
  enum devtype {
    none,
    xwin,
    ps,
    psc,
    xfig,
    pbm,
    png,
    jpeg,
    file
  };

  const static int PLOTBASE = 1;
  const static int PLOTINTERSECT = 2;

  // Class functions
public: 
  DataPlot();

  // Initialise plot
  void InitPlot();
  // Start a new page
  void NextPlot();
  // Close plot
  void FinishPlot();

  // Check if plot is open
  bool IsOpen();

  // Set range using specified values
  void SetRange(PLFLT,PLFLT,PLFLT,PLFLT, PLFLT, PLFLT);
  void AutoRange(DataSet *, int);
  void SetContours(PLFLT, PLFLT, int);

  // Set up plot options
  void SetTitles(char *, char *, char *, char *);
  void SetPlotHill(int);
  void SetPlotLegend(bool);
  void SetPlotContLabel(bool);
  void SetLineOffset(PLFLT);
  void SetLineScale(PLFLT);
  void SetLineWidth(int);
  void SetLineStyle(int);
  void SetVectorStyle(int,bool,bool);
  void SetColour(int);
  void SetVectorScale(PLFLT);
  void SetInvertAxes(bool);
  void SetColScheme(int);
  void SetRunParms(int,int);
  void SetAxisScaling(int,PLFLT);
  void SetDataScaling(PLFLT);
  void SetDrawMesh(bool);
  void SetViewAngle(PLFLT,PLFLT);
  void SetRotateAxes(PLFLT);
  void SetLogContour(bool);
  void SetTextScaling(PLFLT);
  void SetAspectRatio(PLFLT);

  // Set up device and interpolation settings
  void SetDevice(devtype);
  void SetTransform(bool);
  void SetInterpolation(int,int);
  bool NeedDevFile();
  void SetDevFile(const char *);

  // Query if a legend should be plotted
  bool GetPlotLegend();

  // Various types of contour / line / vector plot
  void ColContPlot(DataSet *, int, PLFLT, DataSet *);
  void ColContPlot(PLFLT **, int, int, PLFLT *, PLFLT *, PLFLT *, int, 
		   bool = false, bool = false, bool = false);
  void LineContPlot(DataSet *, int, PLFLT, DataSet *);
  void LineContPlot(PLFLT **, int, int, PLFLT *, PLFLT *, PLFLT *, int);
  void VectorPlot(DataSet *, DataSet *, int, PLFLT, DataSet *);
  void VectorPlot(PLFLT, PLFLT, PLFLT, PLFLT);
  void VectorPlot(PLFLT **, PLFLT **, int, int, PLFLT *, PLFLT *);
  void SurfacePlot(DataSet *, int, PLFLT, DataSet *);
  void SurfacePlot(PLFLT **, int, int, PLFLT *, PLFLT *);
  void PlotHill1D(PLFLT *, PLFLT *, int, PLFLT = 0.0);
  void PlotHill2D(PLFLT, PLFLT **, PLFLT *, PLFLT *, int, int);
  void PlotHill(int, PLFLT **, PLFLT **, PLFLT **, PLFLT **, PLFLT *, 
		PLFLT *, int, int);
  void LinePlot(DataSet *, int, PLFLT, PLFLT, char *, DataSet *);
  void LinePlot(TimeSeries *, char *);
  void LinePlot(DataSet *, int, PLFLT *, PLFLT *, PLFLT *, PLFLT *, PLFLT *, char *);
  void LinePlot(PLFLT *, int, PLFLT *, int, PLFLT, PLFLT, PLFLT, PLFLT);
  void LinePlotLegend(char *, int, int);
  void Hodograph(DataSet *, DataSet *, int, PLFLT, PLFLT, char *, DataSet *);
  void Hodograph(TimeSeries *, TimeSeries *, char *);
  void PlotLabel(char *, PLFLT, PLFLT, PLFLT = 0.5, PLFLT = 0.0);

  PLFLT Separation(DataSet *);

  // Private functions
private:  
  void PlotTitles(char *, int, PLFLT);
  void PlotTitles3D(char *, int, PLFLT);
  void GetExpTitles(char *, int, PLFLT);
  void SetVectorStyle();

private:
  // PlPlot stream 
#ifdef USE_PLPLOT 
  plstream *pls;
#else
  void *pls;
#endif

  // Output device
  devtype dev;
  char devfile[80];

  // Range for graph
  PLFLT xmin, xmax, ymin, ymax, zmin, zmax, dmin, dmax;
  PLFLT tmin, tmax;
  int nlev;

  // Number of plots drawn
  int numplots;
  // Number of lines drawn
  int numlines;
  int maxlines;
  PLFLT offset, scale, rscale;
  int linewidth;

  // Scaling for axes to do units conversion  
  PLFLT xscale,yscale,zscale,dscale;
  // Logarithmic axes?
  bool xlog, ylog, zlog, dlog;

  // Number of contours drawn
  int ncont;

  // Plot a legend?
  bool dolegend;
  bool contlab;
  int plothill;
  bool transform;
  bool invert;
  PLFLT theta;
  int colscheme;
  bool whitezero;
  bool colreverse;
  bool iscol;
  bool autoContourRange;  
  bool autoXRange;
  bool autoYRange;
  bool autoZRange;
  bool autoTRange;
  bool logcontour;
  int linestyle;
  int vectorstyle;
  bool vectorfill;
  bool vectorcentre;
  int linecol;
  PLFLT textsize;
  PLFLT aspect;

  // Settings for 3D surface plot
  bool mesh;
  PLFLT alt, az;

  // Title arrays
  EscString escstr;
  char title[80];
  char xtitle[40];
  char ytitle[40];
  char ztitle[40];

  // Size of interpolation grid
  int nix, niy;

  // Stream for saving data to file
  std::ofstream fout;

  // number of arrow styles defined
  static const int narr = 3;

};

#endif
