// 
// blasvis++ - C++ version of graphics routines
//
// DataSet : Class to encapsulate an array of data, including
//           reading in the data and interpolating it onto an
//           an arbitrary grid.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#include "dataset.h"
#include "datavect.h"
#include <iostream>
#include <fstream>
#include <cstring>
#include <cmath>
#include <stdexcept>

using namespace std;

DataSet::DataSet() 
{
  data = NULL;
  ndim = 0;
  dim[0] = 0;
  dim[1] = 0;
  dim[2] = 0;

  xgrid = NULL;
  ygrid = NULL;
  zgrid = NULL;
  orog = NULL;
  name[0] = '\0';

  isvector = 0;

}

//
// Allocate empty data set
//
DataSet::DataSet(int d, int gnx, int gny, int gnz) 
{
  data = Alloc3DArray<bldata>(gnx,gny,gnz);
  ndim = d;
  dim[0] = gnx;
  dim[1] = gny;
  dim[2] = gnz;

  xgrid = NULL;
  ygrid = NULL;
  zgrid = NULL;
  orog = NULL;
}

DataSet::DataSet(bldata*** d,int n ,int nx,int ny,int nz, bldata *xg, bldata *yg, 
		 bldata *zg, bldata h, bldata **oro) {
  data = d;
  ndim = n;
  dim[0] = nx;
  dim[1] = ny;
  dim[2] = nz;

  SetGrids(xg,yg,zg,nx,ny,nz,h,oro);
}

//
// Set the grid pointers for the data
//
void DataSet::SetGrids(bldata * x, bldata *y, bldata *z, int gnx, int gny, 
		       int gnz, bldata h, DataSet *o) {

  bldata ** tmp;
  
  if (o) 
    tmp = o->Get2D(3);
  else 
    tmp = NULL;
  SetGrids(x, y, z, gnx, gny, gnz, h, tmp);
}

//
// Set the grid pointers for the data
//
void DataSet::SetGrids(bldata * x, bldata *y, bldata *z, int gnx, int gny, 
		       int gnz, bldata h, bldata **o) {

  bldata *** tmp;

  xgrid = x;
  ygrid = y;
  zgrid = z;
  orog = o;
  H = h;
  
  // We may need to tweak the data to get the correct alignment

  // This has extra points in - we need to filter these out. 
  if ((dim[0] == gnx + 2) && (dim[1] == gny + 2)) {
    //    cout << "Correcting array sizes" << endl;
    // Create new array
    // Use dim[2] not gnz as it may be zero
    tmp = Alloc3DArray<bldata>(gnx,gny,dim[2]);
    // Copy data across
    for (int i=0;i<gnx;i++) {
      for (int j=0;j<gny;j++) {
	for (int k=0;k<dim[2];k++) {
	  tmp[i][j][k] = data[i+1][j+1][k];
	}
      }
    }
    Free3DArray(data);
    // Add in new data
    data = tmp;
    dim[0] = gnx;
    dim[1] = gny;
    // Don't want to change this
    //    dim[2] = gnz;
  }

}


//
// Debugging routine to dump the data set to test it.
//
void DataSet::DisplayData() {
  for (int k=0;k<dim[2];k++) {
    for (int j=0;j<dim[1];j++) {
      for (int i=0;i<dim[0];i++) {
	cout << xgrid[i] << "," << ygrid[j] << "," << zgrid[k] << " " << data[i][j][k] << "," << orog[i][j] << endl;
      }
    }
  }
}
//
// Debugging routine to dump the data set to test it.
//
void DataSet::DisplayData(const char *fname) {
  ofstream fout;

  fout.open(fname);

  for (int k=0;k<dim[2];k++) {
    for (int j=0;j<dim[1];j++) {
      for (int i=0;i<dim[0];i++) {
	fout << xgrid[i] << "\t" << ygrid[j] << "\t" << zgrid[k] << "\t" << 
	  data[i][j][k] << "\t" << orog[i][j] << endl;
      }
    }
  }
}

//
// Debugging routine to dump the grids to test it.
//
void DataSet::DisplayGrids() {

  for (int i=0;i<dim[0];i++) {
    cout << i << " " << xgrid[i] << endl;
  }
  for (int j=0;j<dim[1];j++) {
    cout << j << " " << ygrid[j] << endl;
  }
  for (int k=0;k<dim[2];k++) {
    cout << k << " " << zgrid[k] << endl;
  }
}

//
// Display statistics about the data set
//
void DataSet::DisplaySummary() {
  bldata max, min, dz1, dx, dy, dz;
  double mean,absmean,vol,sum,abssum,mean2,stddev, w;  
  int imax,jmax,kmax,imin,jmin,kmin;
  int nx,ny,nz;

  max = data[0][0][0];
  min = max;
  imax = 0;
  jmax = 0;
  kmax = 0;
  imin = 0;
  jmin = 0;
  kmin = 0;
  mean = 0;
  mean2 = 0;
  absmean = 0;
  vol = 0;
  sum = 0;
  abssum = 0;
  stddev = 0;

  nx = dim[0];
  ny = dim[1];
  nz = dim[2];

  for (int k=1;k<nz-1;k++) {
    
    dz1 = (zgrid[k+1]-zgrid[k-1])*0.5;
    
    for (int j=0;j<ny;j++) {
      if (ny == 1)
	dy = 1;
      else if (j==0) 
	dy = ygrid[1]-ygrid[0];
      else if (j == ny-1)
	dy = ygrid[j]-ygrid[j-1];
      else 
	dy = (ygrid[j+1]-ygrid[j-1])*0.5;
      for (int i=0;i<nx;i++) {
	if (nx == 1)
	  dx = 1;
	else if (i==0) 
	  dx = xgrid[1]-ygrid[0];
	else if (i == nx-1)
	  dx = xgrid[i]-xgrid[i-1];
	else 
	  dx = (xgrid[i+1]-xgrid[i-1])*0.5;
	dz = dz1*(1.0-orog[i][j]/H);
	vol += dx*dy*dz;
      }
    }
  }

  // Don't do top and bottom points as these are often not calculated
  for (int k=1;k<nz-1;k++) {
    
    dz1 = (zgrid[k+1]-zgrid[k-1])*0.5;
    
    for (int j=0;j<ny;j++) {
      if (ny == 1)
	dy = 1;
      else if (j==0) 
	dy = ygrid[1]-ygrid[0];
      else if (j == ny-1)
	dy = ygrid[j]-ygrid[j-1];
      else 
	dy = (ygrid[j+1]-ygrid[j-1])*0.5;
      for (int i=0;i<nx;i++) {
	if (nx == 1)
	  dx = 1;
	else if (i==0) 
	  dx = xgrid[1]-ygrid[0];
	else if (i == nx-1)
	  dx = xgrid[i]-xgrid[i-1];
	else 
	  dx = (xgrid[i+1]-xgrid[i-1])*0.5;
	dz = dz1*(1.0-orog[i][j]/H);
	w = dx*dy*dz;
	sum += data[i][j][k]*w;
	abssum += fabs(data[i][j][k])*w;
	mean += data[i][j][k]*w/vol;
	mean2 += data[i][j][k]*data[i][j][k]*w/vol;
	absmean += fabs(data[i][j][k])*w/vol;
	if (data[i][j][k]>max) {	
	  max =  data[i][j][k];
	  imax = i;
	  jmax = j;
	  kmax = k;
	}
	if (data[i][j][k]<min) {
	  min = data[i][j][k];
	  imin = i;
	  jmin = j;
	  kmin = k;
	}
      }
    }
  }
  stddev = sqrt(mean2 - mean*mean);
  cout << "Data maximum " << max << " at (" << xgrid[imax] << "," << ygrid[jmax] << "," << zgrid[kmax] << ")" << endl;
  cout << "Data minimum " << min << " at (" << xgrid[imin] << "," << ygrid[jmin] << "," << zgrid[kmin] << ")" << endl;
  cout << "Mean value is " << mean << endl;
  cout << "Mean absolute value is " << absmean << endl;
  cout << "Volume is " << vol << endl;
  cout << "Sum is " << sum << endl;
  cout << "Absolute Sum is " << abssum << endl;
  cout << "Standard deviation is " << stddev << "\n" << endl;
  //  cout << "C " << max << "\t" << min << "\t" << mean << "\t" << absmean << "\t" << stddev << "\t" << vol << endl;
  
}


// 
// Interpolate a section from a given array of data
//
void DataSet::GetInterpSection(bldata ***s, bldata **x, bldata **y, 
			       bldata **z, bldata **h, bldata dscale, 
			       int nx, int ny, bool invert, bool transform)
{
  bldata wx1, wx2, wy1, wy2, wz1, wz2, w, hloc;
  int ii, jj, kk;
  bldata **sect;

  sect = AllocArray<bldata>(nx,ny);

  for (int i=0; i<nx;i++) {
    for (int j=0;j<ny;j++) {
      if (transform) {
	hloc = h[i][j];
      }
      else {
	hloc = 0;
      }
      if (transform && (z[i][j] < hloc)) {
	sect[i][j] = 0.0;
      }
      else {
	// Get the weighting values
	GetGridWeight(xgrid,dim[0],x[i][j],ii,wx1,wx2,0.0); 
	GetGridWeight(ygrid,dim[1],y[i][j],jj,wy1,wy2,0.0); 
	GetGridWeight(zgrid,dim[2],z[i][j],kk,wz1,wz2,hloc);


	// Add on each contribution where needed
	sect[i][j] = wx1*wy1*wz1*data[ii][jj][kk];
	w = wx1*wy1*wz2;
	if (w > 0) sect[i][j] += w*data[ii][jj][kk+1];
	w = wx1*wy2*wz1;
	if (w > 0) sect[i][j] += w*data[ii][jj+1][kk];
	w = wx2*wy1*wz1;
	if (w > 0) sect[i][j] += w*data[ii+1][jj][kk];
	w = wx1*wy2*wz2;
	if (w > 0) sect[i][j] += w*data[ii][jj+1][kk+1];
	w = wx2*wy1*wz2;
	if (w > 0) sect[i][j] += w*data[ii+1][jj][kk+1];
	w = wx2*wy2*wz1;
	if (w > 0) sect[i][j] += w*data[ii+1][jj+1][kk];
	w = wx2*wy2*wz2;
	if (w > 0) sect[i][j] += w*data[ii+1][jj+1][kk+1];
	sect[i][j] *= dscale;


      }      
    }
  }

  if (invert && ( (isvector == 1) || (isvector == 2))) {
    for (int i=0; i<nx;i++) {
      for (int j=0;j<ny;j++) {
	sect[i][j] = -sect[i][j];
      }
    }
  }


  *s = sect;
}


// ********************************************
// * Get a regular 2-D grid for interpolation *
// ********************************************
//
// This routine takes in the grid dimensions n1,n2 and produces
// a grid using the dimensions [xyz]min/[xyz]max.
// [xyz]pgrid contain the grids for plotting
// [xyz]gri contain the grids for interpolating
// This allows rotation and scaling of axes.
//
void DataSet::CreateInterpGrid(int n1, int n2, 
			       bldata xmin, bldata xmax,
			       bldata ymin, bldata ymax,
			       bldata zmin, bldata zmax,
			       int sdim, bldata **xpgrid, bldata **ypgrid,
			       bldata ***xgri, bldata ***ygri, 
			       bldata ***zgri, bldata ***hh, 
			       bldata xscale, bldata yscale, bldata zscale, 
			       bldata theta, bool invert, bool transform, DataSet *extra) {
  bldata **xgr, **ygr, **zgr, **h;
  bldata *xpgr, *ypgr;
  bldata wx1, wx2, wy1, wy2, wz1, wz2;
  bldata l, d1, d2, lmin, costh, sinth, tmp, xmi, xma, ymi, yma;
  bldata xpmin, xpmax, ypmin, ypmax;
  bldata ***te, tc;
  int ii, jj, kk;

  if (sdim == 5) {
    if (extra == NULL) {
      cerr << "Temperature field need for isentropic surfaces" << endl;
      return;
    }
    else {
      te = extra->GetData();
    }
  }
  

  costh = cos(theta/180.0*M_PI);
  sinth = sin(theta/180.0*M_PI);

  xgr = AllocArray<bldata>(n1,n2);
  ygr = AllocArray<bldata>(n1,n2);
  zgr = AllocArray<bldata>(n1,n2);
  h = AllocArray<bldata>(n1,n2);

  xpgr = new bldata[n1];
  ypgr = new bldata[n2];

  xmi = xmin*costh - ymin*sinth;
  xma = xmax*costh - ymax*sinth;
  ymi = ymin*costh + xmin*sinth;
  yma = ymax*costh + xmax*sinth;
  xpmin = (xmi<xma)?xmi:xma;
  xpmax = (xmi>xma)?xmi:xma;
  ypmin = (ymi<yma)?ymi:yma;
  ypmax = (ymi>yma)?ymi:yma;

  // Invert only works for vertical sections. Use a rotation of 180 
  // degrees for a horizontal section.
  if (invert) {
    tmp = xmin;
    xmin = xmax;
    xmax = tmp;
    tmp = ymin;
    ymin = ymax;
    ymax = tmp;
  }



  for (int i=0;i<n1;i++) {
    for (int j=0;j<n2;j++) {

      switch (sdim) {
      case 1:
      case 2:
	xgr[i][j] = ((xmax-xmin)*i/(n1-1)+xmin);
	ygr[i][j] = ((ymax-ymin)*i/(n1-1)+ymin);
	break;
      case 3:
      case 4:
      case 5:
	xgr[i][j] = ((xpmax-xpmin)*i/(n1-1) + xpmin)*costh +
	  ((ypmax-ypmin)*j/(n2-1) + ypmin)*sinth;
	ygr[i][j] = ((ypmax-ypmin)*j/(n2-1) + ypmin)*costh - 
	  ((xpmax-xpmin)*i/(n1-1) + xpmin)*sinth;
	break;
      default:
	break;
      }

      GetGridWeight(xgrid,dim[0],xgr[i][j],ii,wx1,wx2,0.0);
      GetGridWeight(ygrid,dim[1],ygr[i][j],jj,wy1,wy2,0.0);
      h[i][j] = wx1*wy1*orog[ii][jj];
      if (ii < dim[0]-1) {
	h[i][j] += wx2*wy1*orog[ii+1][jj];
	if (jj < dim[1]-1) {
	  h[i][j] += wx1*wy2*orog[ii][jj+1] + wx2*wy2*orog[ii+1][jj+1];
	}
      }
      else if (jj < dim[1]-1) {
	h[i][j] += wx1*wy2*orog[ii][jj+1];
      }
      switch(sdim) {
      case 1:
      case 2:
	zgr[i][j] = ((zmax-zmin)*j/(n2-1)+zmin);
	break;
      case 3:
	zgr[i][j] = ((zmax-zmin)*j/(n2-1)+zmin);
	break;
      case 4:
	zgr[i][j] = ((zmax-zmin)*j/(n2-1)+zmin);
	if (transform) 
	  zgr[i][j] += h[i][j];
	break;
      case 5:
	extra->GetDataWeight(ii,jj,wx1,wx2,wy1,wy2,((zmax-zmin)*j/(n2-1)+zmin),
			    kk,wz1,wz2);
	if ((wz1 == 0.0) && (wz2 == 0.0)) {
	  zgr[i][j] = -9999.;
	}
	else {
	    zgr[i][j] = (zgrid[kk]*wz1+zgrid[kk+1]*wz2);
	    if (transform) 
	      zgr[i][j] = zgr[i][j] * 
		(1.0 - h[i][j]/zgrid[dim[2]-1]) + h[i][j];	    
	}
	tc = (wx1*wy1*te[ii][jj][kk] + wx2*wy1*te[ii+1][jj][kk] + 
	      wx1*wy2*te[ii][jj+1][kk] + wx2*wy2*te[ii+1][jj+1][kk])*wz1 + 
	  (wx1*wy1*te[ii][jj][kk+1] + wx2*wy1*te[ii+1][jj][kk+1] + 
	   wx1*wy2*te[ii][jj+1][kk+1] + wx2*wy2*te[ii+1][jj+1][kk+1])*wz2;
	//	cout << xgr[i][j] << " " << ygr[i][j] << " " << zgr[i][j] << " " <<tc << endl;
	//	cout << wx1 << " " << wx2 << " " << wy1 << " " << wy2 << " " << wz1 << " " << wz2 << endl;
	    
	break;
      }
    }
  }
  switch (sdim){
  case 1: 
  case 2:
    d1 = xmin*xmin*xscale*xscale + ymin*ymin*yscale*yscale;
    d2 = xmax*xmax*xscale*xscale + ymax*ymax*yscale*yscale;
    l = sqrt( (xmax-xmin)*(xmax-xmin)*xscale*xscale + 
	      (ymax-ymin)*(ymax-ymin)*yscale*yscale );
    lmin = -0.5*((d1 - d2)/l + l);
    for (int i=0;i<n1;i++) {
      xpgr[i] = lmin + l*i/(n1-1);
    }
    for (int j=0;j<n2;j++) {
      ypgr[j] = (zmin + (zmax-zmin)*j/(n2-1))*zscale;
    }
    break;
  case 3:
  case 4:
  case 5:
    for (int i=0; i< n1; i++) {
      xpgr[i] = ((xpmax-xpmin)*i/(n1-1)+xpmin)*xscale;
    }
    for (int j=0; j< n2; j++) {
      ypgr[j] = ((ypmax-ypmin)*j/(n2-1)+ypmin)*xscale;
    }
  }
  *xpgrid = xpgr;
  *ypgrid = ypgr;
  *xgri = xgr;
  *ygri = ygr;
  *zgri = zgr;
  *hh = h;

}


// ********************************************
// * Get a regular 1-D grid for interpolation *
// ********************************************
//
// This routine takes in the grid dimensions n1 and produces
// a 1-D grid using the dimensions [xyz]min/[xyz]max.
// [xyz]pgrid contain the grids for plotting
// [xyz]gri contain the grids for interpolating
// This allows rotation and scaling of axes.
//
void DataSet::CreateInterpGrid1D(int n1,
			       bldata xmin, bldata xmax,
			       bldata ymin, bldata ymax,
			       bldata zmin, bldata zmax,
			       int sdim, bldata **xpgrid,
			       bldata **xgri, bldata **ygri, 
			       bldata **zgri, bldata **hh, 
			       bldata xscale, bldata yscale, bldata zscale, 
			       bldata theta, bool invert, bool transform, DataSet *extra) {
  bldata *xgr, *ygr, *zgr, *h;
  bldata *xpgr;
  bldata wx1, wx2, wy1, wy2, wz1, wz2;
  bldata l, d1, d2, lmin, costh, sinth, tmp, xmi, xma, ymi, yma;
  bldata xpmin, xpmax, ypmin, ypmax;
  bldata ***te, tc;
  int ii, jj, kk;

  // Options for sdim are
  // 1 - x      
  // 2 - y
  // 3 - z
  // 14 - hx (const h in x direction)
  // 24 - hy (const h in y direction)
  // 15 - ix (isentropic)
  // 25 - iy (isentropic)
  // 16 - sx (const streamfunction)
  // 26 - sy (const streamfunction)

  if ((sdim == 5) || (sdim == 6)) {
    if (extra == NULL) {
      cerr << "Extra field need for constant surfaces" << endl;
      return;
    }
    else {
      te = extra->GetData();
    }
  }
  

  costh = cos(theta/180.0*M_PI);
  sinth = sin(theta/180.0*M_PI);

  xgr = new bldata[n1];
  ygr = new bldata[n1];
  zgr = new bldata[n1];
  h = new bldata[n1];

  xpgr = new bldata[n1];

  xmi = xmin*costh - ymin*sinth;
  xma = xmax*costh - ymax*sinth;
  ymi = ymin*costh + xmin*sinth;
  yma = ymax*costh + xmax*sinth;
  xpmin = (xmi<xma)?xmi:xma;
  xpmax = (xmi>xma)?xmi:xma;
  ypmin = (ymi<yma)?ymi:yma;
  ypmax = (ymi>yma)?ymi:yma;

  // Invert only works for vertical sections. Use a rotation of 180 
  // degrees for a horizontal section.
  if (invert) {
    tmp = xmin;
    xmin = xmax;
    xmax = tmp;
    tmp = ymin;
    ymin = ymax;
    ymax = tmp;
  }



  for (int i=0;i<n1;i++) {

    // Set up the horizontal grids
    xgr[i] = ((xmax-xmin)*i/(n1-1)+xmin);
    ygr[i] = ((ymax-ymin)*i/(n1-1)+ymin);
    
    GetGridWeight(xgrid,dim[0],xgr[i],ii,wx1,wx2,0.0);
    GetGridWeight(ygrid,dim[1],ygr[i],jj,wy1,wy2,0.0);
    h[i] = wx1*wy1*orog[ii][jj];
    if (ii < dim[0]-1) {
      h[i] += wx2*wy1*orog[ii+1][jj];
      if (jj < dim[1]-1) {
	h[i] += wx1*wy2*orog[ii][jj+1] + wx2*wy2*orog[ii+1][jj+1];
      }
    }
    else if (jj < dim[1]-1) {
      h[i] += wx1*wy2*orog[ii][jj+1];
    }
    switch(sdim) {
    case 1:
    case 2:
    case 3:
      zgr[i] = ((zmax-zmin)*i/(n1-1)+zmin);
      break;
    case 4:
      zgr[i] = ((zmax-zmin)*i/(n1-1)+zmin);
      if (transform)
	zgr[i] += h[i];
      break;
    case 5:
    case 6:
      extra->GetDataWeight(ii,jj,wx1,wx2,wy1,wy2,((zmax-zmin)*i/(n1-1)+zmin),
			   kk,wz1,wz2);
      if ((wz1 == 0.0) && (wz2 == 0.0)) {
	  zgr[i] = -9999.;
      }
      else {
	zgr[i] = (zgrid[kk]*wz1+zgrid[kk+1]*wz2);
	if (transform) 
	  zgr[i] = zgr[i]*(1.0 - h[i]/zgrid[dim[2]-1]) + h[i];
      }
      // This is tedious as we need to ensure we don't overrun the arrays.
      tc = wx1*wy1*wz1*te[ii][jj][kk];
      tmp = wx2*wy1*wz1;
      if (tmp > 0) tc += wx2*wy1*wz1*te[ii+1][jj][kk];
      tmp = wx1*wy2*wz1;
      if (tmp > 0) tc += wx1*wy2*wz1*te[ii][jj+1][kk];
      tmp = wx2*wy2*wz1;
      if (tmp > 0) tc += wx2*wy2*wz1*te[ii+1][jj+1][kk];
      tmp = wx1*wy1*wz2;
      if (tmp > 0) tc += wx1*wy1*wz2*te[ii][jj][kk+1];
      tmp = wx2*wy1*wz1;
      if (tmp > 0) tc += wx2*wy1*wz2*te[ii+1][jj][kk+1];
      tmp = wx1*wy2*wz1;
      if (tmp > 0) tc += wx1*wy2*wz2*te[ii][jj+1][kk+1];
      tmp = wx2*wy2*wz1;
      if (tmp > 0) tc += wx2*wy2*wz2*te[ii+1][jj+1][kk+1];
      break;
    }
  }
  switch (sdim){
  case 1: 
  case 2:
  case 4:
  case 5:
  case 6:
    d1 = xmin*xmin*xscale*xscale + ymin*ymin*yscale*yscale;
    d2 = xmax*xmax*xscale*xscale + ymax*ymax*yscale*yscale;
    l = sqrt( (xmax-xmin)*(xmax-xmin)*xscale*xscale + 
	      (ymax-ymin)*(ymax-ymin)*yscale*yscale );
    lmin = -0.5*((d1 - d2)/l + l);
    for (int i=0;i<n1;i++) {
      xpgr[i] = lmin + l*i/(n1-1);
    }
    break;
  case 3:
    for (int i=0; i< n1; i++) {
      xpgr[i] = ((zmax-zmin)*i/(n1-1)+zmin)*zscale;
    }
  }
  *xpgrid = xpgr;
  *xgri = xgr;
  *ygri = ygr;
  *zgri = zgr;
  *hh = h;

  //  for (int i=0; i<n1; i++) {
  //    cout << xpgr[i] << " ";
  //  }
  //  cout << endl;

}

// 
// Interpolate a section from a given array of data
//
void DataSet::GetInterpLine2(bldata **s, bldata *x, bldata *y, 
			       bldata *z, bldata *h, bldata dscale,
			       int nx, bool invert)
{

  bldata *sect;

  sect = new bldata[nx];

  for (int i=0; i<nx;i++) {
    try {
      sect[i] = GetVal2(x[i],y[i],z[i]);
    }
    catch (exception e) {
      cout << e.what() << endl;
    }

    sect[i] *= dscale;
  }
  
  if (invert && ( (isvector == 1) || (isvector == 2))) {
    for (int i=0; i<nx;i++) {
      sect[i] = -sect[i];
    }
  }

  *s = sect;

}


// 
// Interpolate a section from a given array of data
//
void DataSet::GetInterpLine(bldata **s, bldata *x, bldata *y, 
			       bldata *z, bldata *h, bldata dscale,
			       int nx, bool invert, bool transform)
{
  bldata wx1, wx2, wy1, wy2, wz1, wz2, w, hloc;
  int ii, jj, kk;
  bldata *sect;

  sect = new bldata[nx];

  for (int i=0; i<nx;i++) {
    if (transform) {
      hloc = h[i];     
    }
    else {
      hloc = 0.0;
    }
    if (z[i] < hloc) {
      sect[i] = 0.0;
    }
    else {
      // Get the weighting values
      GetGridWeight(xgrid,dim[0],x[i],ii,wx1,wx2,0.0); 
      GetGridWeight(ygrid,dim[1],y[i],jj,wy1,wy2,0.0); 
      GetGridWeight(zgrid,dim[2],z[i],kk,wz1,wz2,hloc);

      // Add on each contribution where needed
      sect[i] = wx1*wy1*wz1*data[ii][jj][kk];
      w = wx1*wy1*wz2;
      if (w > 0) sect[i] += w*data[ii][jj][kk+1];
      w = wx1*wy2*wz1;
      if (w > 0) sect[i] += w*data[ii][jj+1][kk];
      w = wx2*wy1*wz1;
      if (w > 0) sect[i] += w*data[ii+1][jj][kk];
      w = wx1*wy2*wz2;
      if (w > 0) sect[i] += w*data[ii][jj+1][kk+1];
      w = wx2*wy1*wz2;
      if (w > 0) sect[i] += w*data[ii+1][jj][kk+1];
      w = wx2*wy2*wz1;
      if (w > 0) sect[i] += w*data[ii+1][jj+1][kk];
      w = wx2*wy2*wz2;
      if (w > 0) sect[i] += w*data[ii+1][jj+1][kk+1];
      sect[i] *= dscale;
    }      
  }

  if (invert && ( (isvector == 1) || (isvector == 2))) {
    for (int i=0; i<nx;i++) {
      sect[i] = -sect[i];
    }
  }

  *s = sect;

  //  for (int i=0; i<nx; i++) {
  //    cout << sect[i] << " ";
  //  }
  //  cout << endl;
}


//
// Get an interpolated line of data from dataset
//
void DataSet::GetInterpLine_old(bldata **d, int sdim, bldata sval1, bldata sval2, int n1, 
			    bldata min1, bldata max1, bldata offset, bldata scale, 
			    bool transform, DataSet *) {

  bldata *line;
  int nx,ny,nz,ii,jj,kk;
  bldata x,y,z,xmin,xmax,ymin,ymax,zmin,zmax,h;
  bldata wx1,wx2,wy1,wy2,wz1,wz2;
  
  if ((xgrid == NULL) || (ygrid == NULL) || (zgrid == NULL)) {
    cerr << "Grids not available - can't interpolate" << endl;
    return;
  }

  // Create array to hold interpolated data
  line = new bldata[n1];
  *d = line;

  // Check on which dimension we are using.

  switch (sdim) {
  case 1:
    xmin = min1;
    xmax = max1;
    nx = n1;
    // Interpolate section along x axis.
    GetGridWeight(ygrid,dim[1],sval1,jj,wy1,wy2,0.0);      

    for (int i=0;i<nx;i++) {
      x = (xmax-xmin)*i/(nx-1)+xmin;      
      GetGridWeight(xgrid,dim[0],x,ii,wx1,wx2,0.0);
      h = wx1*wy1*orog[ii][jj] + wx2*wy1*orog[ii+1][jj];
      if (jj < dim[1]-1) {
	h += wx1*wy2*orog[ii][jj+1] + wx2*wy2*orog[ii+1][jj+1];
      }
      if (transform) {
	GetGridWeight(zgrid,dim[2],sval2,kk,wz1,wz2,h);
      }
      else {
	GetGridWeight(zgrid,dim[2],sval2,kk,wz1,wz2,0);
      }
      line[i] =  wx1*wy1*wz1*data[ii][jj][kk] + wx1*wy1*wz2*data[ii][jj][kk+1] + 
	wx2*wy1*wz1*data[ii+1][jj][kk] + wx2*wy1*wz2*data[ii+1][jj][kk+1];
      if (jj < dim[1]-1) {
	line[i] +=  wx1*wy2*wz1*data[ii][jj+1][kk] + wx1*wy2*wz2*data[ii][jj+1][kk+1] + 
	  wx2*wy2*wz1*data[ii+1][jj+1][kk] + wx2*wy2*wz2*data[ii+1][jj+1][kk+1];
      }
      line[i] = line[i]*scale + offset;
    }
    break;

  case 2:
    ymin = min1;
    ymax = max1;
    ny = n1;
    // Interpolate section along y axis.
    GetGridWeight(xgrid,dim[0],sval1,ii,wx1,wx2,0.0);      

    for (int j=0;j<ny;j++) {
      y = (ymax-ymin)*j/(ny-1)+ymin;      
      GetGridWeight(ygrid,dim[2],y,ii,wy1,wy2,0.0);
      h = wx1*wy1*orog[ii][jj] + wx1*wy2*orog[ii][jj+1];
      if (ii < dim[0]-1) {
	h += wx2*wy1*orog[ii+1][jj] + wx2*wy2*orog[ii+1][jj+1];
      }
      if (transform) {
	GetGridWeight(zgrid,dim[2],sval2,kk,wz1,wz2,h);
      }
      else {
	GetGridWeight(zgrid,dim[2],sval2,kk,wz1,wz2,0);
      }
      line[j] =  wx1*wy1*wz1*data[ii][jj][kk] + wx1*wy1*wz2*data[ii][jj][kk+1] + 
	wx1*wy2*wz1*data[ii][jj+1][kk] + wx1*wy2*wz2*data[ii][jj+1][kk+1];
      if (ii < dim[0]-1) {
	line[j] +=  wx2*wy1*wz1*data[ii+1][jj][kk] + wx2*wy1*wz2*data[ii+1][jj][kk+1] + 
	  wx2*wy2*wz1*data[ii+1][jj+1][kk] + wx2*wy2*wz2*data[ii+1][jj+1][kk+1];
      }
      line[j] = line[j]*scale + offset;
    }
    break;

  case 3:
    zmin = min1;
    zmax = max1;
    nz = n1;

    GetGridWeight(xgrid,dim[0],sval1,ii,wx1,wx2,0.0);
    GetGridWeight(ygrid,dim[1],sval2,jj,wy1,wy2,0.0);
    h = wx1*wy1*orog[ii][jj];
    if (ii < dim[0]-1) {
      h += wx2*wy1*orog[ii+1][jj];
      if (jj < dim[1]-1) {
	h += wx1*wy2*orog[ii][jj+1] + wx2*wy2*orog[ii+1][jj+1];
      }
    }
    else if (jj < dim[1]-1) {
      h += wx1*wy2*orog[ii][jj+1];
    }

    // Interpolate section along z axis.
    for (int k=0; k<nz;k++) {
      z = (zmax-zmin)*k/(nz-1)+zmin;
      if (transform) {
	GetGridWeight(zgrid,dim[2],z,kk,wz1,wz2,h);
      }
      else {
	GetGridWeight(zgrid,dim[2],z,kk,wz1,wz2,0.0);
      }

      line[k] = wx1*wy1*wz1*data[ii][jj][kk] + wx1*wy1*wz2*data[ii][jj][kk+1];
      if (ii < dim[0]-1) {
	line[k] += wx2*wy1*wz1*data[ii+1][jj][kk] + wx2*wy1*wz2*data[ii+1][jj][kk+1];
	if (jj < dim[1]-1) {
	  line[k] += wx1*wy2*wz1*data[ii][jj+1][kk] + wx1*wy2*wz2*data[ii][jj+1][kk+1] +
	    wx2*wy2*wz1*data[ii+1][jj+1][kk] + wx2*wy2*wz2*data[ii+1][jj+1][kk+1];
	}
      }
      else if (jj < dim[1]-1) {
	line[k] += wx1*wy2*wz1*data[ii][jj+1][kk] + wx1*wy2*wz2*data[ii][jj+1][kk+1];
      }
      line[k] = line[k]*scale + offset;
    }
    break;
  default:
    cerr << "Invalid dimension" << endl;
    break;
  }
}

//
// Get 1-D of data (used for grids)
//
bldata *DataSet::Get1D(int d) {
  bldata *vals;

  switch (d) {
  case 1: 
    vals = new bldata[dim[0]];
    for (int i=0;i<dim[0];i++) {
      vals[i] = data[i][0][0];
    }
    break;
  case 2: 
    vals = new bldata[dim[1]];
    for (int i=0;i<dim[1];i++) {
      vals[i] = data[0][i][0];
    }
    break;
  case 3: 
    vals = new bldata[dim[2]];
    for (int i=0;i<dim[2];i++) {
      vals[i] = data[0][0][i];
    }
    break;
  default:
    cerr << "Invalid dimension";
    vals = NULL;
  }

  return vals;
}


// Get 2-D of data (used for orography)
//
bldata **DataSet::Get2D(int d) {
  bldata **vals;

  int nx = dim[0];
  int ny = dim[1];
  int nz = dim[2];

  switch (d) {

  case 1: 
    vals = AllocArray<bldata>(ny,nz);
    for (int i=0;i<ny;i++) {
      for (int j=0;j<nz;j++)
      vals[i][j] = data[0][i][j];
    }
    break;

  case 2: 
    vals = AllocArray<bldata>(nx,nz);
    for (int i=0;i<nx;i++) {
      for (int j=0;j<nz;j++)
      vals[i][j] = data[i][0][j];
    }
    break;

  case 3: 
    vals = AllocArray<bldata>(nx,ny);
    for (int i=0;i<nx;i++) {
      for (int j=0;j<ny;j++)
      vals[i][j] = data[i][j][0];
    }
    break;

  default:
    cerr << "Invalid dimension";
    vals = NULL;
  }

  return vals;
}

//
// Get the weightings associated with a particular grid value
//
void DataSet::GetGridWeight(bldata *grid,int n,bldata v,int & index,
			    bldata & w1,bldata & w2, bldata h) {
  
  bldata val;

  val = (v-h)/(1.0-h/grid[n-1]);

  // Check if this is 1-D
  if (n <= 3) {
    index = 0;
    w1 = 1.0;
    w2 = 0.0;
    //    cout << "Only 1 dimension" << endl;
    return;    
  }

  // Check if out of range
  if ((val < grid[0]) || (val > grid[n-1])) {
    index = 0;
    w1 = 0.0;
    w2 = 0.0;
//     cout << "Value out of range " << val << " " << grid[0] << " " << grid[n-1] << endl;
//     cout << "n = " << n << endl;
//     for (int i=0;i<n;i++) {
//       cout << grid[i] << " ";
//     }
//     cout << endl;
    return;
  }

  
  // Find nearest grid index to point
  index = 0;
  while (val > grid[index]) {
    index++;
  }

  if (index != 0) index--;

  w2 = (val - grid[index]) / (grid[index+1]-grid[index]);
  w1 = (grid[index+1] - val) / (grid[index+1]-grid[index]);
  
}

void DataSet::GetDataWeight(int i, int j, bldata wx1, bldata wx2, bldata wy1, 
			    bldata wy2, bldata val, int & index,
			    bldata & wz1, bldata & wz2) {

  bldata dval1, dval2;

  // Find nearest grid index to point
  index = 0;
  dval1 = 0;
  dval2 = wx1*wy1*data[i][j][0];
  if (i < dim[0]-1) {
    dval2 += wx2*wy1*data[i+1][j][0];
    if (j < dim[1]-1) {
      dval2 += wx2*wy2*data[i+1][j+1][0];
    }
  }
  if (j < dim[1]-1) {
    dval2 += wx1*wy2*data[i][j+1][0];
  }
  while ((val > dval2) && (index < dim[2]-1)) {
    index++;
    dval1 = dval2;
    dval2 = wx1*wy1*data[i][j][index];
    if (i < dim[0]-1) {
      dval2 += wx2*wy1*data[i+1][j][index];
      if (j < dim[1]-1) {
	dval2 += wx2*wy2*data[i+1][j+1][index];
      }
    }
    if (j < dim[1]-1) {
      dval2 += wx1*wy2*data[i][j+1][index];
    }
  }

  if (val > dval2) {
    wz1 = 0;
    wz2 = 0;
    index = dim[2]-1;
    return ;
  }
  if (index == 0) {
    //    cout << val << " " << dval2 << " " << index << endl;
    if (val < dval2) {
      wz1 = 0;
      wz2 = 0;
      return ;
    }
    else {
      wz1 = 1.0;
      wz2 = 0.0;
      return;
    }
  }
  else {
    index--;
  }

  wz2 = (val - dval1) / (dval2-dval1);
  wz1 = (dval2 - val) / (dval2-dval1);
  
}

//
// Get the dimensions of the data
//
int DataSet::GetDim(int i) {
  if ((i<1) || (i>3)) {
    cerr << "Dimension out of range" << endl;
    return 0;
  }
  return dim[i-1];
}

//
// Get the dimensions of the data
//
int DataSet::GetDims() {
  return ndim;
}

//
// Get the raw data 
//
bldata ***DataSet::GetData() {
  return data;
}

//
// Get the sizes of the grid
//

bldata DataSet::GetGridMin(int d) {
  bldata min, max;
  switch (d) {
  case 1:
      min = xgrid[0];
      max = xgrid[dim[0]-1];
    break;
  case 2:
      min = ygrid[0];
      max = ygrid[dim[1]-1];
    break;
  case 3:
      min = GetMin(orog,dim[0],dim[1]);
      max = zgrid[dim[2]-1];
      
    break;
  default:
    cerr << "Invalid dimension" << endl;
    min = 0;
    max = 0;
  }
  return (min<max)?min:max;
}

//
// Get the maximum of the grid
//
bldata DataSet::GetGridMax(int d) {
  bldata min, max;
  switch (d) {
  case 1:
      min = xgrid[0];
      max = xgrid[dim[0]-1];
    break;
  case 2:
      min = ygrid[1];
      max = ygrid[dim[1]-1];
    break;
  case 3:
      min = zgrid[2];
      max = zgrid[dim[2]-1];
    break;
  default:
    cerr << "Invalid dimension" << endl;
    min = 0;
    max = 0;
  }
  return (min>max)?min:max;
}

//
// Convert temperatures to Centigrade.
//
void DataSet::ConvTemp(bldata theta0, bldata g, bool deriv) {

  if (deriv) {
    for (int k=0;k<dim[2];k++) {
      for (int j=0;j<dim[1];j++) {
	for (int i=0;i<dim[0];i++) {
	  data[i][j][k] = data[i][j][k] * theta0/g;
	}
      }
    }
  }
  else {
    for (int k=0;k<dim[2];k++) {
      for (int j=0;j<dim[1];j++) {
	for (int i=0;i<dim[0];i++) {
	  data[i][j][k] = data[i][j][k] * theta0/g + theta0;
	}
      }
    }
  }
}


//
// Kludge to fix the orography where the x dimension is 1.
//

void DataSet::SwapData() {
  bldata *** ptr;
  int nx, ny, nz;

  nx = 1;
  ny = dim[0];
  nz = 1;

  dim[0] = nx;
  dim[1] = ny;
  dim[2] = nz;
  ndim = 2;

  ptr = (bldata ***) new bldata[nx][ny][nz];
    for (int i = 0;i<nx;i++) {
      ptr[i] = (bldata **) new bldata[ny][nz];
      for (int j=0;j<ny;j++) {
	ptr[i][j] = new bldata[nz];
      }
    }

    // Copy data across 
    for (int i = 0; i< ny; i++) {
      ptr[0][i][0] = data[i][0][0];
	
    } 

    // Delete old data
    Free3DArray(data);

    data = ptr;
}

//
// If this is a vector say which dimsension it is
//
void DataSet::SetIsVector(int i) {
  isvector = i;
}

//
// Change the sign of the data to swap axes?
//
int DataSet::GetIsVector() {
  return isvector;
}

//
// Get name of data set
//
char * DataSet::GetName() {
  return name;
}

//
// Set name of data set
//
void DataSet::SetName(const char *n) {
  name[19] = '\0';
  strncpy(name,n,19);
}

//
//  Get the grad of a scalar field
//
DataVect *DataSet::Grad() {
  bldata ***datax, ***datay, ***dataz;
  int nx, ny, nz;
  int i1, i2, j1, j2, k1, k2;
  bldata dx, dy, dz;
  bldata cx1, cx2, cy1, cy2;
  bldata ghalf, ghalf1, ghalf2;

  nx = dim[0];
  ny = dim[1];
  nz = dim[2];
  DataVect *grad = new DataVect(ndim, nx, ny, nz);

  datax = grad->GetX()->data;
  datay = grad->GetY()->data;
  dataz = grad->GetZ()->data;  
  
  for (int k = 0; k < nz; k++) {
    k1 = k-1;
    k2 = k+1;
    if (k1 == -1)
      k1 = 0;
    if (k2 == nz)
      k2 = nz-1;
    dz = zgrid[k2] - zgrid[k1];
    for (int j = 0; j < ny; j++) {
	for (int i = 0; i < nx; i++) {
	ghalf = 1.0 - orog[i][j] / H;
	if (nx == 1) {
	  datax[i][j][k] = 0;
	}
	else {
	  i2 = i+1;
	  i1 = i-1;
	  if (i == 0) {
	    i1 = i;
	  }	
	  if (i == nx-1) {
	    i2 = i;
	  }
	  dx = xgrid[i2]-xgrid[i1];
	  ghalf1 = 1.0 - orog[i1][j] / H;
	  ghalf2 = 1.0 - orog[i2][j] / H;
	  cx1 = -(1.0 - zgrid[k1]/H) *(orog[i2][j] - orog[i1][j])/dx;
	  cx2 = -(1.0 - zgrid[k2]/H) *(orog[i2][j] - orog[i1][j])/dx;
	  datax[i][j][k] = (ghalf2*data[i2][j][k]- ghalf1*data[i1][j][k]) / dx +
	  (cx2*data[i][j][k2] - cx1*data[i][j][k1]) / dz;
	  datax[i][j][k] /= ghalf;
	}
	if (ny == 1) {
	  datay[i][j][k] = 0;
	}
	else {
	  j2 = j+1;
	  j1 = j-1;
	  if (j == 0) {
	    j1 = j;
	  }	
	  if (j == ny-1) {
	    j2 = j;
	  }
	  dy = ygrid[j2]-ygrid[j1];
	  ghalf1 = 1.0 - orog[i][j1] / H;
	  ghalf2 = 1.0 - orog[i][j2] / H;
	  cy1 = -(1.0 - zgrid[k1]/H) *(orog[i][j2] - orog[i][j1])/dy;
	  cy2 = -(1.0 - zgrid[k2]/H) *(orog[i][j2] - orog[i][j1])/dy;
	  datay[i][j][k] = (ghalf2*data[i][j2][k]- ghalf1*data[i][j1][k]) / dy +
	  (cy2*data[i][j][k2] - cy1*data[i][j][k1]) / dz;   	  
	  datay[i][j][k] /= ghalf;
	}
	dataz[i][j][k] = (data[i][j][k2]- data[i][j][k1]) / (dz * ghalf);
      }
    }
  }

  // Set up grids
  grad->GetX()->SetGrids(xgrid, ygrid, zgrid, nx, ny, nz, H, orog);
  grad->GetY()->SetGrids(xgrid, ygrid, zgrid, nx, ny, nz, H, orog);
  grad->GetZ()->SetGrids(xgrid, ygrid, zgrid, nx, ny, nz, H, orog);
  
  grad->GetX()->SetName("Grad X");
  grad->GetX()->SetName("Grad Y");
  grad->GetX()->SetName("Grad Z");

  return grad;
}

//
// Get the grid
//
bldata *DataSet::GetGrid(int i) {
  switch (i) {
  case 1:
    return xgrid;
  case 2:
    return ygrid;
  case 3:
    return zgrid;
  }
  
  cerr << "Invalid dimension" << endl; 
  return NULL;
}

//
// Get the orography
//
bldata **DataSet::GetOrog() {
  return orog;
}

//
// Get the value of the data at a specified point
//
bldata DataSet::GetVal(bldata x,bldata y,bldata z) {
  bldata h, wx1, wx2, wy1, wy2, wz1, wz2, res, tmp;
  int ii, jj, kk;

  // Get the grid weightings
  GetGridWeight(xgrid,dim[0],x,ii,wx1,wx2,0.0); 
  GetGridWeight(ygrid,dim[1],y,jj,wy1,wy2,0.0); 
  h = wx1*wy1*orog[ii][jj];
  if (ii < dim[0]-1) {
    h += wx2*wy1*orog[ii+1][jj];
    if (jj < dim[1]-1) {
      h += wx1*wy2*orog[ii][jj+1] + wx2*wy2*orog[ii+1][jj+1];
    }
  }
  else if (jj < dim[1]-1) {
    h += wx1*wy2*orog[ii][jj+1];
  }
  GetGridWeight(zgrid,dim[2],z,kk,wz1,wz2,h);

  // Check if we are out of bounds in any direction
  // if so the w?1+w?2 < 1
  if (wx1+wx2+wy1+wy2+wz1+wz2 < 2.95) {
    return -9999.0;
  }

  // Add up the contributions from grid points ensuring no array overruns.
  res = wx1*wy1*wz1*data[ii][jj][kk];
  tmp = wx2*wy1*wz1;
  if (tmp > 0) res += wx2*wy1*wz1*data[ii+1][jj][kk];
  tmp = wx1*wy2*wz1;
  if (tmp > 0) res += wx1*wy2*wz1*data[ii][jj+1][kk];
  tmp = wx2*wy2*wz1;
  if (tmp > 0) res += wx2*wy2*wz1*data[ii+1][jj+1][kk];
  tmp = wx1*wy1*wz2;
  if (tmp > 0) res += wx1*wy1*wz2*data[ii][jj][kk+1];
  tmp = wx2*wy1*wz1;
  if (tmp > 0) res += wx2*wy1*wz2*data[ii+1][jj][kk+1];
  tmp = wx1*wy2*wz1;
  if (tmp > 0) res += wx1*wy2*wz2*data[ii][jj+1][kk+1];
  tmp = wx2*wy2*wz1;
  if (tmp > 0) res += wx2*wy2*wz2*data[ii+1][jj+1][kk+1];
  
  return res;

}

// Do 1-D interpolation using given values
// Based on polint from Numerical Recipes
bldata DataSet::Get1DInterp(bldata x, int n, bldata *xa, bldata *ya) {
  bldata diff, tmpdiff;
  bldata c[n], d[n], y;
  bldata ho, hp, w, den, dy;
  int ns;

  diff = fabs(x-xa[0]);
  ns = 0;

  for (int i=0; i<n; i++) {
    if ( (tmpdiff = fabs(x-xa[i-1])) < diff) {
      ns = i;
      diff = tmpdiff;
    }
    c[i] = ya[i];
    d[i] = ya[i];
  }

  y = ya[(ns--)];

  for (int m=1; m<n; m++) {
    for (int i=0; i<n-m; i++) {
      ho = xa[i] - x;
      hp = xa[i+m] - x;
      w = c[i+1] - d[i];
      if ( (den=ho-hp) == 0.0) {
	cerr << "Error interpolating" << endl;
	//	cerr << "n=" << n << " i=" << i << " m=" << m << "\n";
	//	for (int ii=0;ii<n;ii++) {
	//	  cerr << "x[" << i << "]=" << xa[i] << " ";
	//	}       
	//	cerr << endl;
	return 0.0;
      }
      den = w/den;
      d[i] = hp*den;
      c[i] = ho*den;
    }
    y += (dy = (2*(ns+1) < (n-m) ? c[ns+1] : d[(ns--)] ));
  }

  return y;
  
}

//
// Get the value of the data at a specified point
// Uses the 1-D interpolation routine Get1DInterp().
//
bldata DataSet::GetVal2(bldata x,bldata y,bldata z, bool model) {
  
  int xn, yn, zn, xi, yi, zi;
  int ximin, yimin, zimin;
  bldata tmp1[4][4], tmp2[4], tmpdata[4];
  bldata height, ans, zmodel;

  xi = GetIndex(xgrid, dim[0], x);
  yi = GetIndex(ygrid, dim[1], y);

  /* Check if outsize the domain */
  if ( (xi == -1) || (yi == -1) ) {
    //    cout << "Outside horizontal domain in GetVal2()" << xi << " " << yi << endl;
    //    cout << dim[0] << " " << x << endl;
    //    cout << dim[1] << " " << y << endl;
    throw out_of_range("Outside horizontal domain in GetVal2()");
    return -9999.0;
  }

  xn = GetOrder(xi, dim[0], ximin);
  yn = GetOrder(yi, dim[1], yimin);

  for (int i=0; i<xn; i++) {
    for (int j = 0; j<yn; j++) {
      tmpdata[j] = orog[ximin+i][yimin+j];
    }
    tmp2[i] = Get1DInterp( y, yn, &ygrid[yimin], tmpdata);
  }
  
  if (model) {
    zmodel = z;
  }
  else {
    height = Get1DInterp(x, xn, &xgrid[ximin], tmp2);    
    zmodel = (z-height)/(1-height/H);
  }
  zi = GetIndex(zgrid, dim[2], zmodel );
  if (zi == -1) {
    throw out_of_range("Outside domain vertical domain in GetVal2()");
    return -9999.0;
  }
  zn = GetOrder(zi, dim[2], zimin);

  for (int i = 0; i<xn; i++) {
    for (int j = 0; j<yn; j++) {
      for (int k = 0; k<zn; k++) {
	tmpdata[k] = data[ximin+i][yimin+j][zimin+k];
      }
      tmp1[i][j] = Get1DInterp( zmodel, zn, &zgrid[zimin], tmpdata);
    }
  }
  
  for (int i=0; i<xn; i++) {
    for (int j = 0; j<yn; j++) {
      tmpdata[j] = tmp1[i][j];
    }
    tmp2[i] = Get1DInterp( y, yn, &ygrid[yimin], tmpdata);
  }
  
  ans = Get1DInterp(x, xn, &xgrid[ximin], tmp2);

  return ans;
}


//
// Get the index of a specific value 
//
int DataSet::GetIndex(bldata *grid, int size, bldata x) {

  int xi;
  
  // Check if this is just a periodic run.
  if (size <= 3) {
    return 0;
  }

  xi = 0;
  while ( (xi < size) && (grid[xi] < x) ) {
    xi++;
  }
  if (xi == 0) {
    /* x is too small */
    return -1;
  }
  if (xi == size) {
    /* x is too large */
    return -1;
  }
  xi--;

  return xi;
}

//
// Get the order of the interpolation to use
// by checking how near the edge we are.
//
int DataSet::GetOrder(int xi, int size, int &min) {

  if (size == 1) {
    min = xi;
    return 1;
  }

  //  min = xi;
  //  return 2;

    if ((xi == 0) || (xi == size-2)) {
    min = xi;
    return 2;
    }
  if (xi == 0) {
    min = xi;
    return 3;
  }
  if (xi == size-2) {
    min = xi-1;
    return 3;
  }

  min = xi-1;
  return 4;
}

//
// Return new set with data normalised on the upstream value
// 
DataSet * DataSet::Normalise() {
  bldata ***normData;
  DataSet *newSet;

  normData = Alloc3DArray<bldata>(dim[0],dim[1],dim[2]);

  for (int k=0;k<dim[2];k++) {
    for (int j=0;j<dim[1];j++) {
      normData[0][j][k] = 1.0;
      for (int i=0;i<dim[0];i++) {
	if (fabs(data[0][j][k]) > 1E-10) {
	  normData[i][j][k] = data[i][j][k]/data[0][j][k];
	}
	else {
	  normData[i][j][k] = 0;
	}
      }
    }
  }

  newSet = new DataSet(normData,ndim,dim[0],dim[1],dim[2],xgrid,ygrid,zgrid,
		       H,orog);
  
  newSet->SetName("Normalised");
  
  return newSet;

}

DataSet * DataSet::copy() {
  bldata ***dptr;
  DataSet *tmp;

  dptr = Alloc3DArray<bldata>(dim[0],dim[1],dim[2]);
  for (int k=0;k<dim[2];k++) {
    for (int j=0;j<dim[1];j++) {
      for (int i=0;i<dim[0];i++) {
	dptr[i][j][k] = data[i][j][k];
      }
    }
  }

  tmp = new DataSet(dptr,ndim,dim[0],dim[1],dim[2],xgrid,ygrid,zgrid,H,orog);
  tmp->SetName("");
  return tmp;
}
