// 
// blasvis++ - C++ version of graphics routines
//
// DataVect : Class to encapsulate an vector set of DataSet objects
//            and perform arithemtic on them.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#include "datavect.h"
#include <iostream>
//#include <fstream>
//#include <cstring>
//#include <cmath>

//
// Default constructor
//
DataVect::DataVect() {
  datax = NULL;
  datay = NULL;
  dataz = NULL;
  cached = false;
}

//
// Constructor for an empty vector field
//
DataVect::DataVect(int d, int nx, int ny, int nz) {
  datax = new DataSet(d, nx, ny, nz);
  datay = new DataSet(d, nx, ny, nz);
  dataz = new DataSet(d, nx, ny, nz);
  datax->SetIsVector(1);
  datay->SetIsVector(2);
  dataz->SetIsVector(3);
  cached = false;
}

//
// Constructor for a vector field from existing data
//
DataVect::DataVect(DataSet *dx, DataSet *dy, DataSet *dz) {
  datax = dx;
  datay = dy;
  dataz = dz;
  cached = false;
}

//
// Destructor
//
DataVect::~DataVect() {
}

//
// Clear up routine
//
void DataVect::Delete() {
  delete datax;
  delete datay;
  delete dataz;
  datax = NULL;
  datay = NULL;
  dataz = NULL;
}


//
// Get components of vector
//
DataSet *DataVect::GetX() { return datax;}
DataSet *DataVect::GetY() { return datay;}
DataSet *DataVect::GetZ() { return dataz;}

//
// Work out the cross product of two vectors
//
DataVect *DataVect::Cross(DataVect *data2) {
  int ndim, nx, ny, nz;
  bldata ***dx, ***dy, ***dz;
  bldata ***data1x, ***data1y, ***data1z;
  bldata ***data2x, ***data2y, ***data2z;
  
  ndim = datax->ndim;
  nx = datax->dim[0];
  ny = datax->dim[1];
  nz = datax->dim[2];
  DataVect *cross = new DataVect(ndim, nx, ny, nz);

  dx = cross->GetX()->data;
  dy = cross->GetY()->data;
  dz = cross->GetZ()->data;

  data1x = GetX()->data;
  data1y = GetY()->data;
  data1z = GetZ()->data;
  
  data2x = data2->GetX()->data;
  data2y = data2->GetY()->data;
  data2z = data2->GetZ()->data;

  for (int k = 0; k < nz; k++) {
    for (int j = 0; j < ny; j++) {
      for (int i = 0; i < nx; i++) {
	dx[i][j][k] = data1y[i][j][k] * data2z[i][j][k] - 
	  data1z[i][j][k] * data2y[i][j][k];
	dy[i][j][k] = data1z[i][j][k] * data2x[i][j][k] - 
	  data1x[i][j][k] * data2z[i][j][k];
	dz[i][j][k] = data1x[i][j][k] * data2y[i][j][k] - 
	  data1y[i][j][k] * data2x[i][j][k];
      }
    }
  }

  // Set up grids
  cross->GetX()->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);
  cross->GetY()->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);
  cross->GetZ()->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);

  cross->GetX()->SetName("Cross X");
  cross->GetY()->SetName("Cross Y");
  cross->GetZ()->SetName("Cross Z");

  return cross;
}


// Work out the dot product of two vectors
//
DataSet *DataVect::Dot(DataVect *data2) {
  int ndim, nx, ny, nz;
  bldata ***data;
  bldata ***data1x, ***data1y, ***data1z;
  bldata ***data2x, ***data2y, ***data2z;
  
  ndim = datax->ndim;
  nx = datax->dim[0];
  ny = datax->dim[1];
  nz = datax->dim[2];
  DataSet *product = new DataSet(ndim, nx, ny, nz);

  data = product->data;

  data1x = GetX()->data;
  data1y = GetY()->data;
  data1z = GetZ()->data;
  
  data2x = data2->GetX()->data;
  data2y = data2->GetY()->data;
  data2z = data2->GetZ()->data;

  for (int k = 0; k < nz; k++) {
    for (int j = 0; j < ny; j++) {
      for (int i = 0; i < nx; i++) {
	data[i][j][k] = data1x[i][j][k] * data2x[i][j][k] +
	  data1y[i][j][k] * data2y[i][j][k] +
	  data1z[i][j][k] * data2z[i][j][k];
      }
    }
  }

  // Set up grids
  product->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);

  product->SetName("Dot product");

  return product;
}

//
// Work out the divergence of a vector
//
DataSet *DataVect::Div(Grid *grids) {
  int ndim, nx, ny, nz;
  int ndim2,nx2, ny2, nz2;
  int ix,iy,iz,jx,jy,jz,kx,ky,kz;
  bldata wx1, wx2,wy1,wy2,wz1,wz2;
  bldata g1, g2;
  bldata *xgrid2, *ygrid2, *zgrid2;
  bldata ***data;
  bldata ***data1x, ***data1y, ***data1z;
  
  ndim = datax->ndim;
  nx = datax->dim[0];
  ny = datax->dim[1];
  nz = datax->dim[2];

  ndim2 = grids->ndim;
  nx2 = grids->nx;
  ny2 = grids->ny;
  nz2 = grids->nz;
  xgrid2 = grids->xgrid;
  ygrid2 = grids->ygrid;
  zgrid2 = grids->zgrid;

  DataSet *product = new DataSet(ndim2, nx2, ny2, nz2);

  data = product->data;

  data1x = GetX()->data;
  data1y = GetY()->data;
  data1z = GetZ()->data;
  
  kx = 0;
  ky = 0;
  kz = 0;
  for (int k = 0; k < nz2; k++) {
    // Get indices for each of the three vector components
    while ( (kx < nz) && (0.5*(datax->zgrid[kx]+datax->zgrid[kx+1]) 
			    < zgrid2[k]) ) kx++;
    if (kx < nz) kx--;
    while ( (ky < nz) && (0.5*(datay->zgrid[ky]+datay->zgrid[ky+1]) 
			    < zgrid2[k]) ) ky++;
    if (ky < nz) ky--;
    while ( (kz < nz) && (0.5*(dataz->zgrid[kz]+dataz->zgrid[kz+1]) 
			    < zgrid2[k]) ) kz++;
    if (kz < nz) kz--;

    // Work out weightings for each component
    g1 = 0.5*(dataz->zgrid[kz]+dataz->zgrid[kz+1]);
    g2 = 0.5*(dataz->zgrid[kz+1]+dataz->zgrid[kz+2]);
    wz1 = (g2-zgrid2[k])/(g2-g1);
    wz2 = (zgrid2[k]-g1)/(g2-g1);

    jx = 0;
    jy = 0;
    jz = 0;
    for (int j = 0; j < ny2; j++) {
      while ( (jx < ny) && (0.5*(datax->ygrid[jx]+datax->ygrid[jx+1]) 
			    < ygrid2[j]) ) jx++;
      if (jx < ny) jx--;
      while ( (jy < ny) && (0.5*(datay->ygrid[jy]+datay->ygrid[jy+1]) 
			    < ygrid2[j]) ) jy++;
      if (jy < ny) jy--;
      while ( (jz < ny) && (0.5*(dataz->ygrid[jz]+dataz->ygrid[jz+1]) 
			    < ygrid2[j]) ) jz++;
      if (jz < ny) jz--;

      // Work out weightings for each component
      g1 = 0.5*(datay->ygrid[jy]+datay->ygrid[jy+1]);
      g2 = 0.5*(datay->ygrid[jy+1]+datay->ygrid[jy+2]);
      wy1 = (g2-ygrid2[j])/(g2-g1);
      wy2 = (ygrid2[j]-g1)/(g2-g1);

      ix = 0;
      iy = 0;
      iz = 0;
      for (int i = 0; i < nx2; i++) {
	while ( (ix < nx) && (0.5*(datax->xgrid[ix]+datax->xgrid[ix+1]) 
			      < xgrid2[i]) ) ix++;
	if (ix < nx) ix--;
	while ( (iy < nx) && (0.5*(datay->xgrid[iy]+datay->xgrid[iy+1]) 
			      < xgrid2[i]) ) iy++;
	if (iy < nx) iy--;
	while ( (iz < nx) && (0.5*(dataz->xgrid[iz]+dataz->xgrid[iz+1]) 
			      < xgrid2[i]) ) iz++;
	if (iz < nx) iz--;

	// Work out weightings for each component
	g1 = 0.5*(datax->xgrid[ix]+datax->xgrid[ix+1]);
	g2 = 0.5*(datax->xgrid[ix+1]+datax->xgrid[ix+2]);
	wx1 = (g2-xgrid2[i])/(g2-g1);
	wx2 = (xgrid2[i]-g1)/(g2-g1);
 
	data[i][j][k] = (data1x[ix+1][jx][kx]-data1x[ix][jx][kx])*wx1 +
	  (data1x[ix+2][jx][kx]-data1x[ix+1][jx][kx])*wx2 +
	  (data1y[ix][jx+1][kx]-data1y[ix][jx][kx])*wy1 +
	  (data1y[ix][jx+2][kx]-data1y[ix][jx+1][kx])*wy2 +
	  (data1z[ix][jx][kx+1]-data1z[ix][jx][kx])*wz1 +
	  (data1z[ix][jx][kx+2]-data1z[ix][jx][kx+1])*wz2;
      }
    }
  }

  // Set up grids
  product->SetGrids(xgrid2, ygrid2, zgrid2, nx, ny, nz, GetZ()->H, GetZ()->orog);

  product->SetName("Dot product");

  return product;
}

//
// Work out the curl of a vector
//
DataVect *DataVect::Curl() {
  int ndim, nx, ny, nz;
  bldata ***dx, ***dy, ***dz;
  bldata ***data1x, ***data1y, ***data1z;
  
  ndim = datax->ndim;
  nx = datax->dim[0];
  ny = datax->dim[1];
  nz = datax->dim[2];
  DataVect *curl = new DataVect(ndim, nx, ny, nz);

  dx = curl->GetX()->data;
  dy = curl->GetY()->data;
  dz = curl->GetZ()->data;

  data1x = GetX()->data;
  data1y = GetY()->data;
  data1z = GetZ()->data;
  
  for (int k = 0; k < nz; k++) {
    for (int j = 0; j < ny; j++) {
      for (int i = 0; i < nx; i++) {
	dx[i][j][k] = data1y[i][j][k] - 
	  data1z[i][j][k];
	dy[i][j][k] = data1z[i][j][k] - 
	  data1x[i][j][k];
	dz[i][j][k] = data1x[i][j][k] - 
	  data1y[i][j][k];
      }
    }
  }

  // Set up grids
  curl->GetX()->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);
  curl->GetY()->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);
  curl->GetZ()->SetGrids(GetX()->xgrid, GetX()->ygrid, GetX()->zgrid, nx, ny, nz, GetX()->H, GetX()->orog);

  curl->GetX()->SetName("Curl X");
  curl->GetY()->SetName("Curl Y");
  curl->GetZ()->SetName("Curl Z");

  return curl;
}

