//
// blasvis++ - C++ version of graphics routines
//
// Readfile : Abstract Class to read the information in from a 
//            data file.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#include "readfile.h"
#include "models.h"
#include <iostream>
#include <fstream>
#include <cctype>
#include <cstring>
#include <cmath>
#include <stdexcept>

using namespace std;

// Default constructor
ReadFile::ReadFile() {

  Clear();

  endian = false;

  nullgrid = new bldata[1];
  nullgrid[0] = 0;
}

// Constructor with a file name
ReadFile::ReadFile(const char* fname) {
  Clear();
  OpenFile(fname);

  nullgrid = new bldata[1];
  nullgrid[0] = 0;
}

// destructor
ReadFile::~ReadFile() {
  CloseFile();
  delete[] nullgrid;
}

void ReadFile::Clear() {
  xgrid = NULL;
  xngrid = NULL;
  ygrid = NULL;
  yngrid = NULL;
  zgrid = NULL;
  zngrid = NULL;
  orog = NULL;

  pv = NULL;
  pvalt = NULL;
  pvflux = NULL;
  pvfluxalt = NULL;
  bernoulli = NULL;
  windspeed = NULL;
  speedup = NULL;
  deltaT = NULL;
  pcorr = NULL;
  timestep = NULL;
  bldepth = NULL;
  streamfn = NULL;
  gradt = NULL;
  PVsrc = NULL;
  height = NULL;
  turbdiv = NULL;
  uw_res = NULL;
  uu_res = NULL;
  ww_res = NULL;
  e_res = NULL;

  firstItem = NULL;
  lastItem = NULL;
  numitems = 0;
  H = 0;

}


//
// Open the named file and read in the header
//
bool ReadFile::OpenFile(const char* fname) {
//   int rnuml = sizeof(long);
//   char crec[rnuml];
//   char tmp;

  if (IsOpen()) {
    CloseFile();
  }

  // Open file
  file.open(fname,ios::in | ios::binary);

  if (IsOpen()) {
    start = file.tellg();
    if (file.fail()) {
      cerr << "Failure reading file" << endl;
      return false;
    }
//     for (int i=0;i<rnuml;i++) {
//       file >> crec[i];
//     }
    if (norec) {
      irecl = 0;
      reclen = 0;
    }
    else {
      irecl = BinRead<long>();
      reclen = sizeof(long);
    }
    InitFile();
  }
  else {
    cerr << "Error opening file" << endl;
    return false;
  }
  return true;
}

//
// Check if a file is open
//
bool ReadFile::IsOpen() {
  return file.is_open();
}

//
// Output the header list to the screen for debugging purposes.
//
void ReadFile::DisplayHeader() {
  HeaderItem *ptr;

  ptr = firstItem;

  while (ptr != NULL) {
    cout << ptr->name << " " << ptr->ndim << " " << 
      ptr->dim[0] << " " << ptr->dim[1] << " " << ptr->dim[2] << endl;
    ptr = ptr->next;
  }

  //  cout << "Done writing headers" << endl;

}

//
// Get the header item corresponding to a data set
//
HeaderItem * ReadFile::GetHeaderItem(DataSet *data) {
  HeaderItem *ptr;

  ptr = firstItem;

  while ((ptr != NULL) && (ptr->data != data)) {
    ptr = ptr->next;
  }

  if (ptr == NULL) {
    cerr << "Dataset not found." << endl;
    return NULL;
  }

  return ptr;

}


//
// Get the header item corresponding to a name
//
HeaderItem * ReadFile::GetHeaderItem(const char *dname) {
  HeaderItem *ptr;

  ptr = firstItem;

  while ((ptr != NULL) && (strncmp(dname, ptr->name,8))) {
    ptr = ptr->next;
  }

  if (ptr == NULL) {
    cerr << "Dataset not found." << endl;
    return NULL;
  }

  return ptr;

}

//
//
//
HeaderItem *ReadFile::GetFirstHeaderItem() {
  return firstItem;
}

//
// Load in the datasets for PV and calculate it.
//
DataSet *ReadFile::GetPV()
{
  DataSet *vortxd, *vortyd, *vortzd, *tempd, *rhod;
  bldata ***vortx, ***vorty, ***vortz, ***temp, ***rho, ***data;
  bldata **oro;
  bldata fcor;

  int ndim, nx, ny, nz, imin, imax, jmin, jmax;
  bldata cx, cy, dx, dy, dz, ghalf, ghalf1, ghalf2, orog1, orog2, orog0;
  bldata kx, ky, kz, kxr, kyr, kzr, wx1, wx2, wy1, wy2, wz1, wz2, vort;


  if ( pv == NULL ) {
    fcor = GetParm("FCOR");

    tempd = GetData("T");

    if (tempd == NULL) {
      cerr << "Not a stratified flow" << endl;
      return pv;
    }

    rhod = GetData("RHO");

    vortxd = GetData("VORTX");  
    vortyd = GetData("VORTY");
    vortzd = GetData("VORTZ");

    if ( !vortxd || !vortyd || !vortzd || !rhod) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    vortx = vortxd->GetData();
    vorty = vortyd->GetData();
    vortz = vortzd->GetData();
    temp = tempd->GetData();
    rho = rhod->GetData();
    oro = orog->Get2D(3);

    ndim = tempd->GetDims();
    nx = tempd->GetDim(1);
    ny = tempd->GetDim(2);
    nz = tempd->GetDim(3);

    if (nz <= 4) {
      cerr << "Need a thicker boundary layer" << endl;
      return pv;
    }
    
    if (nx == 1) {
      imin = 0;
      imax = 0;
    }
    else {
      imin = 1;
      imax = nx-2;
    }
    if (ny == 1) {
      jmin = 0;
      jmax = 0;
    }
    else {
      jmin = 1;
      jmax = ny-2;
    }
    
    pv = new DataSet(ndim, nx, ny, nz);
    data = pv->GetData();

    cx = 0;
    cy = 0;

    dy = 0;
    ky = 0;
    kyr = 0;
    wx1 = 0;
    wx2 = 0;
    wy1 = 0;
    wy2 = 0;

  // Calculate the PV.
    for (int k=1;k<nz-1;k++) {
      dz = zngrid[k+1]-zngrid[k-1];
      kz = (zngrid[k+1]-zngrid[k])/(zngrid[k]-zngrid[k-1]);
      kzr = 1.0/kz;
      wz1 = (zngrid[k]-zgrid[k-1])/(zgrid[k]-zgrid[k-1]);
      wz2 = 1.0-wz1;
      for (int j=jmin;j<=jmax;j++) {
	if (ny != 1) {
	  dy = yngrid[j+1]-yngrid[j-1];
	  ky = (yngrid[j+1]-yngrid[j])/(yngrid[j]-yngrid[j-1]);
	  kyr = 1.0/ky;
	  wy1 = (yngrid[j]-ygrid[j-1])/(ygrid[j]-ygrid[j-1]);
	  wy2 = 1.0-wy1;
	}
	for (int i=imin;i<=imax;i++) {
	  // Set data to be zero initially
	  data[i][j][k] = 0;
	
	  ghalf = 1.0 - oro[i][j] / H;
	  orog0 = oro[i][j];
	
	  // Check if there are derivatives in the X direction
	  // If so calculate X component of PV
	  if (nx != 1) {
	    dx = xngrid[i+1]-xngrid[i-1];
	    kx = (xngrid[i+1]-xngrid[i])/(xngrid[i]-xngrid[i-1]);
	    kxr = 1.0/kx;
	    wx1 = (xngrid[i]-xgrid[i-1])/(xgrid[i]-xgrid[i-1]);
	    wx2 = 1.0-wx1;
	  
	    orog1 = oro[i-1][j];
	    orog2 = oro[i+1][j];
	  
	    // Check nz-1!
	    cx = -((orog2-orog0)*kxr + (orog0-orog1)*kx)/dx*(1-zngrid[k]/H);
	    ghalf1 = 1.0-orog1/H;
	    ghalf2 = 1.0-orog2/H;
	    if (ny == 1) {
	      vort = wz1*vortx[i][j][k] + wz2*vortx[i][j][k-1];
	    }
	    else {
	      vort = wz1*wy1*vortx[i][j][k] + wz2*wy1*vortx[i][j][k-1] +
		wz1*wy2*vortx[i][j-1][k] + wz2*wy2*vortx[i][j-1][k-1];
	    }
	    data[i][j][k] += vort*( ( (temp[i+1][j][k]-temp[i][j][k])*kxr +
				      (temp[i][j][k]-temp[i-1][j][k])*kx )/dx +
				    ( (temp[i][j][k+1] - temp[i][j][k])*kzr + 
				      (temp[i][j][k] - temp[i][j][k-1])*kz )/dz*(cx/ghalf) );
	  }      
	  // Check if there are any derivatives in the Y direction
	  // If so calculate Y component of PV
	  if (ny != 1) {

	    orog1 = oro[i][j-1];
	    orog2 = oro[i][j+1];

	    cy = -((orog2-orog0)*kyr + (orog0-orog1)*ky)/dy*(1-zngrid[k]/H);
	    ghalf1 = 1.0-orog1/H;
	    ghalf2 = 1.0-orog2/H;
	    if (nx == 1) {
	      vort = wz1*vorty[i][j][k] + wz2*vorty[i][j][k-1];
	    }
	    else {
	      vort = wz1*wx1*vorty[i][j][k] + wz2*wx1*vorty[i][j][k-1] +
		wz1*wx2*vorty[i-1][j][k] + wz2*wx2*vorty[i-1][j][k-1];
	    }
	    data[i][j][k] += vort*
	      ( ( (temp[i][j+1][k]-temp[i][j][k])*kyr +
		  (temp[i][j][k]-temp[i][j-1][k])*ky )/dy +
		( (temp[i][j][k+1] - temp[i][j][k])*kzr + 
		  (temp[i][j][k] - temp[i][j][k-1])*kz )/dz*(cy/ghalf) );
	  }
	  // Calculate Z component of PV
	  if ((nx == 1) || (ny == 1)) {
	    if (nx == 1) {
	      vort = 0.5*(vortz[i][j][k] + vortz[i][j-1][k]);
	    }
	    else {
	      vort = 0.5*(vortz[i][j][k] + vortz[i-1][j][k]);
	    }
	  }
	  else {
	    vort = wx1*wy1*vortz[i][j][k] + wx1*wy2*vortz[i][j-1][k] +
	      wx2*wy1*vortz[i-1][j][k] + wx2*wy2*vortz[i-1][j-1][k];
	  }
	  data[i][j][k] += (vort + fcor)*
	  //  data[i][j][k] += vort*
	    ( (temp[i][j][k+1]-temp[i][j][k])*kzr + (temp[i][j][k]-temp[i][j][k-1])*kz ) 
	    / (dz * ghalf);
	  // Divide by the density.
	  data[i][j][k] /= rho[i][j][k];
	}	
      }
    }
    // Put zeros round the edges
    //  if ((ny > 1) && (nx > 1)) {
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	data[i][j][0] = 0.0;
	data[i][j][nz-1] = 0.0;
      }
    }
    //  }
    if (nx > 1) {
      for (int k=0;k<nz;k++) {
	for (int j=0;j<ny;j++) {
	  data[0][j][k] = 0.0;
	  data[nx-1][j][k] = 0.0;
	}
      }
    }
    if (ny > 1) {
      for (int k=0;k<nz;k++) {
	for (int i=0;i<nx;i++) {
	  data[i][0][k] = 0.0;
	  data[i][ny-1][k] = 0.0;
	}
      }
    }  

    pv->SetGrids(xngrid,yngrid,zngrid,nx,ny,nz,H, orog);
    pv->SetName("PV");

    FreeArray(oro);

  }
  return pv;
}

//
// Get the temperature gradient
//
DataVect *ReadFile::GetGradT() {
  DataSet *tempd;

  if (gradt == NULL) {
    tempd = GetData("T");

    if (tempd == NULL) {
      return NULL;
    }

    gradt = tempd->Grad();
  }
  gradt->GetX()->SetName("GradT X");
  gradt->GetY()->SetName("GradT Y");
  gradt->GetZ()->SetName("GradT Z");

  return gradt;
}

//
// Get the PV flux using generalised Bernoulli equation
//
DataSet *ReadFile::GetPVAlt() {
  DataSet *vortx, *vorty, *vortz;
  DataVect *vort;

  if (pvalt == NULL) {
    vortx = GetData("VORTX");
    vorty = GetData("VORTY");
    vortz = GetData("VORTZ");
    
    if (GetGradT() == NULL) {
      cerr << "Not a stratified flow" << endl;
      return pvalt;
    }

    if ( !vortx || !vorty || !vortz ) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    vort = new DataVect(vortx, vorty, vortz);

    pvalt = gradt->Dot(vort);
    pvalt->SetName("PV");

    delete vort;
  }
  return pvalt;
}

//
// Get the PV flux using generalised Bernoulli equation
//
DataVect *ReadFile::GetPVFlux() {
  DataVect *gradb;

  if (pvflux == NULL) {
    if (GetGradT() == NULL) {
      cerr << "Not a stratified flow" << endl;
      return NULL;
    }
    GetBernoulli();
    if ( !bernoulli ) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    gradb = bernoulli->Grad();
    pvflux = gradt->Cross(gradb);
    pvflux->GetX()->SetName("PV Flux X");
    pvflux->GetY()->SetName("PV Flux Y");
    pvflux->GetZ()->SetName("PV Flux Z");

    gradb->Delete();
    delete gradb;
  }
  return pvflux;
}

DataVect *ReadFile::GetPVFluxAlt() {
  DataSet *u, *v, *w;
  int ndim, dimx, dimy, dimz;
  bldata ***datax, ***datay, ***dataz, ***ud, ***vd, ***wd, ***pvd;
  
  bldata uu, vv, ww;

  if (pvfluxalt == NULL) {

    GetPV();

    ndim = pv->GetDims();
    dimx = pv->GetDim(1);
    dimy = pv->GetDim(2);
    dimz = pv->GetDim(3);
    
    u = GetData("U");
    v = GetData("V");
    w = GetData("W");
    if ( !u || !v || !w || !pv) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();
    pvd = pv->GetData();

    pvfluxalt = new DataVect(ndim, dimx, dimy, dimz);
    datax = pvfluxalt->GetX()->GetData();
    datay = pvfluxalt->GetY()->GetData();
    dataz = pvfluxalt->GetZ()->GetData();

    for (int i = 0;i<dimx;i++) {
      for (int j = 0;j<dimy;j++) {
	for (int k = 0;k<dimz;k++) {
	  uu = ud[i][j][k];
	  if ((dimx > 1) && (i<dimx-1)) {
	    uu = (uu+ud[i+1][j][k])*0.5;
	  }
	  vv = vd[i][j][k];
	  if ((dimy > 1)&&(j<dimy-1)) {
	    vv = (vv+vd[i][j+1][k])*0.5;
	  }
	  ww = wd[i][j][k];
	  if ((dimz > 1)&&(k<dimz-1)) {
	    ww = (ww+wd[i][j][k+1])*0.5;
	  }
	  datax[i][j][k] = uu*pvd[i][j][k]/ 1.0E6;
	  datay[i][j][k] = vv*pvd[i][j][k]/ 1.0E6;
	  dataz[i][j][k] = ww*pvd[i][j][k]/ 1.0E6;
	}
      }
    }

    pvfluxalt->GetX()->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
    pvfluxalt->GetY()->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
    pvfluxalt->GetZ()->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
    pvfluxalt->GetX()->SetName("PV Flux X");
    pvfluxalt->GetY()->SetName("PV Flux Y");
    pvfluxalt->GetZ()->SetName("PV Flux Z");

  }
  return pvfluxalt;
}

//
// Load in the datasets for Bernoulli and calculate it.
//
DataSet *ReadFile::GetBernoulli()
{
  DataSet *ud, *rhod, *pd;
  //DataSet *uid, *vid, *td;

  bldata ***u, ***p, ***rho, ***data;
  //bldata ***ui, ***vi, ***temp;

  // const bldata Cp = 1005.5;

  int ndim, dimx, dimy, dimz;

  if (bernoulli == NULL) {

    ud = GetWindspeed();
    //uid = GetData("UI");
    //vid = GetData("VI");
    rhod = GetData("RHO");
    pd = GetPressureCorr();
    //td = GetDeltaT();
    //    if ( !ud || !uid || !vid || !rhod || !pd || !td) {
    if ( !ud || !rhod || !pd) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    u = ud->GetData();
    // Background profile may not be available
    // If not the just ignore it
    //if (uid && vid && temp) {
    //  ui = uid->GetData();
    //  vi = vid->GetData();
    //  temp = td->GetData();
    //}
    p = pd->GetData();
    rho = rhod->GetData();
    
    ndim = ud->GetDims();
    dimx = ud->GetDim(1);
    dimy = ud->GetDim(2);
    dimz = ud->GetDim(3);
    
    bernoulli = new DataSet(ndim, dimx, dimy, dimz);
    data = bernoulli->GetData();

    for (int i = 0;i<dimx;i++) {
      for (int j = 0;j<dimy;j++) {
	for (int k = 0;k<dimz;k++) {
	  // Calculate perturbation to Bernoulli potential
//	  if (uid && vid && temp) {
//	    data[i][j][k] = 0.5*(u[i][j][k]*u[i][j][k] 
//				 - ui[k][0][0]*ui[k][0][0] 
//				 - vi[k][0][0]*vi[k][0][0])
//	      + Cp*temp[i][j][k];
//	  }
//	  else {
	    data[i][j][k] = 0.5*u[i][j][k]*u[i][j][k] + p[i][j][k] / 
	      rho[i][j][k];
//	  }
	}
      }
    }

    //    bernoulli = new DataSet();
    //    bernoulli->CalcBernoulli(ud, rhod, pd);

    bernoulli->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
    bernoulli->SetName("Bernoulli potential");
  }

  return bernoulli;
}


//
// Get the windspeed at a point i.e. (u^2 + v^2 + w^2)^(1/2)
//
DataSet *ReadFile::GetWindspeed() {
  DataSet *u, *v, *w;
  int ndim, dimx, dimy, dimz;
  bldata *** data, ***ud, ***vd, ***wd;
  
  bldata uu, vv, ww;


  if (windspeed == NULL) {
    u = GetData("U");
    v = GetData("V");
    w = GetData("W");
    if ( !u || !v || !w) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();

    ndim = u->GetDims();
    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    // Create an empty data set
    windspeed = new DataSet(ndim, dimx, dimy, dimz);

    data = windspeed->GetData();
    
    // interpolate components onto the (XN,YN,ZN) grid
    for (int i = 0;i<dimx;i++) {
      for (int j = 0;j<dimy;j++) {
	for (int k = 0;k<dimz;k++) {
	  uu = ud[i][j][k];
	  if ((dimx > 1) && (i<dimx-1)) {
	    uu = (uu+ud[i+1][j][k])*0.5;
	  }
	  vv = vd[i][j][k];
	  if ((dimy > 1)&&(j<dimy-1)) {
	    vv = (vv+vd[i][j+1][k])*0.5;
	  }
	  ww = wd[i][j][k];
	  if ((dimz > 1)&&(k<dimz-1)) {
	    ww = (ww+wd[i][j][k+1])*0.5;
	  }
	  data[i][j][k] = sqrt(uu*uu + vv*vv + ww*ww);
	}
      }
    }
    windspeed->SetName("Windspeed");
    windspeed->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
  }

  return windspeed;
  
}

//
// Get the streamfunction at a point
//
DataSet *ReadFile::GetStreamFn() {
  DataSet *u, *v;
  int ndim, dimx, dimy, dimz;
  bldata *** data, ***ud, ***vd;
  bldata **oro;
  
  bldata uu;

  if (streamfn == NULL) {
    if (useLES) {
      u = GetData("UAV");
      v = GetData("VAV");
    }
    else {
      u = GetData("U");
      v = GetData("V");
    }
    if ( !u || !v) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    ud = u->GetData();
    vd = v->GetData();

    ndim = u->GetDims();
    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    if ((dimx > 1) && (dimy > 1)) {
      cerr << "Streamfunction only supported in 2-D" << endl;
      return NULL;
    }

    oro = orog->Get2D(3);


    // Create an empty data set
    streamfn = new DataSet(ndim, dimx, dimy, dimz);

    data = streamfn->GetData();
    
    if (dimy == 1) {
      for (int i = 0;i<dimx;i++) {
	data[i][0][0] = 0.0;
	for (int k = 1;k<dimz;k++) {
	  uu = ud[i][0][k];	
	  data[i][0][k] = data[i][0][k-1] + 
	    (zgrid[k]-zgrid[k-1])*(1-oro[i][0]/zgrid[dimz-1])*uu;
	}
      }
      streamfn->SetGrids(xgrid,yngrid,zgrid,dimx,dimy,dimz,H, orog);
    }
    else if (dimx == 1) {
      for (int j = 0;j<dimy;j++) {
	data[0][j][0] = 0.0;
	for (int k = 1;k<dimz;k++) {
	  uu = vd[0][j][k];
	  data[0][j][k] = data[0][j][k-1] + 
	    (zgrid[k]-zgrid[k])*(1-oro[0][j]/zgrid[dimz-1])*uu;	
	}
      }
      streamfn->SetGrids(xngrid,ygrid,zgrid,dimx,dimy,dimz,H, orog);
    }
    streamfn->SetName("Streamfn");
    FreeArray(oro);
  }

  return streamfn;
  
}

//
// Get the CFL timestep at a point
//
DataSet *ReadFile::GetTimestep() {
  DataSet *u, *v, *w, *r, *vi, *vih;
  int ndim, dimx, dimy, dimz;
  bldata *** data, ***ud, ***vd, ***wd, ***rigd, ***vis, ***vish;
  
  bldata uu, vv, ww;
  bldata rpr, prandtl, busdyr, cdx, cdy, cdz, vs, vsh;

  if (timestep == NULL) {
    u = GetData("U");
    v = GetData("V");
    w = GetData("W");
    r = GetData("RIGRAD");
    vi = GetData("VIS");
    vih = GetData("VISH");

    // Use this for old data sets where VISH does not exist.
    if ((vih == NULL) || (vih->GetDims() < 3)) vih = vi;

    // Check for NULL pointers
    if ( !u || !v || !w || !vi) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();
    if (r) rigd = r->GetData();
    vis = vi->GetData();
    vish = vih->GetData();

    prandtl = GetParm("PR");
    rpr = ((prandtl<1.0)&&(prandtl>0.0))?1/prandtl:1;
	   
    busdyr = GetParm("BUSDYR");

    ndim = u->GetDims();
    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    // Create an empty data set
    timestep = new DataSet(ndim, dimx, dimy, dimz);

    data = timestep->GetData();
    cdx = 0;
    cdy = 0;
    cdz = 0;

    // interpolate components onto the (XN,YN,ZN) grid
    for (int i = 0;i<dimx;i++) {
      if (i != dimx-1) {
	cdx = 1/(xngrid[i+1]-xngrid[i]);
      }
      for (int j = 0;j<dimy;j++) {
	if (j != dimy-1) {
	  cdy = 1/(yngrid[j+1]-yngrid[j]);
	}
	for (int k = 0;k<dimz;k++) {
	  if (k!= dimz-1) {
	    cdz = 1/(zngrid[k+1]-zngrid[k]);
	  }
	  if ( r && (rigd[i][j][k] < 0.0)) {
	    vs = fabs(rpr*pow((1.0-busdyr*rpr*rigd[i][j][k]),0.25)*vis[i][j][k]);
	  }
	  else {
	    vs = fabs(rpr*vis[i][j][k]);
	  }
	  vsh = fabs(rpr*vish[i][j][k]);
	  uu = ud[i][j][k];
	  vv = vd[i][j][k];
	  ww = wd[i][j][k];
 	  data[i][j][k] = 0.25/(sqrt(uu*uu+2*vsh)*cdx + 
				sqrt(vv*vv+2*vsh)*cdy + 
 				sqrt(ww*ww+2*vs)*cdz + 
 				2*(vsh*(cdx*cdx+cdy*cdy) + vs*cdz*cdz));
	}
      }
    }
    timestep->SetName("CFL timestep");
    timestep->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
  }

  return timestep;
  
}

//
// Get the speedup over a hill i.e. u / u_0
//
DataSet *ReadFile::GetSpeedup() {
  DataSet *u, *v, *w;
  int ndim, dimx, dimy, dimz;
  bldata *** data, ***ud, ***vd, ***wd;
  
  bldata uu, vv, ww;


  if (speedup == NULL) {
    u = GetData("U");
    v = GetData("V");
    w = GetData("W");
    if (!u || !v || !w) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();

    ndim = u->GetDims();
    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    // Create an empty data set
    speedup = new DataSet(ndim, dimx, dimy, dimz);

    data = speedup->GetData();

    for (int k = 0;k<dimz;k++) {      
      for (int j = 0;j<dimy;j++) {
	for (int i = 0;i<dimx;i++) {
	  uu = ud[i][j][k];
	  if ((dimx > 1) && (i<dimx-1)) {
	    uu = (uu+ud[i+1][j][k])*0.5;
	  }
	  vv = vd[i][j][k];
	  if ((dimy > 1)&&(j<dimy-1)) {
	    vv = (vv+vd[i][j+1][k])*0.5;
	  }
	  ww = wd[i][j][k];
	  if ((dimz > 1)&&(k<dimz-1)) {
	    ww = (ww+wd[i][j][k+1])*0.5;
	  }
	  data[i][j][k] = sqrt(uu*uu + vv*vv + ww*ww);
	  if (i > 0) {
	    data[i][j][k] /= data[0][j][k];
	  }
	}
	data[0][j][k] = 1;
      }
    }

    speedup->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
    speedup->SetName("Speedup");
  }

  return speedup;
  
}

//
// Get the temperature perturbation at a point
//
DataSet *ReadFile::GetDeltaT() {
  DataSet *T, *Tref;
  int ndim, dimx, dimy, dimz;
  bldata *** t, ***tref, ***data;
  
  if (deltaT == NULL) {
    T = GetData("T");
    Tref = GetData("THETAR");
    if (!T || !Tref) {
      cerr << "Unable to read all required data" << endl;
      return NULL;
    }

    t = T->GetData();
    tref = Tref->GetData();

    ndim = T->GetDims();
    dimx = T->GetDim(1);
    dimy = T->GetDim(2);
    dimz = T->GetDim(3);

    // Create an empty data set
    deltaT = new DataSet(ndim, dimx, dimy, dimz);

    data = deltaT->GetData();
    
    // Calculate the temperature perturbation
    for (int i = 0;i<dimx;i++) {
      for (int j = 0;j<dimy;j++) {
	for (int k = 1;k<dimz;k++) {
	  data[i][j][k] = t[i][j][k] - tref[i][j][k];
	}
	// No thetar at the point below the surface - set perturbation to be zero.
	data[i][j][0] = 0;
      }
    }
    deltaT->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H, orog);
    deltaT->SetName("Temperature anomaly");
  }

  return deltaT;
  
}

//
// Swap the endianness of the data read in.
//
void ReadFile::SwapEndian(bool e) {
  if (IsOpen()) {
    cerr << "Cannot change endianness while file is open" << endl;
  }
  else {
    endian = e;
  }
}

//
// No Fortran records in the data file.
//
void ReadFile::NoFortranRecords(bool nr) {
  if (IsOpen()) {
    cerr << "Cannot change record type while file is open" << endl;
  }
  else {
    norec = nr;
  }
}

//
// Use LES average not instantaneous fields
//
void ReadFile::UseLESAv(bool les) {
  useLES = les;
}


//
// Get the depth of the boundary layer at a point
//
DataSet *ReadFile::GetBLDepth()
{
  bldata ***wsp, ***data, **oro;
  int ndim, nx, ny, nz, k;
  bldata umax, w1, w2;
  
  if (bldepth == NULL) {
    if (GetWindspeed() == NULL) {
      return bldepth;
    }    
    wsp = windspeed->GetData();
    ndim = windspeed->GetDims();
    nx = windspeed->GetDim(1);
    ny = windspeed->GetDim(2);
    nz = windspeed->GetDim(3);

    oro = orog->Get2D(3);

    bldepth = new DataSet(ndim-1,nx,ny,1);
    data = bldepth->GetData();

    for (int i = 0; i < nx; i++) {
      for (int j = 0; j < ny; j++) {
	umax = wsp[i][j][nz-1];
	k = nz-2;
	while ((k>=0) && (fabs(umax - wsp[i][j][k])/umax < 0.05)) {
	  k--;
	}
	if (k < 0) {
	  data[i][j][0] = 0.0;
	}
	else {
	  if (umax < wsp[i][j][k]) {
	    umax *= 1.05;
	  }
	  else {
	    umax *= 0.95;
	  }
	  w1 = (umax-wsp[i][j][k])/(wsp[i][j][k+1]-wsp[i][j][k]);
	  w2 = 1.0-w1;
	  data[i][j][0] = (zngrid[k+1]*w1+zngrid[k]*w2)*(1.0-oro[i][j]/H);
	}
      }
    }
    bldepth->SetName("BLDepth");
    bldepth->SetGrids(xngrid,yngrid,zgrid,nx,ny,nz,H, orog);
  }

  return bldepth;
}

//
// Calculate a streamline trajectory from the U, V and W fields.
//
void ReadFile::GetTraj(bldata x0, bldata y0, bldata z0, bldata dt, 
		       int nt, int nmax,bldata **xgr, bldata **ygr, 
		       bldata **zgr, bldata **sgr, bldata **hgr, 
		       int *npts, bool back) {

  bldata *xgrid, *ygrid, *zgrid, *hgrid, *dist;
  bldata x, y, z, s, dx, dy, dz, ds, u, v, w;
  bldata k1x, k1y, k1z;
  bldata k2x, k2y, k2z;
  bldata k3x, k3y, k3z;
  bldata k4x, k4y, k4z;
  int n, nstep, k;
  bool finish = false;
  bool skip = false;
  DataSet *uds, *vds, *wds;
  bldata cdx, cdy, cdz;

  if (useLES) {
    uds = GetData("UAV");
    vds = GetData("VAV");
    wds = GetData("WAV");
  }
  else {
    uds = GetData("U");
    vds = GetData("V");
    wds = GetData("W");
  }
  if (!uds || !vds || !wds) {
    cerr << "Unable to read all required data" << endl;
    return;
  }

  xgrid = new bldata[nmax];
  ygrid = new bldata[nmax];
  zgrid = new bldata[nmax];
  hgrid = new bldata[nmax];
  dist = new bldata[nmax];

  n = 1;
  nstep = 0;
  xgrid[0] = x0;
  ygrid[0] = y0;
  zgrid[0] = z0;
  dist[0] = 0;
  x = x0;
  y = y0;
  z = z0;
  s = 0;

  nt = 5;

  try {
    hgrid[0] = orog->GetVal2(x,y,0);
  }
  catch (exception e) {    
    cout << "Getting orog height 1 " << e.what() << endl;
    return;
  }

  if (nxg > 1) {
    cdx = nxg/(this->xgrid[nxg-1]-this->xgrid[0]);
    
  }
  else {
    cdx = 0.0;
  }
  if (nyg > 1) {
    cdy = nyg/(this->ygrid[nyg-1]-this->ygrid[0]);
  }
  else {
    cdy = 0.0;
  }

  if (back) cout << "Doing back trajectory" << endl;
  
  while ((n < nmax) && !finish) {
    
    skip = false;
    try {
      // Get velocity at the current point.
      u = uds->GetVal2(x,y,z);
      v = vds->GetVal2(x,y,z);
      w = wds->GetVal2(x,y,z);
      // *** Transfer the checking of periodic bc's to the DataSet class ***
      // Should make it work transparantly in all cases
    

      k = 1;
      while (this->zgrid[k] < z) {
	k++;
      }

      cdz = 1.0/(this->zgrid[k]-this->zgrid[k-1]);
      
      // Set dt based on CFL condition
      dt = 1.0/(fabs(u)*cdx + fabs(v)*cdy + fabs(w)*cdz);
      dt /= 20;
      
      // Get the associated displacements
      // Use midpoint method to do the intergration
      k1x = u*dt;
      k1y = v*dt;
      k1z = w*dt;

      // Do back trajectory if required
      if (back) {
	k1x = -k1x;
	k1y = -k1y;
	k1z = -k1z;
      }

      k2x = uds->GetVal2(x+0.5*k1x,y+0.5*k1y,z+0.5*k1z)*dt;
      k2y = vds->GetVal2(x+0.5*k1x,y+0.5*k1y,z+0.5*k1z)*dt;
      k2z = wds->GetVal2(x+0.5*k1x,y+0.5*k1y,z+0.5*k1z)*dt;
      
      if (back) {
	k2x = -k2x;
	k2y = -k2y;
	k2z = -k2z;
      }

      k3x = uds->GetVal2(x+0.5*k2x,y+0.5*k2y,z+0.5*k2z)*dt;
      k3y = vds->GetVal2(x+0.5*k2x,y+0.5*k2y,z+0.5*k2z)*dt;
      k3z = wds->GetVal2(x+0.5*k2x,y+0.5*k2y,z+0.5*k2z)*dt;

      if (back) {
	k3x = -k3x;
	k3y = -k3y;
	k3z = -k3z;
      }

      k4x = uds->GetVal2(x+k2x,y+k2y,z+k2z)*dt;
      k4y = vds->GetVal2(x+k2x,y+k2y,z+k2z)*dt;
      k4z = wds->GetVal2(x+k2x,y+k2y,z+k2z)*dt;

      if (back) {
	k4x = -k4x;
	k4y = -k4y;
	k4z = -k4z;
      }

    }
    catch (exception e) {
      //      cout << e.what() << endl;
      skip = true;
      if ((nyg > 1) && (y < this->ygrid[0])) {
	y += (this->ygrid[nyg-1] - this->ygrid[0]);
      }
      else if ((nyg > 1) && (y > this->ygrid[nyg-1])) {
	y -= (this->ygrid[nyg-1] - this->ygrid[0]);
      }
      else {
	finish = true;
	//	cerr << "x=" << x << " y=" << y << " z=" << z << " u=" << u
	//		  << " v=" << v << " w=" << w << endl;
      }    
    }
    
    
    if (! skip) {
      
      dx = (0.5*k1x + k2x + k3x + 0.5*k4x)/3.0;
      dy = (0.5*k1y + k2y + k3y + 0.5*k4y)/3.0;
      dz = (0.5*k1z + k2z + k3z + 0.5*k4z)/3.0;
      
      ds = sqrt(dx*dx+dy*dy+dz*dz);
      
      // Update location
      x += dx;
      // Only update y field if it is a 3D run
      // Nope! Might be useful in 2-D too.
      //      if (nyg > 1) {
      y += dy;
      //      }
      z += dz;
      s += ds;
      nstep++;              
  
      // Update number of steps and save if necessary
      if (nstep == nt) {
	xgrid[n] = x;
	ygrid[n] = y;
	zgrid[n] = z;
	try {
	  hgrid[n] = orog->GetVal2(x,y,0);
	}
	catch (exception e) {
	  cout << "Getting orog height 2 " << e.what() << endl;
	}

	dist[n] = s;
	//	cout << x << " " << y << " " << z << " " << u << " " << v <<
	//	  " " << w << " " << dt << " " << hgrid[n] << " " << s << 
	//	  " " << n << endl;
	n++;
	nstep = 0;
      }
    }
  }
      
      
  if (n == nmax) {
    cout << "Array filled calculating trajectory" << endl;
  }
  
  if (xgr)  *xgr = xgrid;
  if (ygr)  *ygr = ygrid;
  if (zgr)  *zgr = zgrid;
  if (sgr)  *sgr = dist;
  if (hgr)  *hgr = hgrid;  
  *npts = n;

  cout << "Trajectory completed with " << n << " points." << endl;
  
}

//
// Calculate a linear streamline trajectory from the U and V fields.
// i.e. remaining on model surfaces.
//
void ReadFile::GetLinearTraj(bldata x0, bldata y0, bldata z0, bldata dt, 
		       int nt, int nmax,bldata **xgr, bldata **ygr, 
		       bldata **zgr, bldata **sgr, bldata **hgr, 
		       int *npts, bool back) {

  bldata *xgrid, *ygrid, *zgrid, *hgrid, *dist;
  bldata x, y, z, h, H, s, dx, dy, ds, u, v;
  bldata k1x, k1y;
  bldata k2x, k2y;
  bldata k3x, k3y;
  bldata k4x, k4y;
  int n, nstep, k;
  bool finish = false;
  bool skip = false;
  DataSet *uds, *vds;
  bldata cdx, cdy;

  if (useLES) {
    uds = GetData("UAV");
    vds = GetData("VAV");
  }
  else {
    uds = GetData("U");
    vds = GetData("V");
  }
  if (!uds || !vds ) {
    cerr << "Unable to read all required data" << endl;
    return;
  }

  xgrid = new bldata[nmax];
  ygrid = new bldata[nmax];
  zgrid = new bldata[nmax];
  hgrid = new bldata[nmax];
  dist = new bldata[nmax];

  n = 1;
  nstep = 0;
  xgrid[0] = x0;
  ygrid[0] = y0;
  zgrid[0] = z0;
  dist[0] = 0;
  x = x0;
  y = y0;
  s = 0;
  H = this->zngrid[nzg-1];

  nt = 5;

  // Get orography height at first point
  try {
    h = orog->GetVal2(x,y,0);
    hgrid[0] = h;
  }
  catch (exception e) {    
    cout << "Getting orog height 1 " << e.what() << endl;
    return;
  }
  // Get z model level
  z = (z0-h)/(1-h/H);

  if (nxg > 1) {
    cdx = nxg/(this->xgrid[nxg-1]-this->xgrid[0]);
    
  }
  else {
    cdx = 0.0;
  }
  if (nyg > 1) {
    cdy = nyg/(this->ygrid[nyg-1]-this->ygrid[0]);
  }
  else {
    cdy = 0.0;
  }

  if (back) cout << "Doing back trajectory" << endl;
  
  while ((n < nmax) && !finish) {
    
    skip = false;
    try {
      // Get velocity at the current point.
      u = uds->GetVal2(x,y,z,true);
      v = vds->GetVal2(x,y,z,true);
      // *** Transfer the checking of periodic bc's to the DataSet class ***
      // Should make it work transparantly in all cases
    

      k = 1;
      while (this->zgrid[k] < z) {
	k++;
      }

      // Set dt based on CFL condition
      dt = 1.0/(fabs(u)*cdx + fabs(v)*cdy);
      dt /= 20;
      
      // Get the associated displacements
      // Use midpoint method to do the intergration
      k1x = u*dt;
      k1y = v*dt;

      // Do back trajectory if required
      if (back) {
	k1x = -k1x;
	k1y = -k1y;
      }

      k2x = uds->GetVal2(x+0.5*k1x,y+0.5*k1y,z,true)*dt;
      k2y = vds->GetVal2(x+0.5*k1x,y+0.5*k1y,z,true)*dt;
      
      if (back) {
	k2x = -k2x;
	k2y = -k2y;
      }

      k3x = uds->GetVal2(x+0.5*k2x,y+0.5*k2y,z,true)*dt;
      k3y = vds->GetVal2(x+0.5*k2x,y+0.5*k2y,z,true)*dt;

      if (back) {
	k3x = -k3x;
	k3y = -k3y;
      }

      k4x = uds->GetVal2(x+k2x,y+k2y,z,true)*dt;
      k4y = vds->GetVal2(x+k2x,y+k2y,z,true)*dt;

      if (back) {
	k4x = -k4x;
	k4y = -k4y;
      }

    }
    catch (exception e) {
      //      cout << e.what() << endl;
      skip = true;
      if ((nyg > 3) && (y < this->ygrid[0])) {
	y += (this->ygrid[nyg-1] - this->ygrid[0]);
      }
      else if ((nyg > 3) && (y > this->ygrid[nyg-1])) {
	y -= (this->ygrid[nyg-1] - this->ygrid[0]);
      }
      else {
	finish = true;
	//	cerr << "x=" << x << " y=" << y << " z=" << z << " u=" << u
	//		  << " v=" << v << " w=" << w << endl;
      }    
    }
    
    
    if (! skip) {
      
      dx = (0.5*k1x + k2x + k3x + 0.5*k4x)/3.0;
      dy = (0.5*k1y + k2y + k3y + 0.5*k4y)/3.0;
      
      ds = sqrt(dx*dx+dy*dy);
      
      // Update location
      x += dx;
      // Only update y field if it is a 3D run
      // Nope! Might be useful in 2-D too.
      //      if (nyg > 1) {
      y += dy;
      //      }
      s += ds;
      nstep++;              
  
      // Update number of steps and save if necessary
      if (nstep == nt) {
	xgrid[n] = x;
	ygrid[n] = y;
	try {
	  hgrid[n] = orog->GetVal(x,y,0);
	}
	catch (exception e) {
	  cout << "Getting orog height 2 " << e.what() << endl;
	}
	zgrid[n] = hgrid[n] + z*(1-hgrid[n]/H);	
	dist[n] = s;
	//	cout << x << " " << y << " " << z << " " << u << " " << v <<
	//	  " " << dt << " " << hgrid[n] << " " << s << 
	//	  " " << n << endl;
	n++;
	nstep = 0;
      }
    }
  }
      
      
  if (n == nmax) {
    cout << "Array filled calculating trajectory" << endl;
  }
  
  if (xgr)  *xgr = xgrid;
  if (ygr)  *ygr = ygrid;
  if (zgr)  *zgr = zgrid;
  if (sgr)  *sgr = dist;
  if (hgr)  *hgr = hgrid;  
  *npts = n;

  //  cout << x << " " << y << " " << z << " " << u << " " << v <<
  //    " " << dt << " " << hgrid[n] << " " << s << 
  //    " " << n << endl;

  cout << "Trajectory completed with " << n << " points." << endl;
  
}

//
// Get the divergence of the turbulent momentum flux for
// use in turbulent Bernoulli equation.
//
DataVect *ReadFile::GetTurbDiv() {

  //  DataSet *tmp1, *tmp2, *tmp3;
  //  DataSet *d1, *d2, *d3;

  // Check if already calculated
  if (turbdiv != NULL) {
    return turbdiv;
  }

  //  tmp1 = GetData("U2BAR");

  return turbdiv;
}

// 
// Calculate the real height on model surfaces
//
DataSet *ReadFile::GetHeight() {
  
  bldata ***h;
  bldata **oro;
  bldata H;
  
  if (height != NULL) 
    return height;

  // Create the data set for the output
  height = new DataSet(3, nxg, nyg, nzg);
  h = height->GetData();
  oro = orog->Get2D(3);

  H = GetParm("H");

  for (int k=0; k<nzg; k++) {
    for (int j=0; j<nyg; j++) {
      for (int i=0; i<nxg; i++) {
	h[i][j][k] = oro[i][j] + zngrid[k]*(1-oro[i][j]/H);
      }
    }
  }

  height->SetGrids(xngrid,yngrid,zngrid,nxg,nyg,nzg,H, orog);
  height->SetName("Height");
  
  return height;
}

//
// Read in an external data set from an ASCII file
//
DataSet * ReadFile::GetASCIIData(const char *fname) {
  izstream afile;
  DataSet *dptr;
  int ni, nj, nk, nlines;
  bldata x,y,z,d,h, x1, y1, z1;
  bldata *xgr, *ygr, *zgr, **ogr, ***data;

  afile.open(fname,ios::in);

  if (!afile.is_open()) {
    cerr << "Error opening ASCII file" << endl;
    return NULL;
  }

  // First determine the dimensions of the data set
  // and to basic sanity check on file size.
  ni = 0;
  nj = 0;
  nk = 0;
  nlines = 0;
  while (!afile.eof()) {
    afile >> x;
    afile >> y;
    afile >> z;
    afile >> d;
    afile >> h;
    if (!afile.eof()) {
      if (nlines == 0) {
	x1 = x;
	y1 = y;
	z1 = z;
      }
      nlines++;
      if ((y == y1) && (z == z1)) {
	ni++;
      }
      if ((x == x1) && (z == z1)) {
	nj++;
      }
      if ((x == x1) && (y == y1)) {
	nk++;
      }
    }
  }
  if (nlines != ni*nj*nk) {
    cerr << "Error in ASCII file format" << endl;
    afile.close();
    return NULL;
  }
  
  // Return to beginning of file to read in data
  afile.seekg(0,ios::beg);
  afile.clear();

  // Create arrays for grids, data and orography
  xgr = new bldata[ni];
  ygr = new bldata[nj];
  zgr = new bldata[nk];
  ogr = AllocArray<bldata>(ni,nj);
  data = Alloc3DArray<bldata>(ni,nj,nk);
  
  for (int k=0;k<nk;k++) {
    for (int j=0;j<nj;j++) {
      for (int i=0;i<ni;i++) {
	afile >> x;
	afile >> y;
	afile >> z;
	afile >> d;
	afile >> h;
	//	cout << i << " " << j << " " << k << " " << x << " " << y << " " << z << " " << d << " " << h << endl;
	data[i][j][k] = d;
	if (k == 0) {
	  ogr[i][j] = h;
	  if (j == 0) {
	    xgr[i] = x;
	  }
	}
      }
      if (k == 0) {
	ygr[j] = y;
      }
    }
    zgr[k] = z;
  }

  dptr = new DataSet(data,3,ni,nj,nk,xgr,ygr,zgr,zgr[nk-1],ogr);
  dptr->SetName("ASCII data set");

  return dptr;
}

//
// Get the pressure drag on the hill
// 
bldata ReadFile::GetDrag() {
  DataSet *pd, *ud, *vd, *uid, *vid, *cand;
  bldata ***p, **o, *xgr, *ygr, *zgr, ***u, ***v, ***ui, ***vi, ***canden ;
  bldata dragx, dragy, psur, dhx, dhy, cx, cy;
  bldata cdragx, cdragy, canht, candn, cancd;
  bldata sdragx, sdragy, blenx;
  bldata u1, v1, speed, speedi, dx, dy, dz;
  int nx, ny, nz, i1, i2, j1, j2, k;
  bool canopy;

  if (useLES) 
    pd = GetData("PRAV");
  else
    pd = GetData("P");
  if ( !pd ) {
    cerr << "Unable to read pressure" << endl;
    return 0; 
  }
  
  p = pd->GetData();
  o = pd->GetOrog();
  nx = pd->GetDim(1);
  ny = pd->GetDim(2);
  nz = pd->GetDim(3);
  xgr = pd->GetGrid(1);
  ygr = pd->GetGrid(2);
  zgr = pd->GetGrid(3);

  dragx = 0;
  dragy = 0;
  cdragx = 0;
  cdragy = 0;

  canht = GetParm("CNPYHT");
  candn = GetParm("CNPYDN");
  cancd = GetParm("CNPYCD");
  cand = GetData("CDADEN");
  if (cand != NULL) {
    canden = cand->GetData();
  }
  else {
    canden = NULL;
  }
  
  blenx = GetParm("BLENX");
  sdragx = blenx*GetParm("TAUXAV");
  sdragy = blenx*GetParm("TAUYAV");

  if ((canht > 0) || (canden)) {
    canopy = 1;
  }
  else {
    canopy = 0;
  }

  if (canopy) {
    if (useLES) {
      ud = GetData("UAV");
      vd = GetData("VAV");
    }
    else {
      ud = GetData("U");
      vd = GetData("V");
    }
    uid = GetData("UI");
    vid = GetData("VI");
    if ( !ud || !vd || !uid || !vid ) {
      cerr << "Unable to read wind speeds" << endl;
      return 0; 
    }
    u = ud->GetData();
    v = vd->GetData();  
    ui = uid->GetData();
    vi = vid->GetData();  
  }
  
  for (int i=0;i<nx;i++) {   
    if (i == 0) {
      i1 = 0;
    }
    else {
      i1 = i-1;
    }
    if (i == nx-1) {
      i2 = nx-1;
    }
    else {
      i2 = i+1;
    }
    cx = (i2==i1)?0:1.0/(i2-i1);
    if (cx == 0) {
      dx = 1.0;
    }
    else {
      dx = cx*(xgr[i2]-xgr[i1]);
    }
    for (int j=0;j<ny;j++) {
      if (j == 0) {
	j1 = 0;
      }
      else {
	j1 = j-1;
      }
      if (j == ny-1) {
	j2 = ny-1;
      }
      else {
	j2 = j+1;
      }
      cy = (j2==j1)?0:1.0/(j2-j1);
      if (cy == 0) {
	dy = 1.0;
      }
      else {
	dy = cy*(ygr[j2]-ygr[j1]);
      }
      dhx = (o[i2][j]-o[i1][j])*cx;
      dhy = (o[i][j2]-o[i][j1])*cy;
      psur = 0.5*(p[i][j][0]+p[i][j][1]);
      dragx += psur*dhx*dy;
      dragy += psur*dhy*dx;

      //      cout << psur << " " << dx << " " << dy << " " << dhx << endl;
      //      cout << i1 << " " << i2 << " " << cx << " " << cy << endl;

      if (canden) {
	k = 0;
	while (canden[i][j][k] > 0 && k < nz-1) {
	  u1 = 0.5*(u[i][j][k]+u[i2][j][k]);
	  v1 = 0.5*(v[i][j][k]+v[i][j2][k]);
	  speed = sqrt(u1*u1+v1*v1);
	  speedi = sqrt(ui[0][0][k]*ui[0][0][k]+vi[0][0][k]*vi[0][0][k]);
	  dz = zgrid[k+1]-zgrid[k];
	  cdragx += canden[i][j][k]*(u1*speed-ui[0][0][k]*speedi)*dx*dy*dz;
	  cdragy += canden[i][j][k]*(v1*speed-vi[0][0][k]*speedi)*dx*dy*dz;
	  k++;
	}
      }
      else if (canopy) {
	k = 0;
	while (zgr[k] <= canht && k < nz-1) {
	  u1 = 0.5*(u[i][j][k]+u[i2][j][k]);
	  v1 = 0.5*(v[i][j][k]+v[i][j2][k]);
	  speed = sqrt(u1*u1+v1*v1);
	  speedi = sqrt(ui[0][0][k]*ui[0][0][k]+vi[0][0][k]*vi[0][0][k]);
	  dz = zgrid[k+1]-zgrid[k];
	  cdragx += cancd*candn*(u1*speed-ui[0][0][k]*speedi)*dx*dy*dz;
	  cdragy += cancd*candn*(v1*speed-vi[0][0][k]*speedi)*dx*dy*dz;
	  k++;
	}
      }
    }
  }    
  

  cout << "Pressure drag x " << dragx << "  y " << dragy << endl;
  cout << "Canopy drag x " << cdragx << "  y " << cdragy << endl;
  cout << "Surface drag x " << sdragx << "  y " << sdragy << endl;
  return dragx;
  
}


void ReadFile::SetVersion(int v, modeltype m) {
  ver = v;
  model = m;
}
