// -*-c++-*-
// 
// blasvis++ - C++ version of graphics routines
//
// ReadFile : Abstract class to read the information in from a
//           data file.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#ifndef _READFILE_H
#define _READFILE_H

#include "blastype.h"
#include "dataset.h"
#include "timeseries.h"
#include "datavect.h"
#include "zstream.h"
#include "arrayutil.h"
#include "models.h"

struct HeaderItem {
  char name[10];
  int id;
  int ndim;
  int dim[3];
  long offset;
  DataSet *data;
  TimeSeries *tdata;
  bldata value;
  HeaderItem *next;
};



class ReadFile : public ArrayUtil {
 public:
  ReadFile();
  ReadFile(const char*);
  virtual ~ReadFile();

  void SetVersion(int,modeltype);
  void Clear();
  virtual void CloseFile() {};
  virtual void DisplayParms() = 0;
  void DisplayHeader();
  HeaderItem *GetHeaderItem(DataSet *);
  HeaderItem *GetHeaderItem(const char *);
  HeaderItem *GetFirstHeaderItem();
  bool OpenFile(const char*);
  bool IsOpen();
  void SwapEndian(bool);
  void NoFortranRecords(bool);
  void UseLESAv(bool);

  virtual DataSet *GetData(const char[]) = 0;
  DataSet *GetASCIIData(const char*);
  virtual bldata GetParm(const char[]) = 0;
  DataSet *GetPV();
  DataSet *GetPVAlt();
  DataVect *GetPVFlux();
  DataVect *GetPVFluxAlt();
  DataVect *GetGradT();
  DataSet *GetBernoulli();
  DataSet *GetWindspeed();
  DataVect *GetTurbDiv();
  virtual DataSet *GetSpeedup();
  DataSet *GetDeltaT();
  virtual DataSet *GetPressureCorr() = 0;
  virtual DataVect *GetTau() = 0;
  virtual DataSet *GetPVSource() = 0;
  DataSet *GetHeight();
  DataSet *GetTimestep();
  DataSet *GetBLDepth();
  DataSet *GetStreamFn();
  bldata GetDrag();
  void GetTraj(bldata, bldata, bldata, bldata, int, int, bldata **, 
	       bldata **, bldata **, bldata **, bldata **, int *, bool = false);
  void GetLinearTraj(bldata, bldata, bldata, bldata, int, int, bldata **, 
	       bldata **, bldata **, bldata **, bldata **, int *, bool = false);
  virtual DataSet *GetLESuw() = 0;
  virtual DataSet *GetLESuu() = 0;
  virtual DataSet *GetLESvv() = 0;
  virtual DataSet *GetLESww() = 0;
  virtual DataSet *GetLESenergy() = 0;

  // Protected member functions for class
 protected:
  virtual void InitFile() = 0;
  // Template functions

  // Read in binary data from file
  template <class T> T ReadFile::BinRead() {
    int tlen = sizeof(T);
    char tmp[tlen];
    char tmpswap;
    
    file.read(tmp,sizeof(T));    
    // Swap byte order if necessary
    if (endian) {
      for (int i = 0; i<tlen/2; i++) {
	tmpswap = tmp[i];
	tmp[i] = tmp[tlen-1-i];
	tmp[tlen-1-i] = tmpswap;
      }
    } 
    return *((T *) tmp);
  };

  // Protected data for class

  // Pointer to first header item
  HeaderItem *firstItem;
  HeaderItem *lastItem;
  int numitems;

  // Name of data file and associated ifstream 
  izstream file;

  // Pointers for calculated quantities 
  DataSet *pv, *pvalt, *bernoulli, *windspeed, *speedup, *deltaT, *pcorr,
    *timestep, *bldepth, *streamfn, *PVsrc, *height, *uw_res, *uu_res, 
    *vv_res, *ww_res, *e_res;
  DataVect *pvflux, *pvfluxalt, *gradt, *turbdiv;

  // Record length
  long irecl;
  std::streampos start;

  // Swap the endianness of data
  bool endian;
  // Don't use fortran records
  bool norec;
  int reclen;
  // Use LES average not instantaneous fields
  bool useLES;

  // Grids and orography
  bldata *xgrid,*xngrid;
  bldata *ygrid,*yngrid;
  bldata *zgrid,*zngrid;
  bldata *nullgrid;
  int nxg,nyg,nzg;
  DataSet *orog;
  bldata H;
  
  // version and model information 
  int ver;
  modeltype model;
};

#endif

