//
// blasvis++ - C++ version of graphics routines
//
// Vomfile : Class to read the information in from a 
//            3dnvom1.6 restart file.
//
// (C) 2001 Andrew Ross <aross@env.leeds.ac.uk>

#include "vomfile.h"
#include "models.h"
#include <iostream>
#include <fstream>
#include <cctype>
#include <cstring>
#include <cmath>

using namespace std;

// Default constructor
VomFile::VomFile() : ReadFile() {
  u = NULL;
  v = NULL;
  w = NULL;
  uprime = NULL;
  vprime = NULL;
  ui = NULL;
  vi = NULL;
  temp = NULL;
  pressure = NULL;
  rho = NULL;
  theta0 = NULL;
  visco = NULL;

  vortx = NULL;
  vorty = NULL;
  vortz = NULL;

  tau11 = NULL;
  tau22 = NULL;
  tau33 = NULL;
  tau12 = NULL;
  tau13 = NULL;
  tau23 = NULL;

  ver = 265;
  model = vom;
}

// Constructor with a file name
VomFile::VomFile(char* fname) : ReadFile(fname) {
}

// destructor
VomFile::~VomFile() {
  CloseFile();
  delete[] nullgrid;
}

// 
// Do any initialisation associated with opening the file
//
void VomFile::InitFile() {
  ReadData();
  CreateGrids();
}

//
// Close the current file
//
void VomFile::CloseFile() {

  if (!file.is_open()) 
    return;

  file.close();

  // Delete read in data sets
  if (u) {
    delete u;
    u = NULL;
  }
  if (v) {
    delete v;
    v = NULL;
  }
  if (w) {
    delete w;
    w = NULL;
  }
  if (uprime) {
    delete uprime;
    uprime = NULL;
  }
  if (vprime) {
    delete vprime;
    vprime = NULL;
  }
  if (ui) {
    delete u;
    u = NULL;
  }
  if (vi) {
    delete v;
    v = NULL;
  }
  if (temp) {
    delete temp;
    temp = NULL;
  }
  if (pressure) {
    delete pressure;
    pressure = NULL;
  }
  if (rho) {
    delete rho;
    rho = NULL;
  }
  if (theta0) {
    delete theta0;
    theta0 = NULL;
  }
  if (visco) {
    delete visco;
    visco = NULL;
  }

  // Delete any derived data sets
  if (pv) {
    delete pv;
    pv = NULL;
  }
  if (pvflux) {
    pvflux->Delete();
    delete pvflux;
    pvflux = NULL;
  }  
  if (gradt) {
    gradt->Delete();
    delete gradt;
    gradt = NULL;
  }  
  if (pvfluxalt) {
    pvfluxalt->Delete();
    delete pvfluxalt;
    pvflux = NULL;
  }
  if (bernoulli) {
    delete bernoulli;
    bernoulli = NULL;
  }
  if (windspeed) {
    delete windspeed;
    windspeed = NULL;
  }
  if (speedup) {
    delete speedup; 
    speedup = NULL;
  }
  if (deltaT) {
    delete deltaT;
    deltaT = NULL;
  }
  if (pcorr) {
    delete pcorr;
    pcorr = NULL;
  }
  if (timestep) {
    delete timestep;
    timestep = NULL;
  }
  if (bldepth) {
    delete bldepth;
    bldepth = NULL;
  }

  delete[] xgrid;
  delete[] xngrid;
  delete[] ygrid;
  delete[] yngrid;
  delete[] zgrid;
  delete[] zngrid;
  delete[] orog;

  xgrid = NULL;
  xngrid = NULL;
  ygrid = NULL;
  yngrid = NULL;
  zgrid = NULL;
  zngrid = NULL;
  orog = NULL;

}

//
// Read in the data from file
//
void VomFile::ReadData() {
  int nx, ny, nz;
  long offset;
  bldata ***ud, ***vd, ***wd, ***tempd, ***pd, ***od, ***t0d, ***rhod, ***visd;
  bldata ***umd, ***vmd, ***workd, ***upd, ***vpd;
  bldata dx, dy, xmin, ymin, ***data;
  
  file.seekg(0,ios::beg);

  // Read in record length 
  BinRead<long>();
  nx = (int) BinRead<long>();
  ny = (int) BinRead<long>();
  nz = (int) BinRead<long>();

  // Linear version of the code 3Dvom
  if (model == vom) {
    // Get number of steps
    nstep = (bldata) BinRead<long>();    
    // Get dx and dy values
    dx = GetVal();
    dy = GetVal();
    // Get misc other parameters
    dt = GetVal();
    // Z0 is a 3D field for more recent versions
    if (ver >= 271) {
      z0 = 0;
      SkipArray(nx+2,ny+2,1,1);
    } 
    else {
      z0 = GetVal();
    }
    fcor = GetVal();
    umd = GetArray(1,1,nz+2);
    vmd = GetArray(1,1,nz+2);
    workd = GetArray(1,1,nz+2);
    // Create 3-D theta0 field
    t0d = Alloc3DArray<bldata>(nx+2,ny+2,nz+2);
    for (int i=0;i<nx+2;i++) {
      for (int j=0;j<ny+2;j++) {
	for (int k=0;k<nz+2;k++) {
	  t0d[i][j][k] = workd[0][0][k];
	}
      }
    }
    Free3DArray(workd);
    // Get rho
    workd = GetArray(1,1,nz+2); 
    // Create 3-D rho field
    rhod = Alloc3DArray<bldata>(nx+2,ny+2,nz+2);
    for (int i=0;i<nx+2;i++) {
      for (int j=0;j<ny+2;j++) {
	for (int k=0;k<nz+2;k++) {
	  rhod[i][j][k] = workd[0][0][k];
	}
      }
    }
    Free3DArray(workd);
    // Get z grid
    zngrid = GetVector(nz+2);
    visd = GetArray(nx+2,ny+2,nz+2);

    // Get u,v,w and temperature
    ud = GetArray(nx+2,ny+2,nz+2);
    vd = GetArray(nx+2,ny+2,nz+2);
    wd = GetArray(nx+2,ny+2,nz+2);
    tempd = GetArray(nx+2,ny+2,nz+2);
    SkipArray(nx+2,ny+2,nz+2,9);

    // Get mean part of viscosity
    workd = GetArray(nx+2,ny+2,nz+2); 
    for (int i=0;i<nx+2;i++) {
      for (int j=0;j<ny+2;j++) {
	for (int k=0;k<nz+2;k++) {
	  visd[i][j][k] += workd[0][0][k];
	}
      }
    }
    Free3DArray(workd);

    // Get pressure
    pd = GetArray(nx+2,ny+2,nz+2);

    // Get orography and copy into orography array
    od = GetArray(nx+2,ny+2,1);

    // Skip over a and al
    SkipArray(1,1,nz+2,1);
    SkipArray(nx+2,ny+2,1,1);

    // Skip over alpha, beta, gamma, delta
    SkipArray(nx+2,ny+2,nz+2,4);

    upd = Alloc3DArray<bldata>(nx+2,ny+2,nz+2);
    vpd = Alloc3DArray<bldata>(nx+2,ny+2,nz+2);
    for (int i=0;i<nx+2;i++) {
      for (int j=0;j<ny+2;j++) {
	for (int k=0;k<nz+2;k++) {
	  upd[i][j][k] = ud[i][j][k];
	  vpd[i][j][k] = vd[i][j][k];
	  ud[i][j][k] += umd[0][0][k];
	  vd[i][j][k] += vmd[0][0][k];
	  //	  tempd[i][j][k] += t0d[0][0][k];
	}
      }
    }
  }
  // Non-Linear version of the code 3Dnvom
  else {
    // Get u,v,w and temperature
    ud = GetArray(nx+2,ny+2,nz+2);
    vd = GetArray(nx+2,ny+2,nz+2);
    wd = GetArray(nx+2,ny+2,nz+2);
    tempd = GetArray(nx+2,ny+2,nz+2);

    // Skip over u2...t2 arrays
    SkipArray(nx+2,ny+2,nz+2,4);
  
    // Get pressure
    pd = GetArray(nx+2,ny+2,nz+2);

    // Skip over alpha, beta, gamma, delta
    SkipArray(nx+2,ny+2,nz+2,4);
  
    // Get orography and copy into orography array
    od = GetArray(nx+2,ny+2,1);

    // Skip over a and ar
    SkipArray(1,1,nz+2,1);
    SkipArray(nx+2,ny+2,1,1);
    
    // Get reference temperature
    t0d = GetArray(nx+2,ny+2,nz+2);
    
    // Skip over ug and vg
    SkipArray(nx+2,ny+2,nz+2,2);
    
    // Get rho
    rhod = GetArray(nx+2,ny+2,nz+2);    
    
    // Skip over rho0 and pg
    SkipArray(1,1,nz+2,1);
    SkipArray(nx+2,ny+2,nz+2,1);
    
    // Skip p0, thetam
    GetVal();
    GetVal();
    
    // Get the coordinate arrays
    dx = GetVal();
    dy = GetVal();
    zgrid = GetVector(nz+2);
    GetVal();
    zngrid = GetVector(nz+2);

    // Get misc other parameters
    dt = GetVal();
    fcor = GetVal();
    

    offset = file.tellg();
    if ((offset + sizeof(long)) % (irecl+2*sizeof(long)) == 0) {
      offset = offset + 2*sizeof(long);
      file.seekg(2*sizeof(long),ios::cur);	  
    }
    nstep = (bldata) BinRead<long>();
    
    visd = GetArray(nx+2,ny+2,nz+2);
  }
    
  time = nstep*dt;

  if (model = nvom) {
    // Create the staggered x and y grids
    xgrid = new bldata[nx+2];
    xngrid = new bldata[nx+2];
    ygrid = new bldata[ny+2];
    yngrid = new bldata[ny+2];
    xmin = -(nx+2)*dx/2.0;
    ymin = -(ny+2)*dy/2.0;  
    for (int i=0;i<nx+2; i++) {
      xngrid[i] = xmin + i*dx;
      xgrid[i] = xngrid[i] - dx/2;
    }
    for (int j=0;j<ny+2; j++) {
      yngrid[j] = ymin + j*dy;
      ygrid[j] = yngrid[j] - dy/2;
    }
  }
  else {
    // non-staggered grids
    xmin = -(nx+2)*dx/2.0;
    ymin = -(ny+2)*dy/2.0;  
    xngrid = new bldata[nx+2];
    yngrid = new bldata[ny+2];
    for (int i=0;i<nx+2; i++) {
      xngrid[i] = xmin + i*dx;
    }
    for (int j=0;j<ny+2; j++) {
      yngrid[j] = ymin + j*dy;
    }
    xgrid = xngrid;
    ygrid = yngrid;
    zgrid = zngrid;    

    ui = new DataSet(umd,1,1,1,nz+2);
    ui->SetGrids(nullgrid,nullgrid,zngrid,1,1,nz+2,H,orog);
    ui->SetName("UI");
    vi = new DataSet(vmd,1,1,1,nz+2);
    vi->SetGrids(nullgrid,nullgrid,zngrid,1,1,nz+2,H,orog);
    vi->SetName("VI");
  }

  H = zgrid[nz-1];

  orog = new DataSet(od,2,nx+2,ny+2,1);
  orog->SetGrids(xngrid,yngrid,nullgrid,nx+2,ny+2,1,H,orog);

  u = new DataSet(ud,3,nx+2,ny+2,nz+2);
  u->SetGrids(xgrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  u->SetName("U");
  v = new DataSet(vd,3,nx+2,ny+2,nz+2);
  v->SetGrids(xngrid,ygrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  v->SetName("V");
  w = new DataSet(wd,3,nx+2,ny+2,nz+2);
  w->SetGrids(xngrid,yngrid,zgrid,nx+2,ny+2,nz+2,H,orog);
  w->SetName("W");
  if (upd && vpd) {
    uprime = new DataSet(upd,3,nx+2,ny+2,nz+2);
    uprime->SetGrids(xgrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
    uprime->SetName("UPRIME");
    vprime = new DataSet(vpd,3,nx+2,ny+2,nz+2);
    vprime->SetGrids(xngrid,ygrid,zngrid,nx+2,ny+2,nz+2,H,orog);
    vprime->SetName("VPRIME");
  }

  deltaT = new DataSet(tempd,3,nx+2,ny+2,nz+2);
  deltaT->SetGrids(xngrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  deltaT->SetName("Temperature anomoly");
  pressure = new DataSet(pd,3,nx+2,ny+2,nz+2);
  pressure->SetGrids(xngrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  pressure->SetName("P");
  rho = new DataSet(rhod,3,nx+2,ny+2,nz+2);
  rho->SetGrids(xngrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  rho->SetName("RHO");
  theta0 = new DataSet(t0d,3,nx+2,ny+2,nz+2);
  theta0->SetGrids(xngrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  theta0->SetName("THETAR");
  if (model == vom) {
    visco = new DataSet(visd,3,nx+2,ny+2,nz+2);
    visco->SetGrids(xngrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
    visco->SetName("VIS");
  }
  else {
    visco = NULL;
  }

  nxg = nx + 2;
  nyg = ny + 2;
  nzg = nz + 2;

  GetVorticity();
  CalcTau();

  // Create the temperature field as thetar + deltat
  temp = new DataSet(3,nx+2,ny+2,nz+2);
  data = temp->GetData();
  // Calculate the temperature
  for (int i = 0;i<nx+2;i++) {
    for (int j = 0;j<ny+2;j++) {
      for (int k = 1;k<nz+2;k++) {
	data[i][j][k] = tempd[i][j][k] + t0d[i][j][k];
      }
    }
  }
  temp->SetGrids(xngrid,yngrid,zngrid,nx+2,ny+2,nz+2,H,orog);
  temp->SetName("TEMP");  
}


//
// Read in a vector from file
//
bldata *VomFile::GetVector(int n) {
 bldata *dptr;
  long offset;

  offset = file.tellg();

  dptr = new bldata[n];

  for (int i=0;i<n;i++) {
    if ((offset + sizeof(long)) % (irecl+2*sizeof(long)) == 0) {
      offset = offset + 2*sizeof(long);
      file.seekg(2*sizeof(long),ios::cur);	  
    }
    dptr[i] = BinRead<double>();
    offset = offset + sizeof(double);
  }
  return dptr;
}

//
// Read in an array from file
//
bldata ***VomFile::GetArray(int nx,int ny,int nz) {
  bldata ***dptr;
  long offset;

  offset = file.tellg();

  dptr = (bldata ***) new bldata[nx][ny][nz];
  for (int i = 0;i<nx;i++) {
    dptr[i] = (bldata **) new bldata[ny][nz];
    for (int j=0;j<ny;j++) {
      dptr[i][j] = new bldata[nz];
    }
  }

  for (int k=0;k<nz;k++) {
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	if ((offset + sizeof(long)) % (irecl+2*sizeof(long)) == 0) {
	  offset = offset + 2*sizeof(long);
	  file.seekg(2*sizeof(long),ios::cur);	  
	}
	dptr[i][j][k] = BinRead<double>();
	//	cout << i << "," << j << "," << k << " " << dptr[i][j][k] << endl;
	//	  cout << hex << offset << dec << endl;
	offset = offset + sizeof(double);
	// Check if we have reached the end of a record and update
	// the file pointer accordingly.
      }
    }
  }
  
  return dptr;
}

//
// Read in a double from file with proper checking 
//
bldata VomFile::GetVal() {
  bldata val;
  long offset;

  offset = file.tellg();

  if ((offset + sizeof(long)) % (irecl+2*sizeof(long)) == 0) {
    offset = offset + 2*sizeof(long);
    file.seekg(2*sizeof(long),ios::cur);	  
  }
  val = BinRead<double>();
  return val;
}

//
// Skip over an array from file
//
void VomFile::SkipArray(int nx,int ny,int nz,int na) {
  long offset, size, tonext, nrec;

  offset = file.tellg();

  size = nx*ny*nz*na*sizeof(double);
  tonext = irecl + sizeof(long) - offset % (irecl + 2*sizeof(long));
  if (size >= tonext) {
    nrec = (size-tonext) / irecl + 1;
    size += nrec*2*sizeof(long);
  }
  file.seekg(size,ios::cur);  
}

//
// Get the specified dataset. If it is not cached then load it from file.
//
DataSet * VomFile::GetData(const char name[8]) {
  if (strncmp(name,"U",8) == 0) {
    return u;
  }
  else if (strncmp(name,"V",8) == 0) {
    return v;
  }
  else if (strncmp(name,"W",8) == 0) {
    return w;
  }
  else if (strncmp(name,"UPRIME",8) == 0) {
    return uprime;
  }
  else if (strncmp(name,"VPRIME",8) == 0) {
    return vprime;
  }
  else if (strncmp(name,"UI",8) == 0) {
    return ui;
  }
  else if (strncmp(name,"VI",8) == 0) {
    return vi;
  }
  else if (strncmp(name,"T",8) == 0) {
    return temp;
  }
  else if (strncmp(name,"P",8) == 0) {
    return pressure;
  }
  else if (strncmp(name,"THETAR",8) == 0) {
    return theta0;
  }
  else if (strncmp(name,"RHO",8) == 0) {
    return rho;
  }
  else if ((strncmp(name,"VIS",8) == 0) || 
	   (strncmp(name,"VISH",8) == 0)) {
    return visco;
  }
  else if (strncmp(name,"VORTX",8) == 0) {
    return vortx;
  }
  else if (strncmp(name,"VORTY",8) == 0) {
    return vorty;
  }
  else if (strncmp(name,"VORTZ",8) == 0) {
    return vortz;
  }
  else if (strncmp(name,"TAU11",8) == 0) {
    return tau11;
  }
  else if (strncmp(name,"TAU22",8) == 0) {
    return tau22;
  }
  else if (strncmp(name,"TAU33",8) == 0) {
    return tau33;
  }
  else if (strncmp(name,"TAU12",8) == 0) {
    return tau12;
  }
  else if (strncmp(name,"TAU13",8) == 0) {
    return tau13;
  }
  else if (strncmp(name,"TAU23",8) == 0) {
    return tau23;
  }
  else if (strncmp(name,"ZSN",8) == 0) {
    return orog;
  }
  return NULL;
}

//
// Get the specified parameter value
//
bldata VomFile::GetParm(const char name[8]) {
  if (strncmp(name,"FCOR",8) == 0) {
    return fcor;
  }
  else if (strncmp(name,"TIME",8) == 0) {
    return time;
  }
  else if (strncmp(name,"DT",8) == 0) {
    return dt;
  }
  else if (strncmp(name,"NX",8) == 0) {
    return nxg;
  }
  else if (strncmp(name,"NY",8) == 0) {
    return nyg;
  }
  else if (strncmp(name,"NZ",8) == 0) {
    return nzg;
  }
  else if (strncmp(name,"NSTEP",8) == 0) {
    return nstep;
  }
  else if (strncmp(name,"PR",8) == 0) {
    return 1.0;
  }
  else if (strncmp(name,"BUSDYR",8) == 0) {
    return 16.0;
  }
  else if (strncmp(name,"HTIMAV",8) == 0) {
    return 0.0;
  }
  else if (strncmp(name,"Z0",8) == 0) {
    return z0;
  }
  else if (strncmp(name,"H",8) == 0) {
    return zngrid[nzg-1];
  }
  
  return 0;
}


//
// Calculate the vorticity from the velocity
//
void VomFile::GetVorticity() {
  int dimx, dimy, dimz;
  bldata *** datax, ***datay, ***dataz, ***ud, ***vd, ***wd, ***od;
  bldata H, ghalf1, ghalf2, ghalf3, gh3z1, gh3z2, dxn, dyn, dzn, 
    ghalfx1, ghalfx2, ghalfy1, ghalfy2;
  
  if ((vortx == NULL) || (vorty == NULL) || (vortz == NULL)) {
    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();
    od = orog->GetData();

    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    H = zgrid[dimz-2];

    // Create an empty data set
    vortx = new DataSet(3, dimx, dimy, dimz);
    vorty = new DataSet(3, dimx, dimy, dimz);
    vortz = new DataSet(3, dimx, dimy, dimz);

    datax = vortx->GetData();
    datay = vorty->GetData();
    dataz = vortz->GetData();
    
    // interpolate components onto the (XN,YN,ZN) grid
    for (int i = 1;i<dimx;i++) {
      dxn = xngrid[i] - xngrid[i-1];
      for (int j = 1;j<dimy;j++) {
	// Work out the metric terms
	dyn = yngrid[j] - yngrid[j-1];
	ghalf1 = 1.0-od[i][j][0]/H;
	ghalf2 = 1.0-od[i][j-1][0]/H;
	ghalf3 = 1.0-od[i-1][j][0]/H;
	ghalfx2 = (1.0 - (od[i][j-1][0]+od[i-1][j-1][0])*0.5/H);
	ghalfy2 = (1.0 - (od[i-1][j][0]+od[i-1][j-1][0])*0.5/H);
	for (int k = 1;k<dimz;k++) {
	  // Do the vorticity calculation properly even though it is rather messy
	  dzn = zngrid[k] - zngrid[k-1];	
	  gh3z1 = zngrid[k]/H-1.0;
	  gh3z2 = zngrid[k-1]/H-1.0;

	  ghalfx1 = (1.0 - (od[i][j][0]+od[i-1][j][0])*0.5/H);
	  ghalfy1 = (1.0 - (od[i][j][0]+od[i][j-1][0])*0.5/H);

	  datax[i][j][k] = -(vd[i][j][k] - vd[i][j][k-1]) / dzn +
	    (ghalf1*wd[i][j][k] - ghalf2*wd[i][j-1][k]) / dyn + 
	    ( gh3z1*(wd[i][j][k]+wd[i][j-1][k]+wd[i][j][k-1]+wd[i][j-1][k-1]) -
	      gh3z2*(wd[i][j][k-1]+wd[i][j-1][k-1]+wd[i][j][k-2]+wd[i][j-1][k-2]) )
	    *0.25*(od[i][j][0]-od[i][j-1][0])/(dyn*dzn);
	  datax[i][j][k] /= ghalfy1;

	  datay[i][j][k] = (ud[i][j][k] - ud[i][j][k-1])/dzn - 
	    (ghalf1*wd[i][j][k] - ghalf2*wd[i-1][j][k])/dxn + 
	    (gh3z1*(wd[i][j][k]+wd[i-1][j][k]+wd[i][j][k-1]+wd[i-1][j][k-1]) -
	     gh3z2*(wd[i][j][k-1]+wd[i-1][j][k-1]+wd[i][j][k-2]+wd[i-1][j][k-2]))*
	    0.25*(od[i][j][0]-od[i-1][j][0])/(dxn*dzn);
	  datay[i][j][k] /= ghalfx1;

	  gh3z1 = zgrid[k]/H-1.0;
	  gh3z2 = zgrid[k-1]/H-1.0;
	  dataz[i][j][k] = -(ghalfx1*ud[i][j][k] - ghalfx2*ud[i][j-1][k])/dyn +
	    (ghalfy1*vd[i][j][k] - ghalfy2*vd[i-1][j][k])/dxn -

	    ( gh3z1*(ud[i][j][k] + ud[i-1][j][k] + ud[i][j][k+1] + ud[i-1][j][k+1]) -
	      gh3z2*(ud[i][j][k] + ud[i-1][j][k] + ud[i][j][k-1] + ud[i-1][j][k-1]) )*
	    0.125*(od[i][j][0]+od[i-1][j][0]-od[i][j-1][0]-od[i-1][j-1][0])/
	    (dyn*(zgrid[k]-zgrid[k-1])) - 

	    ( gh3z1*(vd[i][j][k] + vd[i][j-1][k] + vd[i][j][k+1] + vd[i][j-1][k+1]) -
	      gh3z2*(vd[i][j][k] + vd[i][j-1][k] + vd[i][j][k-1] + vd[i][j-1][k-1]) )*
	    0.125*(od[i][j][0]+od[i][j-1][0]-od[i-1][j][0]-od[i-1][j-1][0])/
	    (dxn*(zgrid[k]-zgrid[k-1]));
	  datay[i][j][k] /= ghalf1;
	}
      }
    }
    for (int i=1;i<dimx;i++) {
      for (int j=1;j<dimy;j++) {
	datax[i][j][0] = 0.0;
	datay[i][j][0] = 0.0;
	dataz[i][j][0] = -(ud[i][j][0] - ud[i][j-1][0])/(yngrid[j]-yngrid[j-1]) +
	  (vd[i][j][0] - vd[i-1][j][0])/(xngrid[i]-xngrid[i-1]);
      }
    }
    for (int i=1;i<dimx;i++) {
      for (int k=1;k<dimz;k++) {
	datax[i][0][k] = 0.0;
	datay[i][0][k] = -(wd[i][0][k] - wd[i-1][0][k])/(xngrid[i]-xngrid[i-1]) +
	  (ud[i][0][k] - ud[i][0][k-1])/(zngrid[k]-zngrid[k-1]);
	dataz[i][0][k] = 0.0;
      }
    }
    for (int j=1;j<dimy;j++) {
      for (int k=1;k<dimz;k++) {
	datax[0][j][k] = -(vd[0][j][k] - vd[0][j][k-1])/(zngrid[k]-zngrid[k-1]) +
	  (wd[0][j][k] - wd[0][j-1][k])/(yngrid[j]-yngrid[j-1]);
	datay[0][j][k] = 0.0;
	dataz[0][j][k] = 0.0;
      }
    }
    datax[0][0][0] = 0.0;
    datay[0][0][0] = 0.0;
    dataz[0][0][0] = 0.0;
    vortx->SetName("VORTX");
    vortx->SetGrids(xngrid,ygrid,zgrid,dimx,dimy,dimz,H,orog);
    vorty->SetName("VORTY");
    vorty->SetGrids(xgrid,yngrid,zgrid,dimx,dimy,dimz,H,orog);
    vortz->SetName("VORTZ");
    vortz->SetGrids(xgrid,ygrid,zngrid,dimx,dimy,dimz,H,orog);
  }
  
}

//
// Get the corrected pressure field
//
DataSet *VomFile::GetPressureCorr() {
  return pressure;
}


//
// Get the shear stresses tau_ij
//
void VomFile::CalcTau() {
  bldata ***ud, ***vd, ***wd, ***od, ***visd, ***rhod;
  bldata ***data11, ***data22, ***data33, ***data12, ***data13, ***data23;
  bldata H, dxn, dyn, dzn;
  bldata ghalf, cx, cy;
  int dimx, dimy, dimz;
  int i1, i2, j1, j2, k1, k2;

  // Check if already calculated
  if (tau11 == NULL) {
    ud = u->GetData();
    vd = v->GetData();
    wd = w->GetData();
    od = orog->GetData();
    visd = visco->GetData();
    rhod = rho->GetData();
    
    dimx = u->GetDim(1);
    dimy = u->GetDim(2);
    dimz = u->GetDim(3);

    // Create new data sets
    tau11 = new DataSet(3, dimx, dimy, dimz);
    tau22 = new DataSet(3, dimx, dimy, dimz);
    tau33 = new DataSet(3, dimx, dimy, dimz);
    tau12 = new DataSet(3, dimx, dimy, dimz);
    tau13 = new DataSet(3, dimx, dimy, dimz);
    tau23 = new DataSet(3, dimx, dimy, dimz);

    data11 = tau11->GetData();
    data22 = tau22->GetData();
    data33 = tau33->GetData();
    data12 = tau12->GetData();
    data13 = tau13->GetData();
    data23 = tau23->GetData();
    
    H = zgrid[dimz-2];

    for (int i=0; i<dimx;i++) {
      i1 = i+1;
      i2 = i-1;
      if (i1 == dimx) 
	i1 = dimx-1;
      if (i2 == -1)
	i2 = 0;
      dxn = xngrid[i1] - xngrid[i2];
      if (dxn < 1E-10) 
	dxn = 1;     
      for (int j = 0;j<dimy;j++) {
	j1 = j+1;
	j2 = j-1;
	if (j1 == dimy) 
	  j1 = dimy-1;
	if (j2 == -1)
	  j2 = 0;
	dyn = yngrid[j1] - yngrid[j2];
	if (dyn < 1E-10) 
	  dyn = 1.0;

	ghalf = 1.0 - od[i][j][0]/H;

	for (int k = 0;k<dimz;k++) {
	  k1 = k+1;
	  k2 = k-1;
	  if (k1 == dimz) 
	    k1 = dimz-1;
	  if (k2 == -1)
	    k2 = 0;
	  dzn = zngrid[k1] - zngrid[k2];
	  if (dzn < 1E-10) 
	    dzn = 1.0;

	  cx = -(1.0 - zngrid[k]/H)*(od[i1][j][k]-od[i2][j][k])/dxn;
	  cy = -(1.0 - zngrid[k]/H)*(od[i][j1][k]-od[i][j2][k])/dyn;

	  data11[i][j][k] = 2.0*rhod[i][j][k] * visd[i][j][k] * 
	    ( (ud[i1][j][k]-ud[i2][j][k])/dxn + 
	      cx/ghalf*(ud[i][j][k1]-ud[i][j][k2])/dzn );
	  data22[i][j][k] = 2.0*rhod[i][j][k] * visd[i][j][k] * 
	    ( (vd[i][j1][k]-vd[i][j2][k])/dyn +
	      cy/ghalf*(vd[i][j][k1]-vd[i][j][k2])/dzn );
	  data33[i][j][k] = 2.0*rhod[i][j][k] * visd[i][j][k] * 
	    (wd[i][j][k1]-wd[i][j][k2])/(dzn*ghalf);
	  data12[i][j][k] = rhod[i][j][k] * visd[i][j][k] * 
	    ( (ud[i][j1][k]-ud[i][j2][k])/dyn +
	      cy/ghalf*(ud[i][j][k1]-ud[i][j][k2])/dzn + 
	      (vd[i1][j][k]-vd[i2][j][k])/dxn +
	      cx/ghalf*(vd[i][j][k1]-vd[i][j][k2])/dzn );
	  data13[i][j][k] = rhod[i][j][k] * visd[i][j][k] * 
	    ( (ud[i][j][k1]-ud[i][j][k2])/(dzn*ghalf) + 
	      (wd[i1][j][k]-wd[i2][j][k])/dxn +
	      cx/ghalf*(wd[i][j][k1]-wd[i][j][2])/dzn );
	  data23[i][j][k] = rhod[i][j][k] * visd[i][j][k] * 
	    ( (vd[i][j][k1]-vd[i][j][k2])/(dzn*ghalf) + 
	      (wd[i][j1][k]-wd[i][j2][k])/dyn +
	      cy/ghalf*(wd[i][j][k1]-wd[i][j][k2])/dzn );
	  
	}
      }
    }
    tau11->SetName("TAU11");
    tau11->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);
    tau22->SetName("TAU22");
    tau22->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);
    tau33->SetName("TAU33");
    tau33->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);
    tau12->SetName("TAU12");
    tau12->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);
    tau13->SetName("TAU13");
    tau13->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);
    tau23->SetName("TAU23");
    tau23->SetGrids(xngrid,yngrid,zngrid,dimx,dimy,dimz,H,orog);

    
  }

}

DataVect *VomFile::GetTau() {

  return NULL;

}

//
// Create the grids
//
void VomFile::CreateGrids() {
}

//
// Display parms
//
void VomFile::DisplayParms() {
}

